package railcraft.common.blocks.detector;

import buildcraft.api.tools.IToolWrench;
import cpw.mods.fml.common.registry.GameRegistry;
import java.util.List;
import net.minecraft.block.Block;
import net.minecraft.block.BlockContainer;
import net.minecraft.block.BlockRail;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.world.IBlockAccess;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.block.material.Material;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.world.World;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.common.MinecraftForge;
import railcraft.common.api.core.items.ITrackItem;
import railcraft.common.core.RailcraftConstants;
import railcraft.common.items.IActivationBlockingItem;
import railcraft.common.modules.ModuleManager;
import railcraft.common.modules.ModuleManager.Module;
import railcraft.common.plugins.forge.CreativePlugin;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

public class BlockDetector extends BlockContainer {

    protected final static int[][] texture = new int[Detector.values().length][3];

    public enum Detector {

        ITEM,
        ANY,
        EMPTY,
        MOB,
        POWERED,
        PLAYER,
        EXPLOSIVE,
        ANIMAL,
        TANK,
        ADVANCED,
        ENERGY,
        AGE,
        TRAIN,
        SHEEP,
        VILLAGER,
        LOCOMOTIVE;

        public static Detector fromMeta(int meta) {
            if (meta >= Detector.values().length) {
                meta = 0;
            }
            return values()[meta];
        }
    }

    public BlockDetector(int blockID) {
        super(blockID, 12, Material.rock);

        for (int i = 0; i < texture.length - 1; i++) {
            for (int j = 0; j < texture[i].length; j++) {
                texture[i][j] = 13 + i * 16 + j;
            }
        }
        for (int j = 0; j < texture[15].length; j++) {
            texture[15][j] = 10 + 14 * 16 + j;
        }

        setResistance(4.5F);
        setHardness(2.0F);
        setStepSound(soundStoneFootstep);
        setRequiresSelfNotify();

        setCreativeTab(CreativePlugin.TAB);

        GameRegistry.registerTileEntity(TileDetector.class, "RCDetectorTile");
        GameRegistry.registerTileEntity(TileDetectorItem.class, "RCDetectorItemTile");
        GameRegistry.registerTileEntity(TileDetectorAdvanced.class, "RCDetectorAdvancedTile");
        GameRegistry.registerTileEntity(TileDetectorTank.class, "RCDetectorTankTile");
        GameRegistry.registerTileEntity(TileDetectorAnimal.class, "RCDetectorAnimalTile");
        GameRegistry.registerTileEntity(TileDetectorTrain.class, "RCDetectorTrainTile");
        GameRegistry.registerTileEntity(TileDetectorSheep.class, "RCDetectorSheepTile");
        GameRegistry.registerTileEntity(TileDetectorVillager.class, "RCDetectorVillagerTile");
        GameRegistry.registerTileEntity(TileDetectorLocomotive.class, "RCDetectorLocomotiveTile");

        MinecraftForge.setBlockHarvestLevel(this, "pickaxe", 2);
        MinecraftForge.setBlockHarvestLevel(this, "crowbar", 0);
    }

    @Override
    public String getTextureFile() {
        return RailcraftConstants.MAIN_TEXTURE_FILE;
    }

    public static String getBlockNameFromMetadata(int meta) {
        StringBuilder name = new StringBuilder("detector.");
        name.append(Detector.fromMeta(meta).name().toLowerCase());
        return name.toString();
    }

    @Override
    public boolean isBlockNormalCube(World world, int i, int j, int k) {
        return false;
    }

    @Override
    public boolean isBlockSolidOnSide(World world, int i, int j, int k, ForgeDirection side) {
        return true;
    }

    @Override
    public int damageDropped(int meta) {
        return meta;
    }

    @Override
    public TileEntity createNewTileEntity(World var1) {
        return null;
    }

    @Override
    public TileEntity createNewTileEntity(World var1, int meta) {
        switch (Detector.values()[meta]) {
            case ITEM:
                return new TileDetectorItem();
            case ADVANCED:
                return new TileDetectorAdvanced();
            case TANK:
                return new TileDetectorTank();
            case ANIMAL:
                return new TileDetectorAnimal();
            case TRAIN:
                return new TileDetectorTrain();
            case SHEEP:
                return new TileDetectorSheep();
            case VILLAGER:
                return new TileDetectorVillager();
            case LOCOMOTIVE:
                return new TileDetectorLocomotive();
            default:
                return new TileDetector();
        }
    }

    // Determine direction here
    @Override
    public void onBlockPlacedBy(World world, int i, int j, int k, EntityLiving entityliving) {
        TileEntity tile = world.getBlockTileEntity(i, j, k);
        if (tile instanceof TileDetector) {
            ((TileDetector) tile).direction = MiscTools.getSideClosestToPlayer(world, i, j, k, entityliving);
            ((TileDetector) tile).onBlockPlacedBy(entityliving);
        }
    }

    @Override
    public boolean onBlockActivated(World world, int x, int y, int z, EntityPlayer player, int side, float u1, float u2, float u3) {
        if (player.isSneaking()) {
            return false;
        }
        ItemStack current = player.getCurrentEquippedItem();
        if (current != null) {
            Item item = current.getItem();
            if (item instanceof IActivationBlockingItem) {
                return false;
            } else if (item instanceof IToolWrench) {
                if (Game.isNotHost(world)) {
                    return true;
                }
                IToolWrench wrench = (IToolWrench) item;
                if (wrench.canWrench(player, x, y, z)) {
                    TileEntity t = world.getBlockTileEntity(x, y, z);
                    if (t instanceof TileDetector) {
                        TileDetector tile = (TileDetector) t;
                        if (tile.direction.ordinal() == side) {
                            tile.direction = tile.direction.getOpposite();
                        } else {
                            tile.direction = ForgeDirection.getOrientation(side);
                        }
                        wrench.wrenchUsed(player, x, y, z);
                        world.markBlockForUpdate(x, y, z);
                        return true;
                    }
                }
            } else if ((item.itemID < Block.blocksList.length && BlockRail.isRailBlock(item.itemID))
                    || item instanceof ITrackItem) {
                return false;
            }
        }
        TileEntity tile = world.getBlockTileEntity(x, y, z);
        if (tile instanceof TileDetector) {
            return ((TileDetector) tile).blockActivated(player);
        }
        return false;
    }

    @Override
    public int getBlockTexture(IBlockAccess world, int i, int j, int k, int side) {
        int meta = world.getBlockMetadata(i, j, k);
        if (meta >= texture.length || meta < 0) {
            meta = 0;
        }
        TileEntity t = world.getBlockTileEntity(i, j, k);
        if (t instanceof TileDetector && ((TileDetector) t).direction.ordinal() == side) {
            if (((TileDetector) t).powered) {
                return texture[meta][2];
            }
            return texture[meta][1];
        }
        return texture[meta][0];
    }

    @Override
    public int getBlockTextureFromSideAndMetadata(int side, int meta) {
        if (meta >= texture.length || meta < 0) {
            return texture[0][1];
        }
        if (side == 3) {
            return texture[meta][2];
        }
        return texture[meta][0];
    }

    @Override
    public boolean canProvidePower() {
        return true;
    }

    @Override
    public boolean isProvidingWeakPower(IBlockAccess world, int i, int j, int k, int side) {
        TileEntity t = world.getBlockTileEntity(i, j, k);
        if (t instanceof TileDetector) {
            TileDetector tile = (TileDetector) t;
            if (tile.direction == MiscTools.getOppositeSide(side)) {
                return tile.powered;
            }
        }
        return false;
    }

    @Override
    public boolean isProvidingStrongPower(IBlockAccess world, int i, int j, int k, int side) {
        return isProvidingWeakPower(world, i, j, k, side);
    }

    @Override
    public void onBlockAdded(World world, int i, int j, int k) {
        super.onBlockAdded(world, i, j, k);
        world.markBlockForUpdate(i, j, k);
        if (Game.isNotHost(world)) {
            return;
        }
        world.notifyBlocksOfNeighborChange(i + 1, j, k, blockID);
        world.notifyBlocksOfNeighborChange(i - 1, j, k, blockID);
        world.notifyBlocksOfNeighborChange(i, j, k + 1, blockID);
        world.notifyBlocksOfNeighborChange(i, j, k - 1, blockID);
        world.notifyBlocksOfNeighborChange(i, j - 1, k, blockID);
        world.notifyBlocksOfNeighborChange(i, j + 1, k, blockID);
        world.scheduleBlockUpdate(i, j, k, blockID, tickRate());
    }

    @Override
    public void breakBlock(World world, int i, int j, int k, int id, int metadata) {
        super.breakBlock(world, i, j, k, id, metadata);
        if (Game.isNotHost(world)) {
            return;
        }
        TileEntity t = world.getBlockTileEntity(i, j, k);
        if (t instanceof TileDetector) {
            TileDetector tile = (TileDetector) t;
            tile.onBlockRemoval();
        }
        world.notifyBlocksOfNeighborChange(i + 1, j, k, blockID);
        world.notifyBlocksOfNeighborChange(i - 1, j, k, blockID);
        world.notifyBlocksOfNeighborChange(i, j, k + 1, blockID);
        world.notifyBlocksOfNeighborChange(i, j, k - 1, blockID);
        world.notifyBlocksOfNeighborChange(i, j - 1, k, blockID);
        world.notifyBlocksOfNeighborChange(i, j + 1, k, blockID);
    }

    @Override
    public boolean canConnectRedstone(IBlockAccess world, int i, int j, int k, int dir) {
        TileEntity t = world.getBlockTileEntity(i, j, k);
        if (t instanceof TileDetector) {
            TileDetector tile = (TileDetector) t;
            if (dir == 1 && tile.direction.ordinal() == 5) {
                return true;
            }
            if (dir == 3 && tile.direction.ordinal() == 4) {
                return true;
            }
            if (dir == 2 && tile.direction.ordinal() == 3) {
                return true;
            }
            if (dir == 0 && tile.direction.ordinal() == 2) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void getSubBlocks(int id, CreativeTabs tab, List list) {
        for (Detector d : Detector.values()) {
            if (d == Detector.ENERGY && !ModuleManager.isModuleLoaded(Module.IC2)) {
                continue;
            }
            if (d == Detector.TRAIN && !ModuleManager.isModuleLoaded(Module.TRAIN)) {
                continue;
            }
            list.add(new ItemStack(blockID, 1, d.ordinal()));
        }
    }
}
