package railcraft.common.blocks.detector;

import ic2.api.IWrenchable;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import net.minecraft.entity.EntityAgeable;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.monster.IMob;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.api.carts.CartTools;
import railcraft.common.api.carts.IExplosiveCart;
import railcraft.common.blocks.RailcraftBlocks;
import railcraft.common.blocks.RailcraftTileEntity;
import railcraft.common.blocks.detector.BlockDetector.Detector;
import railcraft.common.carts.EntityCartEnergy;
import railcraft.common.lang.RailcraftLanguage;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.misc.SafeNBTWrapper;

public class TileDetector extends RailcraftTileEntity implements IWrenchable {

    public static final float SENSITIVITY = 0.2f;
    public ForgeDirection direction = ForgeDirection.UP;
    public boolean powered = false;
    private int update = MiscTools.getRand().nextInt();

    public TileDetector() {
    }

    @Override
    public String getInvName() {
        return RailcraftLanguage.translate(BlockDetector.getBlockNameFromMetadata(getBlockMetadata()));
    }

    public List<EntityMinecart> getCarts() {
        return CartTools.getMinecartsOnAllSides(worldObj, xCoord, yCoord, zCoord, SENSITIVITY);
    }

    public boolean testForCarts() {
        int meta = getBlockMetadata();
        List<EntityMinecart> carts = getCarts();
        if (meta == Detector.ANY.ordinal()) {
            return !carts.isEmpty();
        } else if (meta == Detector.EMPTY.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart.canBeRidden() && cart.riddenByEntity == null) {
                    return true;
                }
            }
        } else if (meta == Detector.MOB.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart.riddenByEntity != null && cart.riddenByEntity instanceof IMob) {
                    return true;
                }
            }
        } else if (meta == Detector.POWERED.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart.isPoweredCart()) {
                    return true;
                }
            }
        } else if (meta == Detector.PLAYER.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart.riddenByEntity instanceof EntityPlayer) {
                    return true;
                }
            }
        } else if (meta == Detector.EXPLOSIVE.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart instanceof IExplosiveCart) {
                    return true;
                }
            }
        } else if (meta == Detector.ENERGY.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart instanceof EntityCartEnergy) {
                    return true;
                }
            }
        } else if (meta == Detector.AGE.ordinal()) {
            for (EntityMinecart cart : carts) {
                if (cart.riddenByEntity instanceof EntityAgeable) {
                    EntityAgeable ageable = (EntityAgeable) cart.riddenByEntity;
                    return !ageable.isChild();
                }
                if (cart.riddenByEntity != null) {
                    return true;
                }
            }
        }
        return false;
    }

    public boolean blockActivated(EntityPlayer player) {
        return false;
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        data.setByte("direction", (byte) direction.ordinal());
        data.setBoolean("powered", powered);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        SafeNBTWrapper safe = new SafeNBTWrapper(data);

        direction = ForgeDirection.getOrientation(safe.getByte("direction"));
        powered = data.getBoolean("powered");
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);

        data.writeBoolean(powered);
        data.writeByte(direction.ordinal());
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        powered = data.readBoolean();
        direction = ForgeDirection.getOrientation(data.readByte());

        markBlockForUpdate();
    }

    @Override
    public boolean canUpdate() {
        return true;
    }

    @Override
    public void updateEntity() {
        update++;
        if (Game.isNotHost(getWorld())) {
            return;
        }
        if (updateInterval() == 0 || update % updateInterval() == 0) {
            boolean cartDetected = testForCarts();
            if (cartDetected != powered) {
                powered = cartDetected;
                sendUpdateToClient();
                worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord, zCoord, RailcraftBlocks.getBlockDetector().blockID);
                MiscTools.notifyBlocksOfNeighborChangeOnSide(worldObj, xCoord, yCoord, zCoord, RailcraftBlocks.getBlockDetector().blockID, direction);
            }
        }
    }

    protected short updateInterval() {
        return 0;
    }

    @Override
    public boolean wrenchCanSetFacing(EntityPlayer player, int side) {
        if (direction == ForgeDirection.getOrientation(side)) {
            return false;
        }
        return true;
    }

    @Override
    public void setFacing(short facing) {
        direction = ForgeDirection.getOrientation(facing);
        sendUpdateToClient();
    }

    @Override
    public short getFacing() {
        return (short) direction.ordinal();
    }

    @Override
    public boolean wrenchCanRemove(EntityPlayer entityPlayer) {
        return true;
    }

    @Override
    public float getWrenchDropRate() {
        return 1;
    }

    @Override
    public short getId() {
        return (short) getBlockMetadata();
    }

    public void onBlockRemoval() {
    }

    @Override
    public ItemStack getWrenchDrop(EntityPlayer entityPlayer) {
        return getBlock().getBlockDropped(worldObj, xCoord, yCoord, zCoord, getBlockMetadata(), 0).get(0);
    }
}
