package railcraft.common.blocks.detector;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import railcraft.common.api.carts.CartTools;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.gui.slots.ISlotController;
import railcraft.common.lang.RailcraftLanguage;
import railcraft.common.util.inventory.InvTools;
import railcraft.common.util.misc.Game;
import railcraft.common.util.network.IGuiReturnHandler;

public class TileDetectorItem extends TileDetectorFilter implements IGuiReturnHandler
{

    public enum PrimaryMode
    {

        EMPTY("Empty"), FULL("Full"), ANYTHING("Anything"), FILTERED("Filtered"), NOT_EMPTY("Not Empty");
        public final String name;

        private PrimaryMode(String name)
        {
            this.name = name;
        }

        @Override
        public String toString()
        {
            return RailcraftLanguage.translate("gui.detector.item." + name().toLowerCase());
        }
    };

    public enum FilterMode
    {

        AT_LEAST("At Least"), AT_MOST("At Most"), EXACTLY("Exactly");
        public final String name;

        private FilterMode(String name)
        {
            this.name = name;
        }

        @Override
        public String toString()
        {
            return RailcraftLanguage.translate("gui.detector.item." + name().toLowerCase());
        }
    };
    private PrimaryMode primaryMode = PrimaryMode.ANYTHING;
    private FilterMode filterMode = FilterMode.AT_LEAST;

    public TileDetectorItem()
    {
        super(9);
    }

    @Override
    public boolean testForCarts()
    {
        List<EntityMinecart> carts = CartTools.getMinecartsOnAllSides(worldObj, xCoord, yCoord, zCoord, SENSITIVITY);
        for(EntityMinecart cart : carts) {
            if(cart.getSizeInventory() > 0) {
                switch (primaryMode) {
                    case ANYTHING:
                        return true;
                    case EMPTY:
                        if(InvTools.isInventoryEmpty(cart)) {
                            return true;
                        }
                        continue;
                    case FULL:
                        if(InvTools.isInventoryFull(cart)) {
                            return true;
                        }
                        continue;
                    case FILTERED:
                        if(matchesFilter(cart)) {
                            return true;
                        }
                        continue;
                    case NOT_EMPTY:
                        if(!InvTools.isInventoryEmpty(cart)) {
                            return true;
                        }
                        continue;
                }
            }
        }
        return false;
    }

    private boolean matchesFilter(EntityMinecart cart)
    {
        for(int i = 0; i < getFilters().getSizeInventory(); i++) {
            ItemStack filter = getFilters().getStackInSlot(i);
            if(filter == null) {
                continue;
            }
            int amountFilter = InvTools.countItems(getFilters(), filter);
            int amountCart = InvTools.countItems(cart, filter);

            switch (filterMode) {
                case EXACTLY:
                    if(amountCart != amountFilter) {
                        return false;
                    }
                    break;
                case AT_LEAST:
                    if(amountCart < amountFilter) {
                        return false;
                    }
                    break;
                case AT_MOST:
                    if(amountCart > amountFilter) {
                        return false;
                    }
                    break;
            }
        }
        return true;
    }

    @Override
    public void writeToNBT(NBTTagCompound data)
    {
        super.writeToNBT(data);

        data.setByte("primaryMode", (byte)primaryMode.ordinal());
        data.setByte("filterMode", (byte)filterMode.ordinal());
    }

    @Override
    public void readFromNBT(NBTTagCompound data)
    {
        super.readFromNBT(data);

        primaryMode = PrimaryMode.values()[data.getByte("primaryMode")];
        filterMode = FilterMode.values()[data.getByte("filterMode")];
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException
    {
        super.writePacketData(data);

        data.writeByte((byte)primaryMode.ordinal());
        data.writeByte((byte)filterMode.ordinal());
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException
    {
        super.readPacketData(data);

        primaryMode = PrimaryMode.values()[data.readByte()];
        filterMode = FilterMode.values()[data.readByte()];
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException
    {
        data.writeByte((byte)primaryMode.ordinal());
        data.writeByte((byte)filterMode.ordinal());
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException
    {
        primaryMode = PrimaryMode.values()[data.readByte()];
        filterMode = FilterMode.values()[data.readByte()];
    }

    @Override
    public boolean blockActivated(EntityPlayer player)
    {
        GuiHandler.openGui(EnumGui.DETECTOR_ITEM, player, worldObj, xCoord, yCoord, zCoord);
        return true;
    }

    public PrimaryMode getPrimaryMode()
    {
        return primaryMode;
    }

    public FilterMode getFilterMode()
    {
        return filterMode;
    }

    public void setPrimaryMode(PrimaryMode primaryMode)
    {
        this.primaryMode = primaryMode;
    }

    public void setFilterMode(FilterMode filterMode)
    {
        this.filterMode = filterMode;
    }

    public ISlotController getSlotController()
    {
        return new ISlotController()
        {

            @Override
            public boolean isSlotEnabled()
            {
                return getPrimaryMode() == PrimaryMode.FILTERED;
            }
        };
    }
}
