package railcraft.common.blocks.detector;

import railcraft.common.gui.util.MultiButtonController;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.liquids.ITankContainer;
import net.minecraftforge.liquids.LiquidStack;
import railcraft.common.api.carts.CartTools;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.gui.util.IMultiButtonState;
import railcraft.common.liquids.LiquidManager;
import railcraft.common.liquids.TankToolkit;
import railcraft.common.util.network.IGuiReturnHandler;

public class TileDetectorTank extends TileDetectorFilter implements IGuiReturnHandler {

    public enum ButtonState implements IMultiButtonState {

        VOID("L = ?"),
        EMPTY("L = 0%"),
        NOT_EMPTY("L > 0%"),
        FULL("L = 100%"),
        QUARTER("L >= 25%"),
        HALF("L >= 50%"),
        MOST("L >= 75%"),
        LESS_THAN_QUARTER("L < 25%"),
        LESS_THAN_HALF("L < 50%"),
        LESS_THAN_MOST("L < 75%"),
        LESS_THAN_FULL("L < 100%");
        private final String label;

        private ButtonState(String label) {
            this.label = label;
        }

        @Override
        public String getLabel() {
            return label;
        }
    }
    private final MultiButtonController<ButtonState> buttonController = MultiButtonController.getController(0, ButtonState.values());

    public TileDetectorTank() {
        super(1);
    }

    public MultiButtonController getButtonController() {
        return buttonController;
    }

    public LiquidStack getFilterLiquid() {
        ItemStack filter = getFilters().getStackInSlot(0);
        if (filter != null) {
            return LiquidManager.getInstance().getLiquidInContainer(filter);
        }
        return null;
    }

    @Override
    public boolean testForCarts() {
        List<EntityMinecart> carts = CartTools.getMinecartsOnAllSides(worldObj, xCoord, yCoord, zCoord, SENSITIVITY);
        for (EntityMinecart cart : carts) {
            if (cart instanceof ITankContainer) {
                TankToolkit tank = new TankToolkit((ITankContainer) cart);
                boolean liquidMatches = false;
                LiquidStack filterLiquid = getFilterLiquid();
                LiquidStack tankLiquid = tank.drain(ForgeDirection.UNKNOWN, 1, false);
                if (filterLiquid == null) {
                    liquidMatches = true;
                } else if (filterLiquid.isLiquidEqual(tankLiquid)) {
                    liquidMatches = true;
                } else if (tank.canPutLiquid(ForgeDirection.UNKNOWN, filterLiquid)) {
                    liquidMatches = true;
                }
                boolean quantityMatches = false;
                ButtonState state = buttonController.getButtonState();
                switch (state) {
                    case VOID:
                        quantityMatches = true;
                        break;
                    case EMPTY:
                        if (filterLiquid != null && tank.isTankEmpty(filterLiquid)) {
                            quantityMatches = true;
                        } else if (filterLiquid == null && tank.areTanksEmpty()) {
                            quantityMatches = true;
                        }
                        break;
                    case NOT_EMPTY:
                        if (filterLiquid != null && tank.getLiquidQty(filterLiquid) > 0) {
                            quantityMatches = true;
                        } else if (filterLiquid == null && tank.isLiquidInTank()) {
                            quantityMatches = true;
                        }
                        break;
                    case FULL:
                        if (filterLiquid != null && tank.isTankFull(filterLiquid)) {
                            quantityMatches = true;
                        } else if (filterLiquid == null && tank.areTanksFull()) {
                            quantityMatches = true;
                        }
                        break;
                    default:
                        float level = filterLiquid != null ? tank.getLiquidLevel(filterLiquid) : tank.getLiquidLevel();
                        switch (state) {
                            case QUARTER:
                                quantityMatches = level >= 0.25f;
                                break;
                            case HALF:
                                quantityMatches = level >= 0.5f;
                                break;
                            case MOST:
                                quantityMatches = level >= 0.75f;
                                break;
                            case LESS_THAN_QUARTER:
                                quantityMatches = level < 0.25f;
                                break;
                            case LESS_THAN_HALF:
                                quantityMatches = level < 0.5f;
                                break;
                            case LESS_THAN_MOST:
                                quantityMatches = level < 0.75f;
                                break;
                            case LESS_THAN_FULL:
                                quantityMatches = level < 1f;
                                break;
                        }
                }
                return liquidMatches && quantityMatches;
            }
        }
        return false;
    }

    @Override
    public boolean blockActivated(EntityPlayer player) {
        GuiHandler.openGui(EnumGui.DETECTOR_TANK, player, worldObj, xCoord, yCoord, zCoord);
        return true;
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        data.setByte("state", (byte) buttonController.getCurrentState());
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        buttonController.setCurrentState(data.getByte("state"));
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);

        data.writeByte(buttonController.getCurrentState());
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        buttonController.setCurrentState(data.readByte());
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException {
        data.writeByte(buttonController.getCurrentState());
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException {
        buttonController.setCurrentState(data.readByte());
    }
}
