package railcraft.common.blocks.machine;

import buildcraft.api.core.SafeTimeTracker;
import cpw.mods.fml.common.FMLCommonHandler;
import cpw.mods.fml.common.network.PacketDispatcher;
import cpw.mods.fml.relauncher.Side;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.List;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EnumCreatureType;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.AxisAlignedBB;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.network.PacketTileRequest;

public abstract class TileMultiBlock extends TileMachineBase {

    private enum MultiBlockState {

        VALID, INVALID, UNKNOWN
    };
    private static final int UNKNOWN_STATE_RECHECK = 256;
    private static final int NETWORK_RECHECK = 64;
    private static final int NEIGHBOR_CHECK = 32;
    protected boolean isMaster;
    private byte patternX;
    private byte patternY;
    private byte patternZ;
    private boolean checkNeighbors = true;
    private boolean tested;
    private boolean requestPacket;
    protected int update = MiscTools.getRand().nextInt();
    private SafeTimeTracker timeTracker = new SafeTimeTracker();
    private MultiBlockState state;
    private TileMultiBlock masterBlock;
    private MultiBlockPattern currentPattern;
    private final List<? extends MultiBlockPattern> patterns;
    private final TileMultiBlock[] neighbors = new TileMultiBlock[6];

    public TileMultiBlock(List<? extends MultiBlockPattern> patterns) {
        this.patterns = patterns;
        currentPattern = patterns.get(0);
        tested = FMLCommonHandler.instance().getEffectiveSide() != Side.SERVER;
    }

    protected void onPatternLock(MultiBlockPattern pattern) {
    }

    public final char getPatternMarker() {
        if (currentPattern == null || !isStructureValid()) {
            return 'O';
        }
        return currentPattern.getPatternMarker(patternX, patternY, patternZ);
    }

    public final int getPatternPositionX() {
        return patternX;
    }

    public final int getPatternPositionY() {
        return patternY;
    }

    public final int getPatternPositionZ() {
        return patternZ;
    }

    private void setPatternPosition(byte x, byte y, byte z) {
        patternX = x;
        patternY = y;
        patternZ = z;
    }

    public final void setPattern(MultiBlockPattern pattern) {
        this.currentPattern = pattern;
        onPatternLock(pattern);
    }

    public final MultiBlockPattern getPattern() {
        return currentPattern;
    }

    public final byte getPatternIndex() {
        return (byte) patterns.indexOf(currentPattern);
    }

    protected int getMaxRecursionDepth() {
        return 12;
    }

    @Override
    public final boolean canUpdate() {
        return true;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();
        update++;
        if (Game.isHost(worldObj)) {
            if (checkNeighbors && update % NEIGHBOR_CHECK == 0) {
                grabNeighbors();
            }
            if (!tested && (state != MultiBlockState.UNKNOWN || update % UNKNOWN_STATE_RECHECK == 0)) {
                testIfMasterBlock();
//                ClientProxy.getMod().totalMultiBlockUpdates++;
            }
        } else {
            if (requestPacket && timeTracker.markTimeIfDelay(worldObj, NETWORK_RECHECK)) {
                PacketTileRequest pkt = new PacketTileRequest(this);
                PacketDispatcher.sendPacketToServer(pkt.getPacket());
                requestPacket = false;
            }
        }
    }

    private void testIfMasterBlock() {
//        System.out.println("testing structure");
        state = getMasterBlockState();
        tested = true;

        if (state == MultiBlockState.UNKNOWN) {
            tested = false;
        } else if (state == MultiBlockState.VALID) {
            isMaster = true;
//             System.out.println("structure complete");

            int xWidth = currentPattern.getPatternWidthX();
            int zWidth = currentPattern.getPatternWidthZ();
            int height = currentPattern.getPatternHeight();

            int xOffset = xCoord - currentPattern.getMasterOffsetX();
            int yOffset = yCoord - currentPattern.getMasterOffsetY();
            int zOffset = zCoord - currentPattern.getMasterOffsetZ();

            for (byte px = 0; px < xWidth; px++) {
                for (byte py = 0; py < height; py++) {
                    for (byte pz = 0; pz < zWidth; pz++) {

                        char marker = currentPattern.getPatternMarker(px, py, pz);
                        if (isMapPositionOtherBlock(marker)) {
                            continue;
                        }

                        int x = px + xOffset;
                        int y = py + yOffset;
                        int z = pz + zOffset;

                        TileEntity tile = worldObj.getBlockTileEntity(x, y, z);
                        if (tile instanceof TileMultiBlock) {
                            TileMultiBlock multiBlock = (TileMultiBlock) tile;
                            multiBlock.masterBlock = this;
                            multiBlock.tested = true;
                            multiBlock.checkNeighbors = true;
                            multiBlock.setPattern(currentPattern);
                            multiBlock.setPatternPosition(px, py, pz);
                            multiBlock.sendUpdateToClient();
                        } else {
                            return;
                        }
                    }
                }
            }
        } else {
            if (isMaster) {
                isMaster = false;
                onMasterReset();
                sendUpdateToClient();
            }
        }
    }

    protected void onMasterReset() {
    }

    protected boolean isMapPositionOtherBlock(char mapPos) {
        switch (mapPos) {
            case 'A':
            case 'O':
                return true;
            default:
                return false;
        }
    }

    protected boolean isMapPositionValid(int i, int j, int k, char mapPos) {
        int id = worldObj.getBlockId(i, j, k);
        switch (mapPos) {
            case 'O': // Other
                if (id == getBlockId() && worldObj.getBlockMetadata(i, j, k) == getBlockMetadata()) {
                    return false;
                }
                break;
            case 'W': // Window
            case 'B': // Block
                if (id != getBlockId() || worldObj.getBlockMetadata(i, j, k) != getBlockMetadata()) {
                    return false;
                }
                break;
            case 'A': // Air
                if (!worldObj.isAirBlock(i, j, k)) {
                    return false;
                }
                break;
        }
        return true;
    }

    private MultiBlockState getMasterBlockState() {
        for (MultiBlockPattern map : patterns) {
            boolean valid = true;

            int xWidth = map.getPatternWidthX();
            int zWidth = map.getPatternWidthZ();
            int height = map.getPatternHeight();

            int xOffset = xCoord - map.getMasterOffsetX();
            int yOffset = yCoord - map.getMasterOffsetY();
            int zOffset = zCoord - map.getMasterOffsetZ();

            for (int x = 0; x < xWidth && valid; x++) {
                for (int y = 0; y < height && valid; y++) {
                    for (int z = 0; z < zWidth; z++) {
                        int xx = x + xOffset;
                        int yy = y + yOffset;
                        int zz = z + zOffset;
                        if (!worldObj.blockExists(xx, yy, zz)) {
                            return MultiBlockState.UNKNOWN;
                        }
                        if (!isMapPositionValid(xx, yy, zz, map.getPatternMarker(x, y, z))) {
                            valid = false;
                            break;
                        }
                    }
                }
            }

            if (valid) {
                AxisAlignedBB entityCheckBounds = map.getEntityCheckBounds(xCoord, yCoord, zCoord);
//                if(entityCheckBounds != null) {
//                    System.out.println("test entitys: " + entityCheckBounds.toString());
//                }
                if (entityCheckBounds != null && !worldObj.getEntitiesWithinAABB(Entity.class, entityCheckBounds).isEmpty()) {
                    return MultiBlockState.INVALID;
                }
                currentPattern = map;
                return MultiBlockState.VALID;
            }
        }

        return MultiBlockState.INVALID;
    }

    private void grabNeighbors() {
        checkNeighbors = false;
        for (int side = 0; side < 6; side++) {
            neighbors[side] = null;
            ForgeDirection dir = ForgeDirection.getOrientation(side);
            if (!MiscTools.blockExistsOnSide(worldObj, xCoord, yCoord, zCoord, dir)) {
                checkNeighbors = true;
                continue;
            }
            TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, xCoord, yCoord, zCoord, dir);
            if (tile != null && isStructureTile(tile)) {
                neighbors[side] = (TileMultiBlock) tile;
            }
        }
    }

    @Override
    public void onBlockAdded() {
        super.onBlockAdded();

        if (Game.isNotHost(worldObj)) return;

        grabNeighbors();
        onBlockChange();
    }

    @Override
    public void onBlockRemoval() {
        super.onBlockRemoval();

        if (Game.isNotHost(worldObj)) return;

        onBlockChange();
        isMaster = false;
    }

    @Override
    public void onChunkUnload() {
        super.onChunkUnload();

        if (Game.isNotHost(worldObj)) return;

        tested = false;
        retestMasterBlock();
    }

    @Override
    public void invalidate() {
        if (worldObj != null && Game.isNotHost(worldObj)) return;
        tested = false;
        retestMasterBlock();
        super.invalidate();
    }

    private void onBlockChange() {
        for (int side = 0; side < 6; side++) {
            TileMultiBlock tile = neighbors[side];
            if (tile != null) {
                tile.grabNeighbors();
                tile.onBlockChange(getMaxRecursionDepth());
            }
        }
    }

    private void onBlockChange(int depth) {
        depth--;
        if (depth < 0) {
            return;
        }
        if (tested) {
            tested = false;
            for (int side = 0; side < 6; side++) {
                TileMultiBlock tile = neighbors[side];
                if (tile != null && !tile.isInvalid()) {
                    TileMultiBlock mBlock = tile.getMasterBlock();
                    if (mBlock == null) {
                        tile.onBlockChange(depth);
                    } else {
                        mBlock.grabNeighbors();
                        mBlock.onBlockChange(getMaxRecursionDepth());
                    }
                } else {
                    neighbors[side] = null;
                }
            }
        }
    }

    protected boolean isStructureTile(TileEntity tile) {
        return tile.getClass() == getClass();
    }

    @Override
    public void onInventoryChanged() {
        super.onInventoryChanged();
        if (!isMaster) {
            TileMultiBlock mBlock = (TileMultiBlock) getMasterBlock();
            if (mBlock != null) {
                mBlock.onInventoryChanged();
            }
        }
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        data.setBoolean("master", isMaster);
//        data.setByte("pattern", getPatternIndex());
//        data.setByte("patternX", patternX);
//        data.setByte("patternY", patternY);
//        data.setByte("patternZ", patternZ);

    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        isMaster = data.getBoolean("master");
//        byte patternIndex = data.getByte("pattern");
//        if (patternIndex >= 0 && patternIndex < patterns.size()) {
//            setPattern(patterns.get(patternIndex));
//        }
//        patternX = data.getByte("patternX");
//        patternY = data.getByte("patternY");
//        patternZ = data.getByte("patternZ");
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);
        boolean hasMaster = getMasterBlock() != null;
        data.writeBoolean(hasMaster);
        if (hasMaster) {
            byte patternIndex = getPatternIndex();
            data.writeByte(patternIndex);

            data.writeByte(patternX);
            data.writeByte(patternY);
            data.writeByte(patternZ);
        }
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        requestPacket = false;
        boolean needsUpdate = false;

        boolean hasMaster = data.readBoolean();
        if (hasMaster) {
            byte patternIndex = data.readByte();
            patternIndex = (byte) Math.max(patternIndex, 0);
            patternIndex = (byte) Math.min(patternIndex, patterns.size() - 1);
            MultiBlockPattern pat = patterns.get(patternIndex);

            byte pX = data.readByte();
            byte pY = data.readByte();
            byte pZ = data.readByte();

            if (patternX != pX || patternY != pY || patternZ != pZ) {
                patternX = pX;
                patternY = pY;
                patternZ = pZ;
                needsUpdate = true;
            }

            isMaster = pX == pat.getMasterOffsetX() && pY == pat.getMasterOffsetY() && pZ == pat.getMasterOffsetZ();

            setPattern(pat);

            int masterX = pat.getMasterRelativeX(xCoord, pX);
            int masterY = pat.getMasterRelativeY(yCoord, pY);
            int masterZ = pat.getMasterRelativeZ(zCoord, pZ);

            TileEntity tile = null;
            if (worldObj != null) {
                tile = worldObj.getBlockTileEntity(masterX, masterY, masterZ);
            }
            if (tile != null) {
                if (masterBlock != tile && isStructureTile(tile)) {
                    needsUpdate = true;
                    masterBlock = (TileMultiBlock) tile;
                }
            }
            if (getMasterBlock() == null) {
                requestPacket = true;
            }
        } else if (masterBlock != null) {
            needsUpdate = true;
            masterBlock = null;
            isMaster = false;
        }

        if (needsUpdate) {
            markBlockForUpdate();
        }

//
//        System.out.printf("marker=%c, pattern=%d, x=%d, y=%d, z=%d%n", currentPattern.getPatternMarkerChecked(patternX, patternY, patternZ), patternIndex, patternX, patternY, patternZ);
//        if(masterBlock != null)
//        System.out.printf("tested=%b, invalid=%b, isMaster=%b%n" ,masterBlock.tested, masterBlock.isInvalid(), masterBlock.isMaster());
    }

    public final boolean isMaster() {
        return isMaster;
    }

    public final void setMaster(boolean m) {
        isMaster = m;
    }

    public final void retestMasterBlock() {
        if (masterBlock != null) {
            masterBlock.tested = false;
        }
    }

    public final boolean isStructureValid() {
        return masterBlock != null && masterBlock.tested && masterBlock.isMaster && !masterBlock.isInvalid();
    }

    public final TileMultiBlock getMasterBlock() {
        if (masterBlock != null && !isStructureValid()) {
            masterBlock = null;
            sendUpdateToClient();
        }
        return masterBlock;
    }

    @Override
    public boolean canCreatureSpawn(EnumCreatureType type) {
        return isStructureValid() && getPatternPositionY() < 2 ? false : super.canCreatureSpawn(type);
    }
}
