package railcraft.common.blocks.machine.beta;

import buildcraft.api.power.IPowerProvider;
import buildcraft.api.power.IPowerReceptor;
import buildcraft.api.power.PowerFramework;
import buildcraft.api.tools.IToolWrench;
import buildcraft.api.transport.IPipeConnection;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.blocks.machine.TileMachineBase;
import railcraft.common.gui.indicator.IIndicatorController;
import railcraft.common.gui.util.ToolTip;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public abstract class TileEngine extends TileMachineBase implements IPowerReceptor, IPipeConnection {

    private ForgeDirection direction = ForgeDirection.UP;
    private float pistonProgress = 0;
    public float currentOutput = 0;
    private int pistonStage;
    private boolean powered;
    private boolean isActive;
    private boolean needsInit = true;
    public float energy;
    public float extraEnergy;
    private EnergyStage energyStage = EnergyStage.BLUE;
    private IPowerProvider provider = PowerFramework.currentFramework.createPowerProvider();
    private IIndicatorController energyIndicator = new EnergyIndicator();

    private class EnergyIndicator implements IIndicatorController {

        List<ToolTip> tips = new ArrayList<ToolTip>();
        ToolTip tip = new ToolTip();

        private EnergyIndicator() {
            tips.add(tip);
        }

        @Override
        public List<ToolTip> getToolTip() {
            tip.text = String.format("%.0f MJ", energy);
            return tips;
        }

        @Override
        public int getScaledLevel(int size) {
            float e = Math.min(energy, maxEnergy());
            return (int) (e * size / maxEnergy());
        }
    };

    public IIndicatorController getEnergyIndicator() {
        return energyIndicator;
    }

    public float getCurrentOutput() {
        return currentOutput;
    }

    public float getEnergy() {
        return energy;
    }

    public enum EnergyStage {

        BLUE, GREEN, YELLOW, ORANGE, RED, OVERHEAT
    }

    public TileEngine() {
        initEnergyProvider();
    }

    private void initEnergyProvider() {
        provider.configure(0, 2, maxEnergyReceived(), 1, maxEnergy());
    }

    @Override
    public void updateEntity() {
        super.updateEntity();
        if (Game.isNotHost(worldObj)) {
            if (pistonStage != 0) {
                pistonProgress += getPistonSpeed();

                if (pistonProgress > 1) {
                    pistonStage = 0;
                    pistonProgress = 0;
                }
            } else if (this.isActive) {
                pistonStage = 1;
            }

            return;
        }

        if (needsInit) {
            needsInit = false;
            checkPower();
        }

        if (!powered) {
            if (energy > 1) {
                energy--;
            }
        }

        provider.update(this);

        if (getEnergyStage() == EnergyStage.OVERHEAT) {
            overheat();
        } else {
            if (pistonStage != 0) {
                pistonProgress += getPistonSpeed();

                if (pistonProgress > 0.5 && pistonStage == 1) {
                    pistonStage = 2;

                    TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, xCoord, yCoord, zCoord, direction);

                    if (EngineTools.isPoweredTile(tile)) {
                        IPowerReceptor receptor = (IPowerReceptor) tile;

                        IPowerProvider recProv = receptor.getPowerProvider();

                        float extracted = extractEnergy(recProv.getMinEnergyReceived(), recProv.getMaxEnergyReceived(), true);

                        if (extracted > 0) {
                            recProv.receiveEnergy(extracted, direction.getOpposite());
                        }
                    }
                } else if (pistonProgress >= 1) {
                    pistonProgress = 0;
                    pistonStage = 0;
                }
            } else if (powered) {
                TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, xCoord, yCoord, zCoord, direction);

                if (EngineTools.isPoweredTile(tile)) {
                    IPowerReceptor receptor = (IPowerReceptor) tile;

                    IPowerProvider recProv = receptor.getPowerProvider();

                    if (extractEnergy(recProv.getMinEnergyReceived(), recProv.getMaxEnergyReceived(), false) > 0) {
                        pistonStage = 1;
                        setActive(true);
                    } else {
                        setActive(false);
                    }
                } else {
                    setActive(false);
                }

            } else {
                setActive(false);
            }
        }

        burn();
    }

    protected void overheat() {
        subtractEnergy(5);
    }

    protected abstract void burn();

    private void setActive(boolean isActive) {
        if (this.isActive != isActive) {
            this.isActive = isActive;
            sendUpdateToClient();
        }
    }

    public boolean isActive() {
        return isActive;
    }

    public float getPistonSpeed() {
        if (Game.isHost(worldObj)) {
            return Math.max(0.16f * getEnergyLevel(), 0.01f);
        }
        switch (getEnergyStage()) {
            case BLUE:
                return 0.01F;
            case GREEN:
                return 0.02F;
            case YELLOW:
                return 0.04F;
            case ORANGE:
                return 0.08F;
            case RED:
                return 0.16F;
            default:
                return 0;
        }
    }

    @Override
    public boolean blockActivated(EntityPlayer player, int side) {
        ItemStack current = player.inventory.getCurrentItem();
        if (current != null) {
            if (current.getItem() instanceof IToolWrench) {
                IToolWrench wrench = (IToolWrench) current.getItem();
                if (wrench.canWrench(player, xCoord, yCoord, zCoord)) {
                    if (Game.isHost(worldObj) && getEnergyStage() == EnergyStage.OVERHEAT) {
                        resetEnergyStage();
                        wrench.wrenchUsed(player, xCoord, yCoord, zCoord);
                        return true;
                    }
                    if (switchOrientation()) {
                        wrench.wrenchUsed(player, xCoord, yCoord, zCoord);
                        return true;
                    }
                }
            }
        }
        return super.blockActivated(player, side);
    }

    @Override
    public void onBlockPlacedBy(EntityLiving entityliving) {
        super.onBlockPlacedBy(entityliving);
        switchOrientation();
        checkPower();
    }

    @Override
    public void onNeighborBlockChange(int blockId) {
        if (Game.isNotHost(getWorld())) {
            return;
        }
        checkPower();
    }

    private void checkPower() {
        boolean p = worldObj.isBlockGettingPowered(xCoord, yCoord, zCoord) || worldObj.isBlockIndirectlyGettingPowered(xCoord, yCoord, zCoord);
        if (powered != p) {
            powered = p;
            sendUpdateToClient();
        }
    }

    public boolean isPowered() {
        return powered;
    }

    @Override
    public boolean canConnectRedstone(int dir) {
        return true;
    }

    public boolean switchOrientation() {
        for (int i = direction.ordinal() + 1; i <= direction.ordinal() + 6; ++i) {
            ForgeDirection dir = ForgeDirection.getOrientation(i % 6);

            TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, xCoord, yCoord, zCoord, dir);

            if (EngineTools.isPoweredTile(tile)) {
                direction = dir;
                sendUpdateToClient();
                if (Game.isNotHost(worldObj)) {
                    markBlockForUpdate();
                }
                return true;
            }
        }
        return false;
    }

    public ForgeDirection getOrientation() {
        return direction;
    }

    @Override
    public boolean isBlockSolidOnSide(ForgeDirection side) {
        return direction.getOpposite() == side;
    }

    @Override
    public void setPowerProvider(IPowerProvider provider) {
        this.provider = provider;
    }

    @Override
    public IPowerProvider getPowerProvider() {
        return provider;
    }

    @Override
    public int powerRequest() {
        return 0;
    }

    @Override
    public void doWork() {
        if (Game.isNotHost(worldObj)) {
            return;
        }
        float e = provider.useEnergy(1, maxEnergyReceived(), true) * 0.95F;
        extraEnergy += e;
        addEnergy(e);
    }

    public float getEnergyLevel() {
        return energy / maxEnergy();
    }

    protected EnergyStage computeEnergyStage() {
        float energyLevel = getEnergyLevel();
        if (energyLevel < 0.2f) {
            return EnergyStage.BLUE;
        } else if (energyLevel < 0.4f) {
            return EnergyStage.GREEN;
        } else if (energyLevel < 0.6f) {
            return EnergyStage.YELLOW;
        } else if (energyLevel < 0.8f) {
            return EnergyStage.ORANGE;
        } else if (energyLevel < 1f) {
            return EnergyStage.RED;
        } else {
            return EnergyStage.OVERHEAT;
        }
    }

    public final EnergyStage getEnergyStage() {
        if (Game.isHost(worldObj)) {
            if (energyStage == EnergyStage.OVERHEAT) {
                return energyStage;
            }
            EnergyStage newStage = computeEnergyStage();

            if (energyStage != newStage) {
                energyStage = newStage;
                sendUpdateToClient();
            }
        }

        return energyStage;
    }

    public final void resetEnergyStage() {
        EnergyStage newStage = computeEnergyStage();

        if (energyStage != newStage) {
            energyStage = newStage;
            sendUpdateToClient();
        }
    }

    public void addEnergy(float addition) {
        energy += addition;

        if (energy > maxEnergy()) {
            energy = maxEnergy();
        }
    }

    public void subtractEnergy(float subtraction) {
        energy -= subtraction;
        if (energy < 0) {
            energy = 0;
        }
    }

    public float extractEnergy(int min, int max, boolean doExtract) {
        if (energy < min) {
            return 0;
        }

        float actualMax;

        float combinedMax = maxEnergyExtracted() + extraEnergy * 0.5f;
        if (max > combinedMax) {
            actualMax = combinedMax;
        } else {
            actualMax = max;
        }


        float extracted;

        if (energy >= actualMax) {
            extracted = actualMax;
            if (doExtract) {
                energy -= actualMax;
                extraEnergy -= Math.min(actualMax, extraEnergy);
            }
        } else {
            extracted = energy;
            if (doExtract) {
                energy = 0;
                extraEnergy = 0;
            }
        }

        return extracted;
    }

    public float getProgress() {
        return pistonProgress;
    }

    public abstract int maxEnergy();

    public abstract int maxEnergyExtracted();

    public abstract int maxEnergyReceived();

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        data.setByte("direction", (byte) direction.ordinal());
        data.setBoolean("powered", powered);
        data.setFloat("energy", energy);
        data.setFloat("currentOutput", currentOutput);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        direction = ForgeDirection.getOrientation(data.getByte("direction"));
        powered = data.getBoolean("powered");
        energy = data.getFloat("energy");
        currentOutput = data.getFloat("currentOutput");
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);

        data.writeByte(direction.ordinal());
        data.writeByte(getEnergyStage().ordinal());
        data.writeBoolean(isActive);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        direction = ForgeDirection.getOrientation(data.readByte());
        energyStage = EnergyStage.values()[data.readByte()];
        isActive = data.readBoolean();
    }

    @Override
    public boolean isPipeConnected(ForgeDirection side) {
        return side != direction;
    }
}
