package railcraft.common.blocks.machine.beta;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.liquids.ILiquidTank;
import net.minecraftforge.liquids.ITankContainer;
import net.minecraftforge.liquids.LiquidStack;
import railcraft.common.blocks.machine.IEnumMachine;
import railcraft.common.liquids.LiquidManager;
import railcraft.common.liquids.TankManager;
import railcraft.common.liquids.TankWrapper;
import railcraft.common.liquids.tanks.FakeTank;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public class TileTankIronValve extends TileTankIron implements ITankContainer {

    private static final byte FILL_DURATION = 20;
    private byte filling = 0;

    private void setFilling() {
        boolean needsUpdate = false;
        if (filling == 0) {
            needsUpdate = true;
        }
        filling = FILL_DURATION;
        if(needsUpdate){
            sendUpdateToClient();
        }
    }

    private void decrementFilling() {
        byte f = filling;
        if (filling > 0) {
            filling--;
        }
        if (filling == 0 && f != 0) {
            sendUpdateToClient();
        }
    }

    public boolean isFilling() {
        return filling > 0;
    }

    @Override
    public IEnumMachine getMachineType() {
        return EnumMachineBeta.TANK_IRON_VALVE;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();

        if (Game.isNotHost(worldObj)) {
            return;
        }
        decrementFilling();

        if (isMaster) {
            TileEntity tileBelow = worldObj.getBlockTileEntity(xCoord, yCoord - 1, zCoord);

            TileTankIronValve tankBelow = null;
            if (tileBelow instanceof TileTankIronValve) {
                tankBelow = (TileTankIronValve) tileBelow;
                if (tankBelow.isStructureValid() && tankBelow.getPatternMarker() == 'T') {
                    ILiquidTank tank = tankBelow.getTankManager().getTank(0);
                    LiquidStack liquid = tank.getLiquid();
                    if (liquid != null && liquid.amount >= tank.getCapacity() - LiquidManager.BUCKET_VOLUME) {
                        tankBelow = null;

                        LiquidStack fillStack = liquid.copy();
                        fillStack.amount = LiquidManager.BUCKET_VOLUME - (tank.getCapacity() - liquid.amount);
                        if (fillStack.amount > 0) {
                            int used = fill(0, fillStack, false);
                            if (used > 0) {
                                fillStack = tank.drain(used, true);
                                fill(0, fillStack, true);
                            }
                        }
                    }
                } else {
                    tankBelow = null;
                }
            }

            if (tankBelow != null) {
                LiquidStack available = tankManager.drain(0, LiquidManager.BUCKET_VOLUME, false);
                if (available != null && available.amount > 0) {
                    int used = tankBelow.fill(ForgeDirection.UP, available, true);
                    tankManager.drain(0, used, true);
                }
            }
        }
    }

    @Override
    public int getBlockTexture(int side) {
        if (!isStructureValid() || getPattern() == null) {
            return EnumMachineBeta.TANK_IRON_VALVE.getTexture(side);
        }
        ForgeDirection s = ForgeDirection.getOrientation(side);
        char markerSide = getPattern().getPatternMarkerChecked(MiscTools.getXOnSide(getPatternPositionX(), s), MiscTools.getYOnSide(getPatternPositionY(), s), MiscTools.getZOnSide(getPatternPositionZ(), s));

        if (!isMapPositionOtherBlock(markerSide)) {
            if(side == ForgeDirection.UP.ordinal() || side == ForgeDirection.DOWN.ordinal()){
                return EnumMachineBeta.TANK_IRON_VALVE.getTexture(6);
            }
            return EnumMachineBeta.TANK_IRON_VALVE.getTexture(7);
        }
        return EnumMachineBeta.TANK_IRON_VALVE.getTexture(side);
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);
        data.writeByte(filling);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);
        filling = data.readByte();
    }

    /**
     * Drains liquid out of internal tanks, distribution is left to the
     * ITankContainer.
     *
     * @param from ForgeDirection the liquid is drained to.
     * @param maxDrain Maximum amount of liquid to drain.
     * @param doDrain If false draining will only be simulated.
     * @return LiquidStack representing the liquid and amount actually drained
     * from the ITankContainer
     */
    @Override
    public LiquidStack drain(ForgeDirection from, int maxDrain, boolean doDrain) {
        return drain(0, maxDrain, doDrain);
    }

    /**
     * Drains liquid out of the specified internal tank.
     *
     * @param maxDrain Maximum amount of liquid to drain.
     * @param doDrain If false draining will only be simulated.
     * @return LiquidStack representing the liquid and amount actually drained
     * from the ITankContainer
     */
    @Override
    public LiquidStack drain(int tankIndex, int maxDrain, boolean doDrain) {
        if (getPatternPositionY() - getPattern().getMasterOffsetY() > 1) {
            return null;
        }
        TankManager tMan = getTankManager();
        if (tMan != null) {
            return tMan.drain(tankIndex, maxDrain, doDrain);
        }
        return null;
    }

    /**
     * Fills liquid into internal tanks, distribution is left to the
     * ITankContainer.
     *
     * @param from ForgeDirection the liquid is pumped in from.
     * @param resource LiquidStack representing the maximum amount of liquid
     * filled into the ITankContainer
     * @param doFill If false filling will only be simulated.
     * @return Amount of resource that was filled into internal tanks.
     */
    @Override
    public int fill(ForgeDirection from, LiquidStack resource, boolean doFill) {
        return fill(0, resource, doFill);
    }

    /**
     * Fills liquid into the specified internal tank.
     *
     * @param from ForgeDirection the liquid is pumped in from.
     * @param resource LiquidStack representing the maximum amount of liquid
     * filled into the ITankContainer
     * @param doFill If false filling will only be simulated.
     * @return Amount of resource that was filled into internal tanks.
     */
    @Override
    public int fill(int tankIndex, LiquidStack resource, boolean doFill) {
        if (resource == null) return 0;
        TankManager tMan = getTankManager();
        if (tMan == null) {
            return 0;
        }
        int filled = tMan.fill(tankIndex, resource, doFill);
        if (filled > 0) {
            setFilling();
        }
        return filled;
    }

    @Override
    public ILiquidTank[] getTanks(ForgeDirection side) {
        TankManager tMan = getTankManager();
        if (tMan != null) {
            return tMan.getTanks();
        }
        return FakeTank.ARRAY;
    }

    @Override
    public ILiquidTank getTank(ForgeDirection direction, LiquidStack type) {
        TankManager tMan = getTankManager();
        if (tMan != null) {
            return new IronTankWrapper(tMan.getTank(0));
        }
        return FakeTank.INSTANCE;
    }

    private class IronTankWrapper extends TankWrapper {

        public IronTankWrapper(ILiquidTank tank) {
            super(tank);
        }

        @Override
        public int fill(LiquidStack resource, boolean doFill) {
            if (resource == null) return 0;
            int filled = tank.fill(resource, doFill);
            if (filled > 0) {
                setFilling();
            }
            return filled;
        }

        @Override
        public LiquidStack drain(int maxDrain, boolean doDrain) {
            if (getPatternPositionY() - getPattern().getMasterOffsetY() > 1) {
                return null;
            }
            return tank.drain(maxDrain, doDrain);
        }

        @Override
        public int getTankPressure() {
            if (getPatternPositionY() - getPattern().getMasterOffsetY() > 1) {
                return 0;
            }
            return 1;
        }
    }
}
