package railcraft.common.blocks.machine.gamma;

import buildcraft.api.inventory.ISpecialInventory;
import java.util.Map;
import java.util.Set;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemMinecart;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.api.carts.CartTools;
import railcraft.common.api.core.items.IItemType;
import railcraft.common.api.core.items.IMinecartItem;
import railcraft.common.blocks.machine.IEnumMachine;
import railcraft.common.carts.ItemCartAnchor;
import railcraft.common.core.RailcraftConfig;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.util.inventory.InvTools;
import railcraft.common.util.inventory.InventoryCopy;
import railcraft.common.util.inventory.InventoryMapper;
import railcraft.common.util.inventory.ItemStackSet;
import railcraft.common.util.inventory.PhantomInventory;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

public class TileDispenserTrain extends TileDispenserCart implements ISpecialInventory {

    public final static int PATTERN_SIZE = 9;
    public final static int BUFFER_SIZE = 18;
    private byte patternIndex;
    private boolean spawningTrain = false;
    private EntityMinecart lastCart;
    private PhantomInventory invPattern = new PhantomInventory(PATTERN_SIZE, this);
    private int update = MiscTools.getRand().nextInt();

    public TileDispenserTrain() {
        super();
        setInventorySize(BUFFER_SIZE);
    }

    @Override
    public IEnumMachine getMachineType() {
        return EnumMachineGamma.DISPENSER_TRAIN;
    }

    public PhantomInventory getPattern() {
        return invPattern;
    }

    @Override
    public int getBlockTexture(int side) {
        if (getFacing() == side) {
            return getMachineType().getTexture(3);
        }
        if (side != 0 && side != 1) {
            return getMachineType().getTexture(2);
        }
        return getMachineType().getTexture(1);
    }

    @Override
    public boolean openGui(EntityPlayer player) {
        GuiHandler.openGui(EnumGui.TRAIN_DISPENSER, player, worldObj, xCoord, yCoord, zCoord);
        return true;
    }

    private boolean canBuildTrain() {
        Map<ItemStack, Integer> pattern = InvTools.getManifest(getPattern());
        Map<ItemStack, Integer> buffer = InvTools.getManifest(this);

        for (Map.Entry<ItemStack, Integer> entry : pattern.entrySet()) {
            Integer count = buffer.get(entry.getKey());
            if (count == null || count < entry.getValue()) {
                return false;
            }
        }

        return true;
    }

    private static class MinecartItemType implements IItemType {

        private final ItemStack original;
        private final boolean isAnchor;

        public MinecartItemType(ItemStack cart) {
            original = cart;
            isAnchor = original.getItem() instanceof ItemCartAnchor;
        }

        @Override
        public boolean isItemType(ItemStack stack) {
            if (stack == null) {
                return false;
            }
            if (InvTools.isItemEqual(stack, original)) {
                return true;
            }
            if (isAnchor && stack.getItem() instanceof ItemCartAnchor) {
                return true;
            }
            return false;
        }
    }

    private boolean spawnNextCart() {
        ItemStack spawn = getPattern().getStackInSlot(patternIndex);
        if (spawn == null) {
            resetSpawnSequence();
            return false;
        }
        IItemType filter = new MinecartItemType(spawn);
        if (InvTools.countItems(this, filter) == 0) {
            resetSpawnSequence();
            return false;
        }
        int x = MiscTools.getXOnSide(xCoord, direction);
        int y = MiscTools.getYOnSide(yCoord, direction);
        int z = MiscTools.getZOnSide(zCoord, direction);
        if ((spawn.getItem() instanceof ItemMinecart || spawn.getItem() instanceof IMinecartItem)
                && CartTools.getMinecartOnSide(worldObj, xCoord, yCoord, zCoord, 0, direction) == null) {
            ItemStack cartItem = InvTools.removeOneItem(this, filter);
            if (cartItem != null) {
                EntityMinecart cartPlaced = CartTools.placeCart(getOwner(), cartItem, worldObj, x, y, z);
                if (cartPlaced != null) {
                    CartTools.getLinkageManager(worldObj).createLink(cartPlaced, lastCart);
                    lastCart = cartPlaced;
                    patternIndex++;
                    if (patternIndex >= getPattern().getSizeInventory()) {
                        resetSpawnSequence();
                    }
                    return true;
                } else {
                    InvTools.moveItemStack(cartItem, this);
                }
            }
        }
        return false;
    }

    private void resetSpawnSequence() {
        patternIndex = 0;
        spawningTrain = false;
        timeSinceLastSpawn = 0;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();
        update++;

        if (spawningTrain && update % 4 == 0) {
            spawnNextCart();
        }
    }

    @Override
    public void onNeighborBlockChange(int id) {
        if (Game.isNotHost(getWorld())) {
            return;
        }
        boolean newPower = worldObj.isBlockGettingPowered(xCoord, yCoord, zCoord) || worldObj.isBlockIndirectlyGettingPowered(xCoord, yCoord, zCoord);
        if (!powered && newPower) {
            powered = newPower;
            EntityMinecart cart = CartTools.getMinecartOnSide(worldObj, xCoord, yCoord, zCoord, 0, direction);
            if (cart == null) {
                if (!spawningTrain && canBuildTrain()) {
                    if (timeSinceLastSpawn > RailcraftConfig.getCartDispenserMinDelay() * 20) {
                        spawningTrain = true;
                    }
                }
            }
//            else if(!spawningTrain) {
//                ItemStack cartStack = InventoryTools.moveItemStack(cart.getCartItem(), invBuffer);
//                if(cartStack == null) {
//                    cart.setDead();
//                }
//            }
        } else if (powered ^ newPower) {
            powered = newPower;
        }
    }

    @Override
    public int getStartInventorySide(ForgeDirection side) {
        return 0;
    }

    @Override
    public int getSizeInventorySide(ForgeDirection side) {
        return BUFFER_SIZE;
    }

    @Override
    public int getInventoryStackLimit() {
        return 64;
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        data.setBoolean("spawningTrain", spawningTrain);
        data.setByte("patternIndex", patternIndex);

        getPattern().writeToNBT("invPattern", data);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        spawningTrain = data.getBoolean("spawningTrain");
        patternIndex = data.getByte("patternIndex");

        if (data.hasKey("pattern")) {
            NBTTagCompound pattern = data.getCompoundTag("pattern");
            getPattern().readFromNBT("Items", pattern);
        } else {
            getPattern().readFromNBT("invPattern", data);
        }
    }

    @Override
    public int addItem(ItemStack stack, boolean doAdd, ForgeDirection from) {
        if (InvTools.isInventoryEmpty(getPattern())) {
            return 0;
        }
        IInventory inv = new InventoryMapper(this);
        if (!doAdd) {
            inv = new InventoryCopy(inv);
        }
        ItemStack leftOver = InvTools.moveItemStack(stack, inv);
        if (leftOver == null) {
            return stack.stackSize;
        }
        return stack.stackSize - leftOver.stackSize;
    }

    @Override
    public ItemStack[] extractItem(boolean doRemove, ForgeDirection from, int maxItemCount) {
        Set<ItemStack> patternSet = new ItemStackSet();
        Set<ItemStack> bufferSet = new ItemStackSet();

        for (ItemStack stack : getPattern().getContents()) {
            if (stack != null) {
                patternSet.add(stack);
            }
        }

        for (ItemStack stack : getInventory().getContents()) {
            if (stack != null) {
                bufferSet.add(stack);
            }
        }

        bufferSet.removeAll(patternSet);

        IInventory inv = this;
        if (!doRemove) {
            inv = new InventoryCopy(inv);
        }

        for (ItemStack stack : bufferSet) {
            ItemStack removed = InvTools.removeOneItem(inv, stack);
            return removed != null ? new ItemStack[]{removed} : new ItemStack[0];
        }

        return new ItemStack[0];
    }
}
