package railcraft.common.blocks.machine.gamma;

import ic2.api.Direction;
import ic2.api.energy.tile.IEnergySink;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import railcraft.common.api.carts.CartTools;
import railcraft.common.api.carts.IEnergyTransfer;
import railcraft.common.blocks.machine.IEnumMachine;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.util.misc.Game;
import railcraft.common.util.network.IGuiReturnHandler;

public class TileEnergyLoader extends TileLoaderEnergyBase implements IEnergySink, IGuiReturnHandler {

    private static final short[] INPUT_LEVELS = {512, 2048};
    private boolean transferedEnergy;
    private boolean waitTillFull = false;
    private boolean waitIfEmpty = true;

    @Override
    public IEnumMachine getMachineType() {
        return EnumMachineGamma.ENERGY_LOADER;
    }

    @Override
    public int getBlockTexture(int side) {
        if (getFacing() == side) {
            return getMachineType().getTexture(3);
        }
        if (side != 0 && side != 1) {
            return getMachineType().getTexture(2);
        }
        return getMachineType().getTexture(1);
    }

    @Override
    public boolean openGui(EntityPlayer player) {
        GuiHandler.openGui(EnumGui.LOADER_ENERGY, player, worldObj, xCoord, yCoord, zCoord);
        return true;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();

        if (Game.isNotHost(getWorld())) {
            return;
        }

        transferedEnergy = false;
        transferRate = 0;

        EntityMinecart cart = CartTools.getMinecartOnSide(worldObj, xCoord, yCoord, zCoord, 0.1f, direction);

        if (cart != currentCart) {
            setPowered(false);
            currentCart = cart;
            cartSent();
        }

        if (cart == null) {
            return;
        }

        if (!(cart instanceof IEnergyTransfer)) {
            if (CartTools.cartVelocityIsLessThan(cart, STOP_VELOCITY)) {
                setPowered(true);
            }
            return;
        }

        if (isSendCart()) {
            if (CartTools.cartVelocityIsLessThan(cart, STOP_VELOCITY)) {
                setPowered(true);
            }
            return;
        }

        IEnergyTransfer energyCart = (IEnergyTransfer) cart;

        if (!energyCart.canInjectEnergy() || energyCart.getTier() > getTier()) {
            if (CartTools.cartVelocityIsLessThan(cart, STOP_VELOCITY)) {
                setPowered(true);
            }
            return;
        }

        if (energy > 0 && energyCart.getEnergy() < energyCart.getCapacity()) {
            int usage = (int) (energyCart.getTransferLimit() * Math.pow(1.5, overclockerUpgrades));
            int injection = (int) (energyCart.getTransferLimit() * Math.pow(1.3, overclockerUpgrades));
            if (usage > energy) {
                double ratio = (double) energy / (double) usage;
                usage = energy;
                injection = (int) (injection * ratio);
            }

            transferRate = injection;
            int extra = energyCart.injectEnergy(this, injection, getTier(), true, false, false);
            energy -= usage - extra;
            transferedEnergy = extra != injection;
        }

        if (!transferedEnergy && !isPowered() && CartTools.cartVelocityIsLessThan(cart, STOP_VELOCITY)) {
            if (!waitTillFull && energyCart.getEnergy() > 0) {
                setPowered(true);
            } else if (!waitIfEmpty && !waitTillFull && energyCart.getEnergy() == 0) {
                setPowered(true);
            } else if (energyCart.getEnergy() >= energyCart.getCapacity()) {
                setPowered(true);
            }
        }
    }

    @Override
    public void writeToNBT(NBTTagCompound nbttagcompound) {
        super.writeToNBT(nbttagcompound);
        nbttagcompound.setBoolean("WaitIfEmpty", waitIfEmpty());
        nbttagcompound.setBoolean("WaitTillFull", waitTillFull());
    }

    @Override
    public void readFromNBT(NBTTagCompound nbttagcompound) {
        super.readFromNBT(nbttagcompound);
        setWaitIfEmpty(nbttagcompound.getBoolean("WaitIfEmpty"));
        setWaitTillFull(nbttagcompound.getBoolean("WaitTillFull"));
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);

        byte bits = 0;
        bits |= waitIfEmpty ? 1 : 0;
        bits |= waitTillFull ? 2 : 0;
        data.writeByte(bits);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        byte bits = data.readByte();
        waitIfEmpty = (bits & 1) != 0;
        waitTillFull = (bits & 2) != 0;
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException {
        data.writeBoolean(waitIfEmpty);
        data.writeBoolean(waitTillFull);
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException {
        waitIfEmpty = data.readBoolean();
        waitTillFull = data.readBoolean();
    }

    @Override
    public int injectEnergy(Direction directionFrom, int amount) {
        if (amount > INPUT_LEVELS[0] && transformerUpgrades <= 0) {
            worldObj.createExplosion(null, xCoord, yCoord, zCoord, 2F, true);
            return 0;
        }
        energy += amount;
        int capacity = getCapacity();
        if (energy > capacity) {
            int extra = energy - capacity;
            energy = capacity;
            return extra;
        }
        return 0;
    }

    @Override
    public boolean acceptsEnergyFrom(TileEntity emitter, Direction direction) {
        if (getFacing() == direction.toSideValue()) {
            return false;
        }
        return true;
    }

    public boolean waitTillFull() {
        return waitTillFull;
    }

    public void setWaitTillFull(boolean waitTillFull) {
        this.waitTillFull = waitTillFull;
    }

    public boolean waitIfEmpty() {
        return waitIfEmpty;
    }

    public void setWaitIfEmpty(boolean waitIfEmpty) {
        this.waitIfEmpty = waitIfEmpty;
    }

    @Override
    public int demandsEnergy() {
        return getCapacity() - energy;
    }

    @Override
    public int getMaxSafeInput() {
        return transformerUpgrades <= 0 ? INPUT_LEVELS[0] : INPUT_LEVELS[1];
    }
}
