package railcraft.common.blocks.machine.gamma;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.inventory.IInventory;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.liquids.ILiquidTank;
import net.minecraftforge.liquids.ITankContainer;
import net.minecraftforge.liquids.LiquidStack;
import railcraft.common.liquids.LiquidManager;
import railcraft.common.liquids.TankManager;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

public abstract class TileLoaderLiquidBase extends TileLoaderBase implements IInventory, ITankContainer {

    protected final TankManager tankManager = new TankManager();
    protected int update = MiscTools.getRand().nextInt();

    protected TileLoaderLiquidBase() {
        super();
        setInventorySize(2);
    }

    public TankManager getTankManager() {
        return tankManager;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();
        update++;

        if (Game.isHost(getWorld()) && update % LiquidManager.NETWORK_UPDATE_INTERVAL == 0) {
            sendUpdateToClient();
        }
    }

    /**
     * Fills liquid into internal tanks, distribution is left to the
     * ITankContainer.
     *
     * @param from ForgeDirection the liquid is pumped in from.
     * @param resource LiquidStack representing the maximum amount of liquid
     * filled into the ITankContainer
     * @param doFill If false filling will only be simulated.
     * @return Amount of resource that was filled into internal tanks.
     */
    @Override
    public int fill(ForgeDirection from, LiquidStack resource, boolean doFill) {
        return fill(0, resource, doFill);
    }

    /**
     * Fills liquid into the specified internal tank.
     *
     * @param from ForgeDirection the liquid is pumped in from.
     * @param resource LiquidStack representing the maximum amount of liquid
     * filled into the ITankContainer
     * @param doFill If false filling will only be simulated.
     * @return Amount of resource that was filled into internal tanks.
     */
    @Override
    public int fill(int tankIndex, LiquidStack resource, boolean doFill) {
        return tankManager.fill(tankIndex, resource, doFill);
    }

    /**
     * Drains liquid out of internal tanks, distribution is left to the
     * ITankContainer.
     *
     * @param from ForgeDirection the liquid is drained to.
     * @param maxDrain Maximum amount of liquid to drain.
     * @param doDrain If false draining will only be simulated.
     * @return LiquidStack representing the liquid and amount actually drained
     * from the ITankContainer
     */
    @Override
    public LiquidStack drain(ForgeDirection from, int maxDrain, boolean doDrain) {
        return drain(0, maxDrain, doDrain);
    }

    /**
     * Drains liquid out of the specified internal tank.
     *
     * @param maxDrain Maximum amount of liquid to drain.
     * @param doDrain If false draining will only be simulated.
     * @return LiquidStack representing the liquid and amount actually drained
     * from the ITankContainer
     */
    @Override
    public LiquidStack drain(int tankIndex, int maxDrain, boolean doDrain) {
        return tankManager.drain(tankIndex, maxDrain, doDrain);
    }

    @Override
    public ILiquidTank[] getTanks(ForgeDirection dir) {
        return tankManager.getTanks();
    }

    @Override
    public ILiquidTank getTank(ForgeDirection direction, LiquidStack type) {
        return tankManager.getTank(0);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        if (data.getTag("tanks") instanceof NBTTagCompound) {
            data.setTag("tanks", new NBTTagList());
        }
        tankManager.readTanksFromNBT(data);
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        tankManager.writeTanksToNBT(data);
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);

        tankManager.writePacketData(data);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);

        tankManager.readPacketData(data);
    }
}
