package railcraft.common.blocks.signals;

import railcraft.common.api.signals.SimpleSignalReceiver;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.world.IBlockAccess;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.api.signals.IReceiverTile;
import railcraft.common.api.signals.SignalAspect;
import railcraft.common.api.signals.SignalController;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.network.IGuiReturnHandler;

public class TileBoxReceiver extends TileBoxBase implements IAspectActionManager, IGuiReturnHandler, IReceiverTile {

    protected int update = MiscTools.getRand().nextInt();
    private boolean prevBlinkState;
    private boolean[] powerOnAspects = new boolean[SignalAspect.values().length];
    private final SimpleSignalReceiver receiver = new SimpleSignalReceiver(getInvName(), this);

    public TileBoxReceiver() {
        powerOnAspects[SignalAspect.GREEN.ordinal()] = true;

    }

    @Override
    public EnumSignal getSignalType() {
        return EnumSignal.BOX_RECEIVER;
    }

    @Override
    public boolean blockActivated(int side, EntityPlayer player) {
        if (player.isSneaking()) {
            return false;
        }
        ItemStack current = player.getCurrentEquippedItem();
        if (current != null && current.getItem() instanceof ItemSignalTuner) {
            return super.blockActivated(side, player);
        }
        GuiHandler.openGui(EnumGui.BOX_RECEIVER, player, worldObj, xCoord, yCoord, zCoord);
        return true;
    }

    @Override
    public void updateEntity() {
        update++;
        if (Game.isNotHost(getWorld())) {
            receiver.tickClient();
            if (receiver.getAspect().isBlinkAspect() && prevBlinkState != SignalAspect.isBlinkOn()) {
                prevBlinkState = SignalAspect.isBlinkOn();
                markBlockForUpdate();
            }
            return;
        }
        receiver.tickServer();
        SignalAspect prevAspect = receiver.getAspect();
        if (receiver.isBeingPaired()) {
            receiver.setAspect(SignalAspect.BLINK_YELLOW);
        } else if (!receiver.isPaired()) {
            receiver.setAspect(SignalAspect.BLINK_RED);
        }
        if (prevAspect != receiver.getAspect()) {
            updateNeighbors();
            sendUpdateToClient();
        }
    }

    @Override
    public void onControllerAspectChange(SignalController con, SignalAspect aspect) {
        updateNeighbors();
        sendUpdateToClient();
    }

    private void updateNeighbors() {
        int blockID = getBlockId();
        worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord, zCoord, blockID);
        for (int side = 2; side < 6; side++) {
            ForgeDirection forgeSide = ForgeDirection.getOrientation(side);
            TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, xCoord, yCoord, zCoord, forgeSide);
            if (tile instanceof TileBoxBase) {
                TileBoxBase box = (TileBoxBase) tile;
                box.onNeighborStateChange(this);
            }
        }
//        worldObj.notifyBlocksOfNeighborChange(xCoord + 1, yCoord, zCoord, blockID);
//        worldObj.notifyBlocksOfNeighborChange(xCoord - 1, yCoord, zCoord, blockID);
//        worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord, zCoord + 1, blockID);
//        worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord, zCoord - 1, blockID);
//        worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord - 1, zCoord, blockID);
//        worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord + 1, zCoord, blockID);
    }

    @Override
    public boolean isPoweringTo(IBlockAccess world, int i, int j, int k, int side) {
        TileEntity tile = MiscTools.getBlockTileEntityOnSide(worldObj, i, j, k, MiscTools.getOppositeSide(side));
        if (tile instanceof TileBoxBase) {
            return false;
        }
        return powerOnAspects[receiver.getAspect().ordinal()];
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        byte[] array = new byte[powerOnAspects.length];
        for (int i = 0; i < powerOnAspects.length; i++) {
            array[i] = (byte) (powerOnAspects[i] ? 1 : 0);
        }
        data.setByteArray("PowerOnAspect", array);

        receiver.writeToNBT(data);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        if (data.hasKey("PowerOnAspect")) {
            byte[] array = data.getByteArray("PowerOnAspect");
            for (int i = 0; i < powerOnAspects.length; i++) {
                powerOnAspects[i] = array[i] == 1;
            }
        }

        receiver.readFromNBT(data);

        if (data.hasKey("ControllerX")) {
            int cx = data.getInteger("ControllerX");
            int cy = data.getInteger("ControllerY");
            int cz = data.getInteger("ControllerZ");
            receiver.registerLegacyController(cx, cy, cz);
        }
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);
        receiver.writePacketData(data);

        writeGuiData(data);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);
        receiver.readPacketData(data);

        readGuiData(data);

        markBlockForUpdate();
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException {

        byte bits = 0;
        for (int i = 0; i < powerOnAspects.length; i++) {
            bits |= (powerOnAspects[i] ? 1 : 0) << i;
        }
        data.writeByte(bits);
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException {

        byte bits = data.readByte();
        for (int bit = 0; bit < powerOnAspects.length; bit++) {
            powerOnAspects[bit] = ((bits >> bit) & 1) == 1;
        }
        updateNeighbors();
    }

    @Override
    public boolean connectToBoxAt(int i, int j, int k, int side) {
        TileEntity tile = worldObj.getBlockTileEntity(i, j, k);
        if (tile instanceof TileBoxBase) {
            return ((TileBoxBase) tile).canReceiveAspect();
        }
        return false;
    }

    @Override
    public boolean isEmitingRedstone() {
        return powerOnAspects[receiver.getAspect().ordinal()];
    }

    @Override
    public boolean doesActionOnAspect(SignalAspect aspect) {
        return powerOnAspects[aspect.ordinal()];
    }

    @Override
    public void doActionOnAspect(SignalAspect aspect, boolean trigger) {
        powerOnAspects[aspect.ordinal()] = trigger;
        updateNeighbors();
    }

    @Override
    public SignalAspect getBoxSignalAspect() {
        return receiver.getAspect();
    }

    @Override
    public boolean canTransferAspect() {
        return true;
    }

    @Override
    public SimpleSignalReceiver getReceiver() {
        return receiver;
    }
}
