package railcraft.common.blocks.signals;

import railcraft.common.api.signals.SimpleSignalReceiver;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.block.Block;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import railcraft.common.api.core.items.IToolCrowbar;
import railcraft.common.api.signals.IReceiverTile;
import railcraft.common.api.signals.SignalAspect;
import railcraft.common.api.signals.SignalController;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.network.IGuiReturnHandler;

public class TileSwitchMotor extends TileSwitchBase implements IAspectActionManager, IGuiReturnHandler, IReceiverTile {

    private final SimpleSignalReceiver receiver = new SimpleSignalReceiver(getInvName(), this);
    private int update = MiscTools.getRand().nextInt();
    private boolean[] switchOnAspects = new boolean[SignalAspect.values().length];
    private boolean switchAspect;

    public TileSwitchMotor() {
        switchOnAspects[SignalAspect.RED.ordinal()] = true;
    }

    @Override
    public EnumSignal getSignalType() {
        return EnumSignal.SWITCH_MOTOR;
    }

    @Override
    public boolean blockActivated(int side, EntityPlayer player) {
        ItemStack current = player.getCurrentEquippedItem();
        if (current != null && current.getItem() instanceof IToolCrowbar) {
            IToolCrowbar crowbar = (IToolCrowbar) current.getItem();
            if (crowbar.canWhack(player, current, xCoord, yCoord, zCoord)) {
                GuiHandler.openGui(EnumGui.SWITCH_MOTOR, player, worldObj, xCoord, yCoord, zCoord);
                crowbar.onWhack(player, current, xCoord, yCoord, zCoord);
                return true;
            }
        }
        return false;
    }

    @Override
    public void updateEntity() {
        super.updateEntity();
        if (Game.isNotHost(worldObj)) {
            receiver.tickClient();
            return;
        }
        receiver.tickServer();
        update++;
        boolean active = isSwitchAspect();
        if (switchAspect != active) {
            switchAspect = active;
            switchTrack(switchAspect || isPowered());
        }
    }

    @Override
    public void onControllerAspectChange(SignalController con, SignalAspect aspect) {
    }

    @Override
    public void onNeighborBlockChange(int blockId) {
        boolean power = isBeingPowered(Block.redstoneWire.blockID == blockId);
        if (isPowered() != power) {
            setPowered(power);
            switchTrack(switchAspect || isPowered());
        }
    }

    @Override
    public boolean canConnectRedstone(int dir) {
        return true;
    }

    private boolean isBeingPowered(boolean testRedstone) {
        return worldObj.isBlockGettingPowered(xCoord, yCoord, zCoord)
                || worldObj.isBlockIndirectlyGettingPowered(xCoord, yCoord, zCoord)
                || (testRedstone && isRedstonePowered());
    }

    private boolean isSwitchAspect() {
        return switchOnAspects[receiver.getAspect().ordinal()];
    }

    @Override
    public void writeToNBT(NBTTagCompound data) {
        super.writeToNBT(data);

        byte[] array = new byte[switchOnAspects.length];
        for (int i = 0; i < switchOnAspects.length; i++) {
            array[i] = (byte) (switchOnAspects[i] ? 1 : 0);
        }
        data.setByteArray("PowerOnAspect", array);

        data.setBoolean("switchAspect", switchAspect);

        receiver.writeToNBT(data);
    }

    @Override
    public void readFromNBT(NBTTagCompound data) {
        super.readFromNBT(data);

        if (data.hasKey("PowerOnAspect")) {
            byte[] array = data.getByteArray("PowerOnAspect");
            for (int i = 0; i < switchOnAspects.length; i++) {
                switchOnAspects[i] = array[i] == 1;
            }
        }

        switchAspect = data.getBoolean("switchAspect");

        receiver.readFromNBT(data);

        if (data.hasKey("ControllerX")) {
            int cx = data.getInteger("ControllerX");
            int cy = data.getInteger("ControllerY");
            int cz = data.getInteger("ControllerZ");
            receiver.registerLegacyController(cx, cy, cz);
        }
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);
        receiver.writePacketData(data);

        writeGuiData(data);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);
        receiver.readPacketData(data);

        readGuiData(data);

        markBlockForUpdate();
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException {

        byte bits = 0;
        for (int i = 0; i < switchOnAspects.length; i++) {
            bits |= (switchOnAspects[i] ? 1 : 0) << i;
        }
        data.writeByte(bits);
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException {

        byte bits = data.readByte();
        for (int bit = 0; bit < switchOnAspects.length; bit++) {
            switchOnAspects[bit] = ((bits >> bit) & 1) == 1;
        }
    }

    @Override
    public boolean doesActionOnAspect(SignalAspect aspect) {
        return switchOnAspects[aspect.ordinal()];
    }

    @Override
    public void doActionOnAspect(SignalAspect aspect, boolean trigger) {
        switchOnAspects[aspect.ordinal()] = trigger;
    }

    @Override
    public SimpleSignalReceiver getReceiver() {
        return receiver;
    }
}
