package railcraft.common.blocks.tracks;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import net.minecraft.block.BlockRail;
import net.minecraft.nbt.NBTTagCompound;
import railcraft.common.api.tracks.ITrackSwitch;
import railcraft.common.util.misc.Game;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public abstract class TrackSwitchBase extends TrackBaseRailcraft implements ITrackSwitch {

    private static final int SPRING_DURATION = 20;
    protected boolean mirrored;
    protected boolean switched;
    private byte sprung;
    private byte locked;

    @Override
    public boolean canMakeSlopes() {
        return false;
    }

    @Override
    public boolean canUpdate() {
        return true;
    }

    @Override
    public boolean isFlexibleRail() {
        return false;
    }

    @Override
    public boolean isMirrored() {
        return mirrored;
    }

    @Override
    public boolean isSwitched() {
        return !isLocked() && (switched || isSprung());
    }

    public boolean isLocked() {
        return locked > 0;
    }

    //    @Override
    //    public boolean blockActivated(EntityPlayer player)
    //    {
    //        ItemStack current = player.getCurrentEquippedItem();
    //        if(player.isSneaking() && current != null && current.getItem() instanceof ICrowbar) {
    //            int meta = tileEntity.getBlockMetadata();
    //            getWorld().setBlockMetadata(getX(), getY(), getZ(), meta == 0 ? 1 : 0);
    //            markBlockNeedsUpdate();
    //            if(current.isItemStackDamageable()) {
    //                current.damageItem(1, player);
    //            }
    //            return true;
    //        }
    //        return super.blockActivated(player);
    //    }
    @Override
    public void onBlockPlaced() {
        determineTrackMeta();
        determineMirror();
    }

    protected void determineTrackMeta() {
        int i = tileEntity.xCoord;
        int j = tileEntity.yCoord;
        int k = tileEntity.zCoord;
        int meta = tileEntity.getBlockMetadata();
        if (meta == EnumTrackMeta.NORTH_SOUTH.ordinal()) {
            if (BlockRail.isRailBlockAt(getWorld(), i + 1, j, k) && BlockRail.isRailBlockAt(getWorld(), i - 1, j, k)) {
                getWorld().setBlockMetadata(i, j, k, 1);
            }
        } else if (meta == EnumTrackMeta.EAST_WEST.ordinal()) {
            if (BlockRail.isRailBlockAt(getWorld(), i, j, k + 1) && BlockRail.isRailBlockAt(getWorld(), i, j, k - 1)) {
                getWorld().setBlockMetadata(i, j, k, 0);
            }
        }
    }

    protected void determineMirror() {
        int x = tileEntity.xCoord;
        int y = tileEntity.yCoord;
        int z = tileEntity.zCoord;
        int meta = tileEntity.getBlockMetadata();
        boolean prevValue = isMirrored();
        if (meta == EnumTrackMeta.NORTH_SOUTH.ordinal()) {
            int ii = x;
            if (BlockRail.isRailBlockAt(getWorld(), x - 1, y, z)) {
                ii--;
                mirrored = true; // West
            } else {
                ii++;
                mirrored = false; // East
            }
            if (BlockRail.isRailBlockAt(getWorld(), ii, y, z)) {
                int otherMeta = getWorld().getBlockMetadata(ii, y, z);
                if (otherMeta == EnumTrackMeta.NORTH_SOUTH.ordinal()) {
                    getWorld().setBlockMetadata(ii, y, z, EnumTrackMeta.EAST_WEST.ordinal());
                }
            }
        } else if (meta == EnumTrackMeta.EAST_WEST.ordinal()) {
            if (BlockRail.isRailBlockAt(getWorld(), x, y, z - 1)) {
                mirrored = true; // North
            } else {
                mirrored = false; // South
            }
        }

        if (prevValue != isMirrored()) {
            sendUpdateToClient();
        }
    }

    @Override
    public void onNeighborBlockChange(int id) {
        if (BlockRail.isRailBlock(id) && Game.isHost(getWorld())) {
            determineTrackMeta();
            determineMirror();
        }
        super.onNeighborBlockChange(id);
    }

    @Override
    public void writeToNBT(NBTTagCompound nbttagcompound) {
        super.writeToNBT(nbttagcompound);
        nbttagcompound.setBoolean("Direction", mirrored);
        nbttagcompound.setBoolean("Switched", switched);
    }

    @Override
    public void readFromNBT(NBTTagCompound nbttagcompound) {
        super.readFromNBT(nbttagcompound);
        mirrored = nbttagcompound.getBoolean("Direction");
        switched = nbttagcompound.getBoolean("Switched");
    }

    @Override
    public void writePacketData(DataOutputStream data) throws IOException {
        super.writePacketData(data);
        data.writeBoolean(mirrored);
        data.writeBoolean(switched);
        data.writeByte(locked);
        data.writeByte(sprung);
    }

    @Override
    public void readPacketData(DataInputStream data) throws IOException {
        super.readPacketData(data);
        mirrored = data.readBoolean();
        switched = data.readBoolean();
        locked = data.readByte();
        sprung = data.readByte();
        markBlockNeedsUpdate();
    }

    @Override
    public void setSwitched(boolean switched) {
        if (this.switched != switched) {
            this.switched = switched;
            sendUpdateToClient();
        }
    }

    public boolean isSprung() {
        return sprung > 0;
    }

    @Override
    public void updateEntity() {
        if (Game.isNotHost(getWorld())) {
            return;
        }

        boolean wasLocked = locked == 0;
        if (locked > 0) {
            locked--;
        }
        if (shouldLockSwitch()) {
            locked = SPRING_DURATION;
        }

        boolean springState = sprung == 0;
        if (sprung > 0) {
            sprung--;
        }
        if (!isLocked()) {
            if (shouldSpringSwitch()) {
                sprung = SPRING_DURATION;
            }
        }

        if (springState != (sprung == 0) || wasLocked != (locked == 0)) {
            sendUpdateToClient();
        }
    }

    protected abstract boolean shouldLockSwitch();

    protected abstract boolean shouldSpringSwitch();
}
