package railcraft.common.carts;

import com.google.common.io.ByteArrayDataInput;
import com.google.common.io.ByteArrayDataOutput;
import cpw.mods.fml.common.registry.IEntityAdditionalSpawnData;
import ic2.api.ElectricItem;
import ic2.api.IElectricItem;
import java.util.ArrayList;
import java.util.List;
import net.minecraft.block.Block;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.world.World;
import railcraft.common.api.carts.CartBase;
import railcraft.common.api.carts.CartTools;
import railcraft.common.api.carts.ICartRenderInterface;
import railcraft.common.api.carts.IEnergyTransfer;
import railcraft.common.api.carts.ILinkageManager;
import railcraft.common.core.RailcraftConfig;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.plugins.ic2.IC2Plugin;
import railcraft.common.util.misc.APIErrorHandler;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.IEnergyDevice;
import railcraft.common.util.misc.MiscTools;

public class EntityCartEnergy extends CartBase implements IEnergyDevice, ICartRenderInterface, IEnergyTransfer, IEntityAdditionalSpawnData
{

    private static final int ENERGY_DATA_ID = 20;
    private static final int[] CAPACITY = {40000, 600000, 10000000};
    private static final int[] TRANSFER_RATE = {128, 512, 2048};
    private int update;

    public EntityCartEnergy(World world)
    {
        super(world);
    }

    public EntityCartEnergy(World world, double d, double d1, double d2)
    {
        this(world);
        setPosition(d, d1 + (double)yOffset, d2);
        motionX = 0.0D;
        motionY = 0.0D;
        motionZ = 0.0D;
        prevPosX = d;
        prevPosY = d1;
        prevPosZ = d2;
    }

    protected void entityInit()
    {
        super.entityInit();
        this.dataWatcher.addObject(ENERGY_DATA_ID, new Integer(0));
    }

    @Override
    public List<ItemStack> getItemsDropped()
    {
        List<ItemStack> items = new ArrayList<ItemStack>();
        ItemStack stack = getIC2Item();
        if(RailcraftConfig.doCartsBreakOnDrop() && stack != null) {
            items.add(new ItemStack(Item.minecartEmpty));
            items.add(stack);
        } else {
            items.add(getCartItem());
        }
        return items;
    }

    @Override
    public void onUpdate()
    {
        update++;

        super.onUpdate();

        if(Game.isNotHost(getWorld())) {
            return;
        }

        if(getEnergy() > getCapacity()) {
            setEnergy(getCapacity());
        }

        ItemStack stack = getStackInSlot(0);
        if(stack != null && stack.getItem() instanceof IElectricItem && getEnergy() > 0) {
            setEnergy(getEnergy() - ElectricItem.charge(stack, getEnergy(), getTier(), false, false));
        }

        stack = getStackInSlot(1);
        if(stack != null && stack.getItem() instanceof IElectricItem && ((IElectricItem)stack.getItem()).canProvideEnergy() && getEnergy() < getCapacity()) {
            setEnergy(getEnergy() + ElectricItem.discharge(stack, getCapacity() - getEnergy(), getTier(), false, false));
        }
    }

    @Override
    public int getTier()
    {
        return minecartType + 1;
    }

    @Override
    protected double getDrag()
    {
        return CartConstants.STANDARD_DRAG;
    }

    @Override
    public boolean doInteract(EntityPlayer player)
    {
        if(Game.isHost(getWorld())) {
            GuiHandler.openGui(EnumGui.CART_ENERGY, player, worldObj, this);
        }
        return true;
    }

    @Override
    public boolean isStorageCart()
    {
        return false;
    }

    @Override
    public boolean canBeRidden()
    {
        return false;
    }

    @Override
    public int getSizeInventory()
    {
        return 2;
    }

    @Override
    public String getInvName()
    {
        switch (minecartType) {
            case 1:
                return "MFE Cart";
            case 2:
                return "MFSU Cart";
            default:
                return "BatBox Cart";
        }
    }

    @Override
    protected void readEntityFromNBT(NBTTagCompound nbttagcompound)
    {
        super.readEntityFromNBT(nbttagcompound);
        setEnergy(nbttagcompound.getInteger("energy"));

        update = MiscTools.getRand().nextInt();
    }

    @Override
    protected void writeEntityToNBT(NBTTagCompound nbttagcompound)
    {
        super.writeEntityToNBT(nbttagcompound);
        nbttagcompound.setInteger("energy", getEnergy());
    }

    @Override
    public int injectEnergy(Object source, int amount, int tier, boolean ignoreTransferLimit, boolean simulate, boolean passAlong)
    {
        if(tier < getTier()) {
            return amount;
        }
        int extra = 0;
        if(!ignoreTransferLimit) {
            extra = Math.max(amount - getTransferLimit(), 0);
            amount = Math.min(amount, getTransferLimit());
        }
        int energy = getEnergy() + amount;
        int capacity = getCapacity();
        if(energy > capacity) {
            extra += energy - capacity;
            energy = capacity;
        }
        if(!simulate) {
            setEnergy(energy);
        }

        if(!passAlong) {
            return extra;
        }

        try {
            ILinkageManager lm = CartTools.getLinkageManager(worldObj);

            EntityMinecart linkedCart = lm.getLinkedCartA(this);
            if(extra > 0 && linkedCart != source && linkedCart instanceof IEnergyTransfer) {
                extra = ((IEnergyTransfer)linkedCart).injectEnergy(this, extra, tier, ignoreTransferLimit, simulate, true);
            }

            linkedCart = lm.getLinkedCartB(this);
            if(extra > 0 && linkedCart != source && linkedCart instanceof IEnergyTransfer) {
                extra = ((IEnergyTransfer)linkedCart).injectEnergy(this, extra, tier, ignoreTransferLimit, simulate, true);
            }
        } catch (Throwable t) {
            APIErrorHandler.versionMismatch(IEnergyTransfer.class);
        }

        return extra;
    }

    @Override
    public int extractEnergy(Object source, int amount, int tier, boolean ignoreTransferLimit, boolean simulate, boolean passAlong)
    {
        if(tier < getTier()) {
            return 0;
        }
        if(!ignoreTransferLimit) {
            amount = Math.min(amount, getTransferLimit());
        }
        int provide = 0;
        int energy = getEnergy();
        provide = Math.min(amount, energy);
        energy -= provide;
        if(energy < 0) {
            energy = 0;
        }
        if(!simulate) {
            setEnergy(energy);
        }

        if(!passAlong) {
            return provide;
        }

        ILinkageManager lm = CartTools.getLinkageManager(worldObj);

        EntityMinecart linkedCart = lm.getLinkedCartA(this);
        if(provide < amount && linkedCart != source && linkedCart instanceof IEnergyTransfer) {
            provide += ((IEnergyTransfer)linkedCart).extractEnergy(this, amount - provide, tier, ignoreTransferLimit, simulate, true);
        }

        linkedCart = lm.getLinkedCartB(this);
        if(provide < amount && linkedCart != source && linkedCart instanceof IEnergyTransfer) {
            provide += ((IEnergyTransfer)linkedCart).extractEnergy(this, amount - provide, tier, ignoreTransferLimit, simulate, true);
        }

        return provide;
    }

    public ItemStack getIC2Item()
    {
        switch (minecartType) {
            case 1:
                return IC2Plugin.getItem("mfeUnit");
            case 2:
                return IC2Plugin.getItem("mfsUnit");
            default:
                return IC2Plugin.getItem("batBox");
        }
    }

    @Override
    public Block getBlock()
    {
        ItemStack stack = getIC2Item();
        if(stack != null) {
            return Block.blocksList[stack.itemID];
        }
        return null;
    }

    @Override
    public int getBlockMetadata()
    {
        ItemStack stack = getIC2Item();
        if(stack != null) {
            return stack.getItemDamage();
        }
        return 0;
    }

    @Override
    public int getCapacity()
    {
        return CAPACITY[minecartType];
    }

    @Override
    public int getTransferLimit()
    {
        return TRANSFER_RATE[minecartType];
    }

    @Override
    public int getEnergyBarScaled(int scale)
    {
        return (getEnergy() * scale) / getCapacity();
    }

    @Override
    public int getEnergy()
    {
        return dataWatcher.getWatchableObjectInt(ENERGY_DATA_ID);
    }

    protected void setEnergy(int energy)
    {
        dataWatcher.updateObject(ENERGY_DATA_ID, energy);
    }

    @Override
    public boolean canExtractEnergy()
    {
        return true;
    }

    @Override
    public boolean canInjectEnergy()
    {
        return true;
    }

    @Override
    public void writeSpawnData(ByteArrayDataOutput data)
    {
        data.writeByte((byte)minecartType);
    }

    @Override
    public void readSpawnData(ByteArrayDataInput data)
    {
        minecartType = data.readByte();
    }
}
