package railcraft.common.carts;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import net.minecraft.block.Block;
import net.minecraft.block.BlockFluid;
import net.minecraft.block.BlockRail;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemBlock;
import net.minecraft.item.ItemStack;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import net.minecraftforge.common.ForgeDirection;
import railcraft.common.api.carts.CartTools;
import railcraft.common.blocks.aesthetics.post.ItemPost;
import railcraft.common.blocks.tracks.EnumTrackMeta;
import railcraft.common.gui.EnumGui;
import railcraft.common.gui.GuiHandler;
import railcraft.common.lang.RailcraftLanguage;
import railcraft.common.util.misc.BallastRegistry;
import railcraft.common.util.misc.Game;
import railcraft.common.util.sounds.SoundHelper;

public class EntityCartUndercutter extends MaintanceCartBase {

    public static final Set<Integer> EXCLUDED_BLOCKS = new HashSet<Integer>();
    private static final int[] TEXTURE = new int[]{208};
    private static final int SLOT_EXIST_UNDER_A = 0;
    private static final int SLOT_EXIST_UNDER_B = 1;
    private static final int SLOT_EXIST_SIDE_A = 2;
    private static final int SLOT_EXIST_SIDE_B = 3;
    private static final int SLOT_REPLACE_UNDER = 4;
    private static final int SLOT_REPLACE_SIDE = 5;
    public static final int SLOT_STOCK_UNDER = 0;
    public static final int SLOT_STOCK_SIDE = 1;

    static {
        EXCLUDED_BLOCKS.add(Block.sand.blockID);
    }

    public static boolean isValidBallast(ItemStack stack) {
        if (stack == null) {
            return false;
        }
        if (EntityCartUndercutter.EXCLUDED_BLOCKS.contains(stack.itemID)) {
            return false;
        }
        if (stack.getItem() instanceof ItemBlock && Block.blocksList[stack.itemID].isOpaqueCube()) {
            return true;
        }
        if (stack.getItem() instanceof ItemPost) {
            return true;
        }
        return false;
    }

    public EntityCartUndercutter(World world) {
        super(world);
    }

    public EntityCartUndercutter(World world, double d, double d1, double d2) {
        this(world);
        setPosition(d, d1 + (double) yOffset, d2);
        motionX = 0.0D;
        motionY = 0.0D;
        motionZ = 0.0D;
        prevPosX = d;
        prevPosY = d1;
        prevPosZ = d2;
    }

    @Override
    public int[] getTextureArray() {
        return TEXTURE;
    }

    @Override
    public int getSizeInventory() {
        return 2;
    }

    @Override
    public void onUpdate() {
        super.onUpdate();
        if (Game.isNotHost(getWorld())) {
            return;
        }

        stockItems(SLOT_REPLACE_UNDER, SLOT_STOCK_UNDER);
        stockItems(SLOT_REPLACE_SIDE, SLOT_STOCK_SIDE);

        int x = MathHelper.floor_double(this.posX);
        int y = MathHelper.floor_double(this.posY);
        int z = MathHelper.floor_double(this.posZ);
        if (BlockRail.isRailBlockAt(this.worldObj, x, y - 1, z)) {
            --y;
        }

        int id = this.worldObj.getBlockId(x, y, z);

        if (BlockRail.isRailBlock(id)) {
            EnumTrackMeta trackMeta = EnumTrackMeta.fromMeta(((BlockRail) Block.blocksList[id]).getBasicRailMetadata(worldObj, this, x, y, z));
            y--;
            replaceUnder(x, y, z, SLOT_EXIST_UNDER_A);
            replaceUnder(x, y, z, SLOT_EXIST_UNDER_B);
            replaceSide(x, y, z, SLOT_EXIST_SIDE_A, trackMeta);
            replaceSide(x, y, z, SLOT_EXIST_SIDE_B, trackMeta);
        }
    }

    private void replaceUnder(int x, int y, int z, int slotExist) {
        replaceWith(x, y, z, slotExist, SLOT_STOCK_UNDER);
    }

    private void replaceSide(int x, int y, int z, int slotExist, EnumTrackMeta trackMeta) {
        if (trackMeta.isEastWestTrack()) {
            replaceWith(x, y, z + 1, slotExist, SLOT_STOCK_SIDE);
            replaceWith(x, y, z - 1, slotExist, SLOT_STOCK_SIDE);
        } else if (trackMeta.isNorthSouthTrack()) {
            replaceWith(x + 1, y, z, slotExist, SLOT_STOCK_SIDE);
            replaceWith(x - 1, y, z, slotExist, SLOT_STOCK_SIDE);
        }
    }

    private void replaceWith(int x, int y, int z, int slotExist, int slotStock) {
        ItemStack exist = patternInv.getStackInSlot(slotExist);
        ItemStack stock = getStackInSlot(slotStock);
        if (exist != null && exist.getItem() instanceof ItemBlock && isValidBallast(stock)) {
            ItemBlock existItem = (ItemBlock) exist.getItem();
            int existMeta = 0;
            if (existItem.getHasSubtypes()) {
                existMeta = existItem.getMetadata(exist.getItemDamage());
            }
            int id = worldObj.getBlockId(x, y, z);
            int oldMeta = worldObj.getBlockMetadata(x, y, z);
            if ((exist.itemID == id && oldMeta == existMeta) || (exist.itemID == Block.dirt.blockID && id == Block.grass.blockID)) {
                if (!BallastRegistry.isItemBallast(stock) || (!worldObj.isAirBlock(x, y - 1, z) && !(Block.blocksList[worldObj.getBlockId(x, y - 1, z)] instanceof BlockFluid))) {
                    List<ItemStack> drops = Block.blocksList[id].getBlockDropped(worldObj, x, y, z, oldMeta, 0);
                    ItemBlock item = (ItemBlock) stock.getItem();
                    int newMeta = 0;
                    if (item.getHasSubtypes()) {
                        newMeta = item.getMetadata(stock.getItemDamage());
                    }
                    if (worldObj.setBlockAndMetadataWithNotify(x, y, z, stock.itemID, newMeta)) {
                        SoundHelper.playBlockSound(worldObj, x, y, z, Block.blocksList[stock.itemID].stepSound.getPlaceSound(), (1f + 1.0F) / 2.0F, 1f * 0.8F, stock.itemID, newMeta);
                        decrStackSize(slotStock, 1);
                        for (ItemStack stack : drops) {
                            CartTools.offerOrDropItem(this, stack);
                        }
                        blink();
                    }
                }
            }
        }
    }

    @Override
    public boolean doInteract(EntityPlayer player) {
        if (Game.isHost(getWorld())) {
            GuiHandler.openGui(EnumGui.CART_UNDERCUTTER, player, worldObj, this);
        }
        return true;
    }

    @Override
    public String getInvName() {
        return RailcraftLanguage.translate(EnumCart.UNDERCUTTER.getTag());
    }

    @Override
    public int addItem(ItemStack stack, boolean doAdd, ForgeDirection from) {
        return 0;
    }
}
