/*
 * This code is the property of CovertJaguar
 * and may only be used with explicit written
 * permission unless otherwise specified on the
 * license page at railcraft.wikispaces.com.
 */
package railcraft.common.carts;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.world.World;
import railcraft.common.api.carts.CartBase;
import railcraft.common.api.carts.ILinkableCart;
import railcraft.common.core.RailcraftConfig;
import railcraft.common.items.ItemOveralls;
import railcraft.common.items.ItemWhistleTuner;
import railcraft.common.util.inventory.InvTools;
import railcraft.common.util.misc.EnumColor;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.network.IGuiReturnHandler;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public abstract class EntityLocomotive extends CartBase implements IDirectionalCart, IGuiReturnHandler, ILinkableCart {

    public enum LocoMode {

        RUNNING, IDLE, SHUTDOWN;
        public static final LocoMode[] VALUES = values();
    }

    public enum LocoSpeed {

        MAX, SLOW, SLOWER, SLOWEST;
        public static final LocoSpeed[] VALUES = values();
    }
    private static final byte PRIMARY_COLOR_DATA_ID = 20;
    private static final byte SECONDARY_COLOR_DATA_ID = 21;
    private static final byte LOCOMOTIVE_MODE_DATA_ID = 22;
    private static final byte LOCOMOTIVE_SPEED_DATA_ID = 23;
    private static final double DRAG_FACTOR = 0.9;
    private static final float NORMAL_FORCE = 0.15f;
    private static final float HS_FORCE = 3.5f * NORMAL_FORCE;
    private static final byte IDLE_FUEL_USE_INTERVAL = 8;
    private static final int WHISTLE_INTERVAL = 256;
    private static final int WHISTLE_DELAY = 160;
    private static final int WHISTLE_CHANCE = 4;
    private int update = MiscTools.getRand().nextInt();
    private int whistleDelay;
    private int tempIdle;
    private float whistlePitch = getNewWhistlePitch();
    protected float renderYaw;
    public LocoMode clientMode = LocoMode.SHUTDOWN;
    public LocoSpeed clientSpeed = LocoSpeed.MAX;

    public EntityLocomotive(World world) {
        super(world);
        setPrimaryColor(EnumColor.LIGHT_GRAY.ordinal());
        setSecondaryColor(EnumColor.GRAY.ordinal());
    }

    public EntityLocomotive(World world, double x, double y, double z) {
        this(world);
        setPosition(x, y + (double) yOffset, z);
        prevPosX = x;
        prevPosY = y;
        prevPosZ = z;
        fuel = 0;
    }

    @Override
    protected void entityInit() {
        super.entityInit();

        dataWatcher.addObject(PRIMARY_COLOR_DATA_ID, Byte.valueOf((byte) 0));
        dataWatcher.addObject(SECONDARY_COLOR_DATA_ID, Byte.valueOf((byte) 0));
        dataWatcher.addObject(LOCOMOTIVE_MODE_DATA_ID, Byte.valueOf((byte) 0));
        dataWatcher.addObject(LOCOMOTIVE_SPEED_DATA_ID, Byte.valueOf((byte) 0));
    }

    public void initEntityFromItem(ItemStack item) {
        NBTTagCompound nbt = item.getTagCompound();
        if (nbt == null) {
            return;
        }
        if (nbt.hasKey("primaryColor")) {
            setPrimaryColor(nbt.getByte("primaryColor"));
        }
        if (nbt.hasKey("secondaryColor")) {
            setSecondaryColor(nbt.getByte("secondaryColor"));
        }
        if (nbt.hasKey("whistlePitch")) {
            whistlePitch = nbt.getFloat("whistlePitch");
        }
    }

    private float getNewWhistlePitch() {
        return 1f + (float) rand.nextGaussian() * 0.2f;
    }

    @Override
    public List<ItemStack> getItemsDropped() {
        List<ItemStack> items = new ArrayList<ItemStack>();
        items.add(getCartItem());
        return items;
    }

    @Override
    public ItemStack getCartItem() {
        ItemStack item = getCartItemBase();
        ItemLocomotive.setItemColorData(item, getPrimaryColor(), getSecondaryColor());
        ItemLocomotive.setItemWhistleData(item, whistlePitch);
        return item;
    }

    protected abstract ItemStack getCartItemBase();

    public LocoMode getMode() {
        return LocoMode.VALUES[dataWatcher.getWatchableObjectByte(LOCOMOTIVE_MODE_DATA_ID)];
    }

    public void setMode(LocoMode mode) {
        if (getMode() != mode) {
            dataWatcher.updateObject(LOCOMOTIVE_MODE_DATA_ID, (byte) mode.ordinal());
        }
    }

    public LocoSpeed getSpeed() {
        return LocoSpeed.VALUES[dataWatcher.getWatchableObjectByte(LOCOMOTIVE_SPEED_DATA_ID)];
    }

    public void setSpeed(LocoSpeed speed) {
        if (getSpeed() != speed) {
            dataWatcher.updateObject(LOCOMOTIVE_SPEED_DATA_ID, (byte) speed.ordinal());
        }
    }

    public boolean isRunning() {
        return fuel > 0 && getMode() == LocoMode.RUNNING && !(isIdle() || isShutdown());
    }

    public boolean isIdle() {
        if (isShutdown()) return false;
        return tempIdle > 0 || getMode() == LocoMode.IDLE || Train.isTrainLockedDown(this);
    }

    public boolean isShutdown() {
        return getMode() == LocoMode.SHUTDOWN;
    }

    public void forceIdle(int ticks) {
        tempIdle = Math.max(tempIdle, ticks);
    }

    @Override
    public void reverse() {
        rotationYaw += 180;
        motionX = -motionX;
        motionZ = -motionZ;
    }

    @Override
    public void setRenderYaw(float yaw) {
        renderYaw = yaw;
    }

    public abstract String getWhistle();

    public final void whistle() {
        if (RailcraftConfig.playSounds() && whistleDelay <= 0) {
            worldObj.playSoundAtEntity(this, getWhistle(), 1, whistlePitch);
            whistleDelay = WHISTLE_DELAY;
        }
    }

    @Override
    public void onUpdate() {
        super.onUpdate();
        update++;

        if (Game.isNotHost(worldObj)) {
            return;
        }

//        if (getEntityData().getBoolean("HighSpeed"))
//            System.out.println(CartTools.getCartSpeedUncapped(this));

        if (whistleDelay > 0) {
            whistleDelay--;
        }

        if (tempIdle > 0) {
            tempIdle--;
        }

        if (update % WHISTLE_INTERVAL == 0 && isRunning() && rand.nextInt(WHISTLE_CHANCE) == 0) {
            whistle();
        }

        if (isShutdown()) {
            double yaw = rotationYaw * Math.PI / 180D;
            double cos = Math.cos(yaw);
            double sin = Math.sin(yaw);
            float limit = 0.05f;
            if (motionX > limit && cos < 0) {
                rotationYaw += 180;
            } else if (motionX < -limit && cos > 0) {
                rotationYaw += 180;
            } else if (motionZ > limit && sin < 0) {
                rotationYaw += 180;
            } else if (motionZ < -limit && sin > 0) {
                rotationYaw += 180;
            }
        }
    }

    @Override
    protected void applyDragAndPushForces() {
        motionX *= getDrag();
        motionY *= 0.0D;
        motionZ *= getDrag();

        if (isRunning()) {
            boolean highSpeed = getEntityData().getBoolean("HighSpeed");
            double yaw = rotationYaw * Math.PI / 180D;
            motionX += Math.cos(yaw) * (highSpeed ? HS_FORCE : NORMAL_FORCE);
            motionZ += Math.sin(yaw) * (highSpeed ? HS_FORCE : NORMAL_FORCE);
        }

        LocoSpeed speed = getSpeed();
        if (speed != LocoSpeed.MAX) {
            float limit = 0.4f;
            switch (speed) {
                case SLOWEST:
                    limit = 0.1f;
                    break;
                case SLOWER:
                    limit = 0.2f;
                    break;
                case SLOW:
                    limit = 0.3f;
                    break;
            }
            motionX = Math.copySign(Math.min(Math.abs(motionX), limit), motionX);
            motionZ = Math.copySign(Math.min(Math.abs(motionZ), limit), motionZ);
        }
    }

    @Override
    protected void updateFuel() {
        if (fuel > 0) {
            if (isRunning() || (isIdle() && update % IDLE_FUEL_USE_INTERVAL == 0)) {
                fuel--;
            } else if (isShutdown()) {
                fuel = 0;
            }
        }
        if (fuel <= 0 && !isShutdown()) {
            fuel = getMoreGoJuice();
        }
        setMinecartPowered(fuel > 0);
    }

    private boolean cartVelocityIsGreaterThan(float vel) {
        return Math.abs(motionX) > vel || Math.abs(motionZ) > vel;
    }

    private boolean isKillabledEntity(Entity entity) {
        if (entity.ridingEntity instanceof EntityMinecart) {
            return false;
        }
        if (!(entity instanceof EntityLiving)) {
            return false;
        }
        if (((EntityLiving)entity).getMaxHealth() >= 100){
            return false;
        }
        return true;
    }

    public int getDamageToRoadKill(EntityLiving entity) {
        if (entity instanceof EntityPlayer) {
            if (ItemOveralls.isPlayerWearing((EntityPlayer) entity)) {
                ItemStack pants = ((EntityPlayer) entity).getCurrentItemOrArmor(2);
                InvTools.damageItem(pants, 5);
                return 4;
            }
        }
        return 50;
    }

    @Override
    public void applyEntityCollision(Entity entity) {
        if (Game.isHost(worldObj)) {
            if (entity.isDead) {
                return;
            }
            if (entity != this.riddenByEntity && cartVelocityIsGreaterThan(0.2f) && isKillabledEntity(entity)) {
                EntityLiving living = (EntityLiving) entity;
                living.attackEntityFrom(new DamageSourceTrain(this), getDamageToRoadKill(living));
                return;
            }
            if (entity instanceof EntityLocomotive) {
                LinkageManager lm = LinkageManager.getInstance(worldObj);
                EntityLocomotive otherLoco = (EntityLocomotive) entity;
                if (!lm.areLinked(this, otherLoco) && cartVelocityIsGreaterThan(0.2f) && otherLoco.cartVelocityIsGreaterThan(0.2f)
                        && (Math.abs(motionX - entity.motionX) > 0.3f || Math.abs(motionZ - entity.motionZ) > 0.3f)) {
                    explode();
                    if (!otherLoco.isDead) {
                        otherLoco.explode();
                    }
                    return;
                }
            }
        }
        super.applyEntityCollision(entity);
    }

    protected void explode() {
        CartUtils.explodeCart(this);
        setDead();
    }

    public abstract int getMoreGoJuice();

    @Override
    protected void updatePushForces() {
    }

    @Override
    public double getDrag() {
        return DRAG_FACTOR;
    }

    @Override
    public boolean doInteract(EntityPlayer player) {
        ItemStack current = player.getCurrentEquippedItem();
        if (current != null && current.getItem() instanceof ItemWhistleTuner) {
            if (RailcraftConfig.playSounds() && whistleDelay <= 0) {
                whistlePitch = getNewWhistlePitch();
                whistle();
                current.damageItem(1, player);
            }
            return true;
        }
        return false;
    }

    @Override
    public void writeEntityToNBT(NBTTagCompound data) {
        super.writeEntityToNBT(data);
        data.setBoolean("isInReverse", isInReverse);

        data.setByte("locomode", (byte) getMode().ordinal());
        data.setByte("locospeed", (byte) getSpeed().ordinal());

        data.setByte("primaryColor", getPrimaryColor());
        data.setByte("secondaryColor", getSecondaryColor());

        data.setFloat("whistlePitch", whistlePitch);
    }

    @Override
    public void readEntityFromNBT(NBTTagCompound data) {
        super.readEntityFromNBT(data);
        isInReverse = data.getBoolean("isInReverse");

        setMode(LocoMode.values()[data.getByte("locomode")]);
        setSpeed(LocoSpeed.values()[data.getByte("locospeed")]);

        setPrimaryColor(data.getByte("primaryColor"));
        setSecondaryColor(data.getByte("secondaryColor"));

        whistlePitch = data.getFloat("whistlePitch");
    }

    @Override
    public void writeGuiData(DataOutputStream data) throws IOException {
        data.writeByte(clientMode.ordinal());
        data.writeByte(clientSpeed.ordinal());
    }

    @Override
    public void readGuiData(DataInputStream data) throws IOException {
        setMode(LocoMode.VALUES[data.readByte()]);
        setSpeed(LocoSpeed.VALUES[data.readByte()]);
    }

    @Override
    public boolean canBeRidden() {
        return false;
    }

    @Override
    public int getSizeInventory() {
        return 0;
    }

    @Override
    public boolean isPoweredCart() {
        return true;
    }

    @Override
    public boolean isLinkable() {
        return true;
    }

    @Override
    public boolean canLinkWithCart(EntityMinecart cart) {
        if (cart instanceof EntityLocomotive) {
            return true;
        }

        LinkageManager lm = LinkageManager.getInstance(worldObj);

        int links = 0;
        EntityMinecart linkA = lm.getLinkedCartA(this);
        if (linkA != null && !(linkA instanceof EntityLocomotive)) {
            return false;
        }

        EntityMinecart linkB = lm.getLinkedCartB(this);
        if (linkB != null && !(linkB instanceof EntityLocomotive)) {
            return false;
        }

        return true;
    }

    @Override
    public boolean hasTwoLinks() {
        return true;
    }

    @Override
    public float getLinkageDistance(EntityMinecart cart) {
        return LinkageManager.LINKAGE_DISTANCE;
    }

    @Override
    public float getOptimalDistance(EntityMinecart cart) {
        return 0.9f;
    }

    @Override
    public boolean canBeAdjusted(EntityMinecart cart) {
        return true;
    }

    @Override
    public void onLinkCreated(EntityMinecart cart) {
    }

    @Override
    public void onLinkBroken(EntityMinecart cart) {
    }

    public abstract String[] getLocomotiveTextures();

    public final byte getPrimaryColor() {
        return dataWatcher.getWatchableObjectByte(PRIMARY_COLOR_DATA_ID);
    }

    public final byte getSecondaryColor() {
        return dataWatcher.getWatchableObjectByte(SECONDARY_COLOR_DATA_ID);
    }

    public final void setPrimaryColor(int color) {
        dataWatcher.updateObject(PRIMARY_COLOR_DATA_ID, (byte) color);
    }

    public final void setSecondaryColor(int color) {
        dataWatcher.updateObject(SECONDARY_COLOR_DATA_ID, (byte) color);
    }
}
