package railcraft.common.carts;

import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraftforge.event.ForgeSubscribe;
import net.minecraftforge.event.entity.minecart.MinecartInteractEvent;
import net.minecraftforge.event.entity.minecart.MinecartUpdateEvent;
import railcraft.common.api.carts.ILinkableCart;
import railcraft.common.api.carts.IMinecart;
import railcraft.common.api.core.items.IToolCrowbar;
import railcraft.common.api.tracks.RailTools;
import railcraft.common.modules.ModuleManager;
import railcraft.common.modules.ModuleManager.Module;
import railcraft.common.util.misc.Vec2D;

public class LinkageHandler {

    public static final String LINK_A_TIMER = "linkA_timer";
    public static final String LINK_B_TIMER = "linkB_timer";
    public static final double LINK_DRAG = 0.90;
    public static final float MAX_DISTANCE = 8f;
    private static final float STIFFNESS = 0.7f;
    private static final float HS_STIFFNESS = 0.7f;
//    private static final float TRANSFER = 0.15f;
    private static final float DAMPING = 0.4f;
    private static final float HS_DAMPING = 0.3f;
    private static final float FORCE_LIMITER = 6f;
    private static LinkageHandler instance;

    private LinkageHandler() {
    }

    public static LinkageHandler getInstance() {
        if (instance == null) {
            instance = new LinkageHandler();
        }
        return instance;
    }

    /**
     * Returns the optimal distance between two linked carts that the
     * LinkageHandler will attempt to maintain at all times.
     *
     * @param cart1
     * @param cart2
     * @return The optimal distance
     */
    private float getOptimalDistance(EntityMinecart cart1, EntityMinecart cart2) {
        float dist = 0;
        if (cart1 instanceof ILinkableCart) {
            dist += ((ILinkableCart) cart1).getOptimalDistance(cart2);
        } else {
            dist += LinkageManager.OPTIMAL_DISTANCE;
        }
        if (cart2 instanceof ILinkableCart) {
            dist += ((ILinkableCart) cart2).getOptimalDistance(cart1);
        } else {
            dist += LinkageManager.OPTIMAL_DISTANCE;
        }
        return dist;
    }

    private boolean canCartBeAdjustedBy(EntityMinecart cart1, EntityMinecart cart2) {
        if (cart1 == cart2) {
            return false;
        }
        if (cart1 instanceof ILinkableCart && !((ILinkableCart) cart1).canBeAdjusted(cart2)) {
            return false;
        }
        if (RailTools.isCartLockedDown(cart1)) {
            return false;
        }
        return true;
    }

    protected void adjustVelocity(EntityMinecart cart1, EntityMinecart cart2, char link) {
        String timer = LINK_A_TIMER;
        if (link == 'B') {
            timer = LINK_B_TIMER;
        }
        if (cart1.worldObj.provider.dimensionId != cart2.worldObj.provider.dimensionId) {
            short count = cart1.getEntityData().getShort(timer);
            count++;
            if (count > 200) {
                LinkageManager.getInstance(cart1.worldObj).breakLink(cart1, cart2);
            }
            cart1.getEntityData().setShort(timer, count);
            return;
        }
        cart1.getEntityData().setShort(timer, (short) 0);

        double dist = cart1.getDistanceToEntity(cart2);
        if (dist > MAX_DISTANCE) {
            LinkageManager.getInstance(cart1.worldObj).breakLink(cart1, cart2);
            return;
        }

        boolean adj1 = canCartBeAdjustedBy(cart1, cart2);
        boolean adj2 = canCartBeAdjustedBy(cart2, cart1);

        Vec2D cart1Pos = new Vec2D(cart1.posX, cart1.posZ);
        Vec2D cart2Pos = new Vec2D(cart2.posX, cart2.posZ);

        Vec2D unit = Vec2D.subtract(cart2Pos, cart1Pos);
        unit.normalize();

        // Energy transfer

//        double transX = TRANSFER * (cart2.motionX - cart1.motionX);
//        double transZ = TRANSFER * (cart2.motionZ - cart1.motionZ);
//
//        transX = limitForce(transX);
//        transZ = limitForce(transZ);
//
//        if(adj1) {
//            cart1.motionX += transX;
//            cart1.motionZ += transZ;
//        }
//
//        if(adj2) {
//            cart2.motionX -= transX;
//            cart2.motionZ -= transZ;
//        }

        // Spring force

        float optDist = getOptimalDistance(cart1, cart2);
        double stretch = dist - optDist;
//        if(Math.abs(stretch) > 0.5) {
//            stretch *= 2;
//        }

        boolean highSpeed = cart1.getEntityData().getBoolean("HighSpeed");

        double stiffness = highSpeed ? HS_STIFFNESS : STIFFNESS;
        double springX = stiffness * stretch * unit.getX();
        double springZ = stiffness * stretch * unit.getY();

        springX = limitForce(springX);
        springZ = limitForce(springZ);

        if (adj1) {
            cart1.motionX += springX;
            cart1.motionZ += springZ;
        }

        if (adj2) {
            cart2.motionX -= springX;
            cart2.motionZ -= springZ;
        }

        // Damping

        Vec2D cart1Vel = new Vec2D(cart1.motionX, cart1.motionZ);
        Vec2D cart2Vel = new Vec2D(cart2.motionX, cart2.motionZ);

        double dot = Vec2D.subtract(cart2Vel, cart1Vel).dotProduct(unit);

        double damping = highSpeed ? HS_DAMPING : DAMPING;
        double dampX = damping * dot * unit.getX();
        double dampZ = damping * dot * unit.getY();

        dampX = limitForce(dampX);
        dampZ = limitForce(dampZ);

        if (adj1) {
            cart1.motionX += dampX;
            cart1.motionZ += dampZ;
        }

        if (adj2) {
            cart2.motionX -= dampX;
            cart2.motionZ -= dampZ;
        }
    }

    private float getTrainMaxSpeed(EntityMinecart cart) {
        LinkageManager lm = LinkageManager.getInstance(cart.worldObj);
        EntityMinecart linkA = lm.getLinkedCartA(cart);
        EntityMinecart linkB = lm.getLinkedCartB(cart);

        float speed = cart.getMaxSpeedRail();
        if (cart instanceof IMinecart) {
            speed = ((IMinecart) cart).getCartMaxSpeed();
        }

        float min1 = speed;
        min1 = Math.min(min1, getTrainMaxSpeedRecursive(linkA, cart));

        float min2 = speed;
        min2 = Math.min(min2, getTrainMaxSpeedRecursive(linkB, cart));


        return Math.min(min1, min2);
    }

    private float getTrainMaxSpeedRecursive(EntityMinecart cart, EntityMinecart prev) {
        if (cart == null) {
            return Float.MAX_VALUE;
        }
        LinkageManager lm = LinkageManager.getInstance(cart.worldObj);
        EntityMinecart linkA = lm.getLinkedCartA(cart);
        EntityMinecart linkB = lm.getLinkedCartB(cart);

        float speed = cart.getMaxSpeedRail();
        if (cart instanceof IMinecart) {
            speed = ((IMinecart) cart).getCartMaxSpeed();
        }

        float min1 = speed;
        if (linkA != prev) {
            min1 = Math.min(min1, getTrainMaxSpeedRecursive(linkA, cart));
        }

        float min2 = speed;
        if (linkB != prev) {
            min2 = Math.min(min2, getTrainMaxSpeedRecursive(linkB, cart));
        }

        return Math.min(min1, min2);
    }

    private void setTrainMaxSpeed(EntityMinecart cart, float trainSpeed) {
        LinkageManager lm = LinkageManager.getInstance(cart.worldObj);
        EntityMinecart linkA = lm.getLinkedCartA(cart);
        EntityMinecart linkB = lm.getLinkedCartB(cart);

        setTrainMaxSpeedRecursive(linkA, cart, trainSpeed);

        setTrainMaxSpeedRecursive(linkB, cart, trainSpeed);

        if (cart instanceof IMinecart) {
            ((IMinecart) cart).setTrainSpeed(trainSpeed);
        } else {
            cart.setMaxSpeedRail(trainSpeed);
        }
    }

    private void setTrainMaxSpeedRecursive(EntityMinecart cart, EntityMinecart prev, float trainSpeed) {
        if (cart == null) {
            return;
        }
        LinkageManager lm = LinkageManager.getInstance(cart.worldObj);
        EntityMinecart linkA = lm.getLinkedCartA(cart);
        EntityMinecart linkB = lm.getLinkedCartB(cart);

        if (linkA != prev) {
            setTrainMaxSpeedRecursive(linkA, cart, trainSpeed);
        }

        if (linkB != prev) {
            setTrainMaxSpeedRecursive(linkB, cart, trainSpeed);
        }

        if (cart instanceof IMinecart) {
            ((IMinecart) cart).setTrainSpeed(trainSpeed);
        } else {
            cart.setMaxSpeedRail(trainSpeed);
        }
    }

    private double limitForce(double force) {
        return Math.copySign(Math.min(Math.abs(force), FORCE_LIMITER), force);
    }

    @ForgeSubscribe
    public void onMinecartUpdate(MinecartUpdateEvent event) {
        EntityMinecart cart = event.minecart;

        LinkageManager lm = LinkageManager.getInstance(cart.worldObj);

        if (cart.isDead) {
            lm.breakLinks(cart);
            lm.removeLinkageId(cart);
            return;
        }

        // Causes a link id cache store
        lm.getLinkageId(cart);

        int launched = cart.getEntityData().getInteger("Launched");
        if (launched > 0) {
            return;
        }

        if (isOnElevator(cart)) {
            return;
        }

        boolean linked = false;

        EntityMinecart link_A = lm.getLinkedCartA(cart);
        if (link_A != null) {
            launched = link_A.getEntityData().getInteger("Launched");
            if (link_A.isDead) {
                lm.breakLinkA(cart);
                lm.removeLinkageId(link_A);
            } else if (launched <= 0 && !isOnElevator(link_A)) {
                linked = true;
                adjustVelocity(cart, link_A, 'A');
            }
        }

        EntityMinecart link_B = lm.getLinkedCartB(cart);
        if (link_B != null) {
            launched = link_B.getEntityData().getInteger("Launched");
            if (link_B.isDead) {
                lm.breakLinkB(cart);
                lm.removeLinkageId(link_B);
            } else if (launched <= 0 && !isOnElevator(link_B)) {
                linked = true;
                adjustVelocity(cart, link_B, 'B');
            }
        }

        if (linked && ModuleManager.isModuleLoaded(Module.LOCOMOTIVES)) {
            cart.motionX *= LINK_DRAG;
            cart.motionZ *= LINK_DRAG;
        }

        if (link_A == null && link_B != null || link_A != null && link_B == null) {
            float trainSpeed = getTrainMaxSpeed(cart);
            setTrainMaxSpeed(cart, trainSpeed);
        } else if (link_A == null && link_B == null) {
            setTrainMaxSpeed(cart, 1.2f);
        }

    }

    @ForgeSubscribe
    public void onMinecartInteract(MinecartInteractEvent event) {
        EntityPlayer player = event.player;
        if (player.getCurrentEquippedItem() != null && player.getCurrentEquippedItem().getItem() instanceof IToolCrowbar) {
            event.setCanceled(true);
        }
    }

    private boolean isOnElevator(EntityMinecart cart) {
        int elevator = cart.getEntityData().getByte("elevator");
        return elevator > 0;
    }
}
