package railcraft.common.core;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import net.minecraft.block.Block;
import net.minecraft.item.Item;
import net.minecraftforge.common.ConfigCategory;
import net.minecraftforge.common.Configuration;
import net.minecraftforge.common.Property;
import railcraft.common.blocks.aesthetics.cube.EnumCube;
import railcraft.common.blocks.aesthetics.wall.EnumWall;
import railcraft.common.blocks.machine.alpha.EnumMachineAlpha;
import railcraft.common.blocks.machine.beta.EnumMachineBeta;
import railcraft.common.blocks.machine.gamma.EnumMachineGamma;
import railcraft.common.blocks.ore.EnumOre;
import railcraft.common.blocks.signals.EnumSignal;
import railcraft.common.blocks.tracks.EnumTrack;
import railcraft.common.carts.EntityTunnelBore;
import railcraft.common.modules.ModuleManager;
import railcraft.common.modules.ModuleManager.Module;
import railcraft.common.liquids.LiquidManager;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;
import railcraft.common.util.misc.RailcraftInstallationException;

public class RailcraftConfig {

    private static final String COMMENT_PREFIX = "\n";
    private static final String COMMENT_SUFFIX = "\n";
//    private static final String COMMENT_PREFIX = "\n\n   # ";
    private static final String CAT_ANCHORS = "anchors";
    private static final String CAT_AURAS = "auras";
    private static final String CAT_LOOT = "loot";
    private static final String CAT_WORLD_GEN = "worldgen";
    private static final String CAT_RECIPES = "recipes";
    private static final String CAT_CARTS = "carts";
    private static final String CAT_SUBBLOCKS = "subblocks";
    private static final String CAT_TWEAKS_CARTS = "tweaks.carts";
    private static final String CAT_TWEAKS_TRACKS = "tweaks.tracks";
    private static final String CAT_TWEAKS_BLOCKS = "tweaks.blocks";
    private static int nextItemId = 7500;
    private static boolean needsInit = true;
    private static Map<String, Integer> itemIDs = new HashMap<String, Integer>();
    private static Set<String> newItemIDs = new TreeSet<String>();
    private static Map<String, Integer> blockIDs = new HashMap<String, Integer>();
    private static Map<String, Boolean> carts = new HashMap<String, Boolean>();
    private static Map<String, Boolean> enabledSubBlocks = new HashMap<String, Boolean>();
    private static Map<String, Boolean> worldGen = new HashMap<String, Boolean>();
    private static Map<String, Boolean> recipes = new HashMap<String, Boolean>();
    private static final Map<String, Integer> lootChances = new HashMap<String, Integer>();
    private static float maxHighSpeed = 1.1f;
    private static boolean boreDestroysBlocks;
    private static boolean boreMinesAllBlocks;
    private static boolean printSignalDebug;
    private static boolean deleteAnchors;
    private static boolean anchorCrafting;
    private static boolean anchorCraftingPersonal;
    private static boolean anchorsCanInteractWithPipes;
    private static boolean printAnchors;
    private static boolean minecartsBreakOnDrop;
    private static boolean adjustBasicCartDrag;
    private static boolean minecartsCollideWithItems;
    private static boolean registerCollisionHandler;
    private static boolean cartsAreSolid;
    private static boolean playSounds;
    private static boolean doUpdateCheck;
    private static int minecartTankCapacity = 16;
    private static int minecartTankFillRate = 25;
    private static int launchRailMaxForce;
    private static int cartDispenserDelay;
    private static int minecartStackSize;
    private static int signalUpdateInterval;
    private static int maxTankSize;
    private static int anchorRefuel;
    private static int anchorRefuelPersonal;
    private static int anchorFuelItemID;
    private static int steelCookTime;
    private static boolean allowTankStacking;
    private static Configuration config;

    private static void checkInit() {
        if (needsInit) {
            Locale locale = Locale.getDefault();
            Locale.setDefault(Locale.ENGLISH);

            config = new Configuration(new File(Railcraft.getMod().getConfigFolder(), "railcraft.cfg"));
            config.load();

            config.categories.remove("entity");

            if (config.hasCategory("block.feature")) {
                ConfigCategory oldCat = config.getCategory("block.feature");
                ConfigCategory newCat = config.getCategory(CAT_SUBBLOCKS);
                newCat.putAll(oldCat);
                config.categories.remove("block.feature");
            }

            if (config.hasCategory("dungeon.loot")) {
                ConfigCategory oldCat = config.getCategory("dungeon.loot");
                ConfigCategory newCat = config.getCategory(CAT_LOOT);
                newCat.putAll(oldCat);
                config.categories.remove("dungeon.loot");
            }

            if (config.hasCategory("world.gen")) {
                ConfigCategory oldCat = config.getCategory("world.gen");
                ConfigCategory newCat = config.getCategory(CAT_WORLD_GEN);
                newCat.putAll(oldCat);
                config.categories.remove("world.gen");
            }

            doUpdateCheck = get("check.version.online", true, "change to '{t}=false' to disable latest version checking");

            playSounds = get("play.sounds", true, "change to '{t}=false' to prevent all mod sounds from playing");

            config.addCustomCategoryComment("tweaks", "Here you can change the behavior of various things");

            loadAnchorSettings();
            loadBlockTweaks();
            loadTrackTweaks();
            loadCartTweaks();
            loadRecipeOption();
            loadCarts();
            loadBlocks();
            loadItems();
            loadBoreMineableBlocks();
            loadLoot();
            loadWorldGen();

            config.save();

            Locale.setDefault(locale);
        }
        needsInit = false;
    }

    private static void loadAnchorSettings() {
        deleteAnchors = get(CAT_ANCHORS, "delete.anchors", false, true, "change to '{t}=true' to delete every World Anchor or Anchor Cart in the world.\nValue resets to false after each session.\nTo disable Anchors completely, disable the ChunkLoading Module from 'modules.cfg'");
        anchorRefuel = get(CAT_ANCHORS, "refuel", -1, 12, 480, "the number of hours that a single Ender Pearl will power a World Anchor or Anchor Cart\n"
                + "this is an approximation only, actual duration is affected by number of chunks loaded and tick rate\n"
                + "set it to -1 to disable the fuel requirement, default = 12, max = 480");
        anchorRefuelPersonal = get(CAT_ANCHORS, "personal.refuel", -1, 12, 480, "the number of hours that a single Ender Pearl will power a Personal Anchor or Personal Anchor Cart\n"
                + "this is an approximation only, actual duration is affected by number of chunks loaded and tick rate\n"
                + "set it to -1 to disable the fuel requirement, default = 12, max = 480");
        anchorCrafting = get(CAT_ANCHORS, "craftable", true, "change to {t}=false to disable World Anchor crafting, they will still be available via Creative");
        anchorCraftingPersonal = get(CAT_ANCHORS, "personal.craftable", true, "change to {t}=false to disable Personal Anchor crafting, they will still be available via Creative");
        printAnchors = get(CAT_ANCHORS, "print.locations", true, "change to {t}=true to print Anchor locations to the log on startup");
        anchorFuelItemID = get(CAT_ANCHORS, "fuel.itemID", 368, "this is the itemID of the item to be used as Anchor Fuel, defaults to Ender Pearls, does not support meta");
        anchorsCanInteractWithPipes = get(CAT_ANCHORS, "interact.with.pipes", true, "change to {t}=false to prevent pipes, tubes, or various other things from interacting with Anchors");
    }

    private static void loadBlockTweaks() {
        cartDispenserDelay = get(CAT_TWEAKS_BLOCKS + ".cartdispenser", "delay", 0, 0, Integer.MAX_VALUE, "set the minimum number of seconds between cart dispensing, default=0");

        maxTankSize = get(CAT_TWEAKS_BLOCKS + ".irontank", "maxsize", 3, 9, 9, "Allows you to set the max Iron Tank base dimension, valid values are 3, 5, 7, and 9");

        allowTankStacking = get(CAT_TWEAKS_BLOCKS + ".irontank", "allow.stacking", true, "Change to '{t}=false' to disable the stacking of Iron Tanks");

        printSignalDebug = get(CAT_TWEAKS_BLOCKS + ".signals", "printDebug", false, "change to '{t}=true' to log debug info for Signal Blocks");
        signalUpdateInterval = get(CAT_TWEAKS_BLOCKS + ".signals", "update.interval", 4, "measured in tick, smaller numbers update more often, resulting in more sensitive signals, but cost more cpu power, default = 4");

    }

    private static void loadTrackTweaks() {
        maxHighSpeed = get(CAT_TWEAKS_TRACKS + ".speed", "max.speed", 0.6f, 1.0f, 1.2f, "change '{t}' to limit max speed on high speed rails, useful if your computer can't keep up with chunk loading, min=0.6, default=1.0, max=1.2");

        launchRailMaxForce = get(CAT_TWEAKS_TRACKS + ".launch", "force.max", 5, 30, 50, "change the value to your desired max launch rail force, min=5, default=30, max=50");
    }

    private static void loadCartTweaks() {
        registerCollisionHandler = get(CAT_TWEAKS_CARTS + ".general", "register.collision.handler", true, "change to '{t}=false' to use a minecart collision handler from a different mod or vanilla behavior");
        cartsAreSolid = get(CAT_TWEAKS_CARTS + ".general", "solid.carts", true,
                "change to '{t}=false' to return minecarts to vanilla player vs cart collision behavior\n"
                + "in vanilla minecarts are ghost like can be walked through\n"
                + "but making carts solid also makes them hard to push by hand\n"
                + "this setting is ignored if aren't using the Railcraft Collision Handler");

        minecartStackSize = get(CAT_TWEAKS_CARTS + ".general", "maxStackSize", 1, 3, 64, "change the value to your desired minecart stack size, vanilla=1, default=3, max=64");

        minecartsBreakOnDrop = get(CAT_TWEAKS_CARTS + ".general", "breakOnDrop", false, "change to '{t}=true' to restore vanilla behavior");
        minecartsCollideWithItems = get(CAT_TWEAKS_CARTS + ".general", "collideWithItems", false, "change to '{t}=true' to restore minecart collisions with dropped items, ignored if 'register.handler.collision=false'");

        adjustBasicCartDrag = get(CAT_TWEAKS_CARTS + ".basic", "adjustDrag", true, "change to '{t}=false' to give basic carts the original vanilla drag values, after changing you may need to replace the carts to see any change in game");

        boreDestroysBlocks = get(CAT_TWEAKS_CARTS + ".bore", "destroyBlocks", false, "change to '{t}=true' to cause the Bore to destroy the blocks it mines instead of dropping them");
        boreMinesAllBlocks = get(CAT_TWEAKS_CARTS + ".bore", "mineAllBlocks", true, "change to '{t}=false' to enable mining checks, use true setting with caution, especially on servers");

        boolean minecartTankCustomize = get(CAT_TWEAKS_CARTS + ".tank", "useCustomValues", false, "change to '{t}=true' to adjust the Tank Cart's capacity and fill rate");

        int capacity = get(CAT_TWEAKS_CARTS + ".tank", "capacity", 1, 16, 32, "change the value to your desired Tank Cart capacity in buckets, min=1, default=16, max=32, ignored if 'tweaks.minecarts.tank.useCustomValues=false'");
        if (minecartTankCustomize) {
            minecartTankCapacity = capacity;
        }

        int fillrate = get(CAT_TWEAKS_CARTS + ".tank", "fillrate", 5, 25, 50, " change the value to your desired Tank Cart fill rate in milli-buckets per tick, min=5, default=25, max=50 \n   # there are 1000 milli-buckets in a bucket, ignored if 'tweaks.minecarts.tank.useCustomValues=false'");
        if (minecartTankCustomize) {
            minecartTankFillRate = fillrate;
        }

    }

    private static void loadRecipeOption() {
        config.addCustomCategoryComment(CAT_RECIPES, "You can add or remove various recipes here");

        ConfigCategory cat = config.getCategory(CAT_RECIPES);
        Iterator<String> keys = cat.keySet().iterator();
        while (keys.hasNext()) {
            String key = keys.next();
            if (key.startsWith("recipe")) {
                keys.remove();
            }
        }

        loadRecipeProperty("minecraft.furnace", "creosote", false, "change to '{t}=true' to add smelting recipes for Creosote Oil to the vanilla furnace");
        loadRecipeProperty("railcraft.track", "useAltRecipes", false, "change to '{t}=true' to use track recipes more similar to vanilla minecraft");
        loadRecipeProperty("railcraft.rockCrusher", "ores", true, "change to '{t}=false' to prevent the game from crushing ores into dusts (only available if IC2 installed)");
        loadRecipeProperty("railcraft.misc", "gunpowder", true, "change to '{t}=false' to disable the sulfur, saltpeter, charcoal dust recipe for gunpowder");
        loadRecipeProperty("railcraft.cart", "bronze", true, "change to '{t}=false' to disable the bronze recipe for minecarts");
        loadRecipeProperty("railcraft.cart", "steel", true, "change to '{t}=false' to disable the steel recipe for minecarts");
        loadRecipeProperty("railcraft.cart", "furnace", false, "change to '{t}=true' to enable the Furnace Minecart recipe");
        loadRecipeProperty("ic2.macerator", "obsidian", false, "change to '{t}=false' to disable the IC2 Macerator recipes for Cushed Obsidian and Obsidian Dust");
        loadRecipeProperty("ic2.macerator", "charcoal", true, "change to '{t}=false' to disable the IC2 Macerator recipe for Charcoal Dust");
        loadRecipeProperty("ic2.macerator", "ores", true, "change to '{t}=false' to disable the IC2 Macerator recipes for Ore Dusts");
        loadRecipeProperty("ic2.macerator", "bones", true, "change to '{t}=false' to disable the IC2 Macerator recipe for Bonemeal");
        loadRecipeProperty("ic2.macerator", "blaze", true, "change to '{t}=false' to disable the IC2 Macerator recipe for Blaze Powder");
        loadRecipeProperty("ic2.macerator", "cobble", true, "change to '{t}=false' to disable the IC2 Macerator recipes for Cobblestone");
        loadRecipeProperty("ic2.macerator", "dirt", true, "change to '{t}=false' to disable the IC2 Macerator recipe for Dirt");
        loadRecipeProperty("ic2.refinedIron", "useSteel", false, "change to '{t}=true' to enable the use of Steel in any IC2 recipe that requires Refined Iron");
        loadRecipeProperty("ic2.refinedIron", "disable", false, "change to '{t}=true' to prevent Refined Iron from being used in any IC2 recipe. Use in combination with 'useSteel'");
        steelCookTime = get("ic2.refinedIron", "steelCookTime", 320, 1280, 5120, "change to adjust the time in ticks it takes to make Steel. Ignored if other two options are false'\nmin=320, default=1280, max=5120");
        loadRecipeProperty("forestry.misc", "fertilizer", true, "change to '{t}=false' to disable the saltpeter recipe for Forestry Fertilizer");
        loadRecipeProperty("forestry.carpenter", "ties", true, "change to '{t}=false' to disable the Carptenter Tie recipe");
    }

    private static void loadWorldGen() {
        config.addCustomCategoryComment(CAT_WORLD_GEN, "You can control which Ores generate in the world here.\n"
                + "If wish to disable world gen entirely,\n"
                + "it is recommended that you disable the World Module in 'modules.cfg' instead.");

        worldGen.put("sulfur", get(config, CAT_WORLD_GEN, "world.generate.sulfur", true));
        worldGen.put("saltpeter", get(config, CAT_WORLD_GEN, "world.generate.saltpeter", true));
    }

    private static void loadLoot() {
        config.addCustomCategoryComment(CAT_LOOT, "Loot chances are defined here.\n"
                + "Smaller values are rarer.\n"
                + "Example Loot:\n"
                + "   Bread = 100\n"
                + "   Redstone = 50\n"
                + "   Record = 5\n"
                + "   Golden Apple = 1");

        loadLootProperty("tie.wood", 20);
        loadLootProperty("cart.basic", 10);
        loadLootProperty("cart.chest", 10);
        loadLootProperty("cart.tnt", 5);
        loadLootProperty("cart.work", 8);
        loadLootProperty("fuel.coke", 20);
        loadLootProperty("creosote.bottle", 20);
        loadLootProperty("track.basic", 30);
        loadLootProperty("steel.ingot", 10);
        loadLootProperty("steel.block", 5);
        loadLootProperty("tool.crowbar", 10);
        loadLootProperty("tool.steel.sword", 5);
        loadLootProperty("tool.steel.shovel", 5);
        loadLootProperty("tool.steel.pickaxe", 5);
        loadLootProperty("tool.steel.axe", 5);
        loadLootProperty("armor.steel.helmet", 5);
        loadLootProperty("armor.steel.plate", 5);
        loadLootProperty("armor.steel.legs", 5);
        loadLootProperty("armor.steel.boots", 5);
    }

    private static void loadCarts() {
        config.addCustomCategoryComment(CAT_CARTS, "Disable individual carts here.");

        loadCartProperty("cart.tnt");
        loadCartProperty("cart.pumpkin");
        loadCartProperty("cart.gift");
        loadCartProperty("cart.tank");
        loadCartProperty("cart.bore");
        loadCartProperty("cart.energy.batbox");
        loadCartProperty("cart.energy.mfe");
        loadCartProperty("cart.energy.mfsu");
        loadCartProperty("cart.anchor");
        loadCartProperty("cart.anchor.personal");
        loadCartProperty("cart.work");
        loadCartProperty("cart.track.relayer");
        loadCartProperty("cart.undercutter");
        loadCartProperty("cart.loco.steam");
    }

    private static void loadBlocks() {
        config.addCustomCategoryComment(Configuration.CATEGORY_BLOCK,
                "This is were block IDs are defined.\n"
                + "Changing these will have adverse effects on existing worlds.\n"
                + "For the list of which sub-blocks are on each ID see the sub-block section below.");

        loadBlockProperty("block.detector", 450);

        String tag = "block.machine.alpha";
        Property prop = config.getCategory(Configuration.CATEGORY_BLOCK).remove("block.utility");
        if (prop != null) {
            prop.setName(tag);
            config.getCategory(Configuration.CATEGORY_BLOCK).put(tag, prop);
        }

        loadBlockProperty("block.machine.alpha", 451);
        loadBlockProperty("block.machine.beta", 452);
        loadBlockProperty("block.machine.gamma", 453);

        loadBlockProperty("block.track", 454);
        loadBlockProperty("block.elevator", 455);

        tag = "block.signal";
        prop = config.getCategory(Configuration.CATEGORY_BLOCK).remove("block.structure");
        if (prop != null) {
            prop.setName(tag);
            config.getCategory(Configuration.CATEGORY_BLOCK).put(tag, prop);
        }

        loadBlockProperty("block.signal", 456);
        loadBlockProperty("block.cube", 457);
        loadBlockProperty("block.ore", 458);
        loadBlockProperty("block.post", 459);
        loadBlockProperty("block.post.metal", 460);
        loadBlockProperty("block.wall", 461);
        loadBlockProperty("block.worldlogic", 462);

        config.addCustomCategoryComment(CAT_SUBBLOCKS, "Here is were you can enable/disable various sub-blocks.\n"
                + "Railcraft will attemtpt to compensate for any missing component by providing alternatives (usually).");

        for (EnumTrack type : EnumTrack.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumCube type : EnumCube.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumWall type : EnumWall.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumOre type : EnumOre.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumMachineAlpha type : EnumMachineAlpha.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumMachineBeta type : EnumMachineBeta.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumMachineGamma type : EnumMachineGamma.values()) {
            loadBlockFeature(type.getTag());
        }

        for (EnumSignal type : EnumSignal.values()) {
            if (type.getModule() != null) {
                loadBlockFeature(type.getTag());
            }
        }
    }

    private static void loadBlockFeature(String tag) {
        tag = MiscTools.cleanTag(tag);
        Property prop = config.get(CAT_SUBBLOCKS, tag, true);
        enabledSubBlocks.put(tag, Boolean.valueOf(prop.value));
    }

    private static void loadItems() {
        config.addCustomCategoryComment(Configuration.CATEGORY_ITEM, "Generally, many items can be disabled by setting the id to zero.\n"
                + "This is not true for all item, so some experimentation may be needed.\n"
                + "Some disabled items will cause a substitute to be used in crafting recipes.");

        loadItemProperty("tool.crowbar");

        loadItemProperty("tool.magnifying.glass");

        loadItemProperty("tool.surveyor");

        loadItemProperty("tool.signal.tuner");
        
        loadItemProperty("tool.whistle.tuner");

        loadItemProperty("backpack.trackman.t1");
        loadItemProperty("backpack.trackman.t2");
        loadItemProperty("backpack.iceman.t1");
        loadItemProperty("backpack.iceman.t2");

        loadItemProperty("liquid.steam");
        loadItemProperty("liquid.creosote.cell");
        changeItemProperty("item.creosote.bottle", "liquid.creosote.bottle");
        changeItemProperty("item.creosote.liquid", "liquid.creosote.liquid");
        changeItemProperty("item.creosote.can", "liquid.creosote.can");
        changeItemProperty("item.creosote.wax", "liquid.creosote.wax");
        changeItemProperty("item.creosote.refactory", "liquid.creosote.refactory");
        changeItemProperty("item.creosote.bucket", "liquid.creosote.bucket");

        loadItemProperty("part.signal.lamp");

        loadItemProperty("part.circuit");

        changeItemProperty("item.coke", "fuel.coke");

        loadItemProperty("part.rebar");

        loadItemProperty("part.rail");
        loadItemProperty("part.plate");
        loadItemProperty("part.gear");
        loadItemProperty("part.tie");

        loadItemProperty("part.turbine.blade");
        loadItemProperty("part.turbine.disk");
        loadItemProperty("part.turbine.rotor");

        loadItemProperty("part.ingot.steel");

        loadItemProperty("tool.steel.sword");
        loadItemProperty("tool.steel.shovel");
        loadItemProperty("tool.steel.pickaxe");
        loadItemProperty("tool.steel.axe");
        loadItemProperty("tool.steel.hoe");

        loadItemProperty("armor.goggles");
        loadItemProperty("armor.overalls");

        loadItemProperty("armor.steel.helmet");
        loadItemProperty("armor.steel.plate");
        loadItemProperty("armor.steel.legs");
        loadItemProperty("armor.steel.boots");

        loadItemProperty("dust");
        loadItemProperty("nugget");

        loadItemProperty("part.railbed");

        changeItemProperty("item.ic2.upgrade.lapotron", "ic2.upgrade.lapotron");

        loadItemProperty("tool.bore.head.diamond");
        loadItemProperty("tool.bore.head.iron");
        loadItemProperty("tool.bore.head.steel");

        loadItemProperty("item.cart.tnt");
        loadItemProperty("item.cart.pumpkin");
        loadItemProperty("item.cart.gift");

        loadItemProperty("item.cart.tank");
        loadItemProperty("item.cart.bore");

        loadItemProperty("item.cart.batbox");
        loadItemProperty("item.cart.mfe");
        loadItemProperty("item.cart.mfsu");

        loadItemProperty("item.cart.anchor");
        loadItemProperty("item.cart.anchor.personal");
        loadItemProperty("item.cart.work");
        loadItemProperty("item.cart.track.relayer");
        loadItemProperty("item.cart.undercutter");

        loadItemProperty("item.cart.loco.steam");

        loadNewItems();
    }

    public static void loadBoreMineableBlocks() {
        String tag = "mineableBlocks";
        Property prop = get(CAT_TWEAKS_CARTS + ".bore", tag, "{}", "add block ids to '{t}' in a common seperated list to define non-vanilla blocks mineable by the tunnel bore \n"
                + "ignored if 'tweaks.carts.bore.mineAllBlocks=true' \n"
                + "metadata sensative entries can be defined in the form 'blockid:metadata' \n"
                + "inclusive ranges can be defined in the form 'startID-endID' \n"
                + "Example:{t}= { 123, 134:3, 450-500 }");
        try {
            for (String segment : prop.value.replaceAll("[{} ]", "").split(",")) {
                if (segment.equals("")) {
                    continue;
                }
                if (segment.contains("-")) {
                    String[] sequence = segment.split("-");
                    int start = Integer.valueOf(sequence[0]);
                    int end = Integer.valueOf(sequence[1]);
                    Game.log(Level.FINEST, "Tunnel Bore: Adding block range to mineable list: {0}-{1}", start, end);
                    for (int block = start; block <= end; block++) {
                        EntityTunnelBore.addMineableBlock(block);
                    }
                } else {
                    String[] entry = segment.split(":");
                    int block = Integer.valueOf(entry[0]);
                    if (block <= 0 || block >= Block.blocksList.length) {
                        throw new Exception("Invalid Block ID = " + block);
                    }
                    int meta = entry.length > 1 ? Integer.valueOf(entry[1]) : -1;
                    Game.log(Level.FINEST, "Tunnel Bore: Adding block to mineable list: {0}:{1}", block, meta);
                    EntityTunnelBore.addMineableBlock(block, meta);
                }
            }
        } catch (Exception ex) {
            Game.log(Level.SEVERE, "Error in railcraft.cfg with {0}", tag);
            throw new RailcraftInstallationException("Railcraft: Configuration Error!"
                    + "\n************************************************************************************"
                    + "\n In railcraft.cfg there was an error with " + tag + "."
                    + "\n " + ex.getMessage()
                    + "\n************************************************************************************"
                    + "\n");
        }
    }

    public static boolean getRecipeConfig(String tag) {
        checkInit();
        Boolean recipe = recipes.get(tag);
        if (recipe == null) {
            throw new RuntimeException("Railcraft Recipe Config Entry does not exist: " + tag);
        }
        return recipe;
    }

    public static boolean useOldRecipes() {
        return getRecipeConfig("railcraft.track.useAltRecipes");
    }

    public static boolean canCrushOres() {
        return getRecipeConfig("railcraft.rockCrusher.ores");
    }

    public static boolean addObsidianRecipesToMacerator() {
        return getRecipeConfig("ic2.macerator.obsidian");
    }

    public static boolean useCreosoteFurnaceRecipes() {
        return getRecipeConfig("minecraft.furnace.creosote");
    }
    
    public static int getSteelCookTime(){
        checkInit();
        return steelCookTime;
    }

    public static boolean doUpdateCheck() {
        checkInit();
        return doUpdateCheck;
    }

    public static boolean boreDestroysBlocks() {
        checkInit();
        return boreDestroysBlocks;
    }

    public static boolean boreMinesAllBlocks() {
        checkInit();
        return boreMinesAllBlocks;
    }

    public static boolean printSignalDebug() {
        checkInit();
        return printSignalDebug;
    }

    public static boolean anchorsCanInteractWithPipes() {
        checkInit();
        return anchorsCanInteractWithPipes;
    }

    public static int anchorRefuel() {
        checkInit();
        return anchorRefuel;
    }

    public static int anchorRefuelPersonal() {
        checkInit();
        return anchorRefuelPersonal;
    }

    public static int anchorFuelItemID() {
        checkInit();
        return anchorFuelItemID;
    }

    public static boolean deleteAnchors() {
        checkInit();
        return deleteAnchors || !ModuleManager.isModuleLoaded(Module.CHUNK_LOADING);
    }

    public static boolean canCraftAnchors() {
        checkInit();
        return anchorCrafting;
    }

    public static boolean canCraftPersonalAnchors() {
        checkInit();
        return anchorCraftingPersonal;
    }

    public static boolean printAnchorLocations() {
        checkInit();
        return printAnchors;
    }

    public static boolean doCartsBreakOnDrop() {
        checkInit();
        return minecartsBreakOnDrop;
    }

    public static boolean adjustBasicCartDrag() {
        checkInit();
        return adjustBasicCartDrag;
    }

    public static boolean doCartsCollideWithItems() {
        checkInit();
        return minecartsCollideWithItems;
    }

    public static boolean useCollisionHandler() {
        checkInit();
        return registerCollisionHandler;
    }

    public static boolean areCartsSolid() {
        return cartsAreSolid;
    }

    public static boolean playSounds() {
        checkInit();
        return playSounds;
    }

    public static float getMaxHighSpeed() {
        checkInit();
        return maxHighSpeed;
    }

    public static int getMinecartStackSize() {
        checkInit();
        return minecartStackSize;
    }

    public static int getLaunchRailMaxForce() {
        checkInit();
        return launchRailMaxForce;
    }

    public static int getCartDispenserMinDelay() {
        checkInit();
        return cartDispenserDelay;
    }

    public static int getTankCartFillRate() {
        checkInit();
        return minecartTankFillRate;
    }

    public static int getTankCartCapacity() {
        checkInit();
        return minecartTankCapacity * LiquidManager.BUCKET_VOLUME;
    }

    public static int getMaxTankSize() {
        checkInit();
        return maxTankSize;
    }

    public static boolean allowTankStacking() {
        checkInit();
        return allowTankStacking;
    }

    public static int getSignalUpdateInterval() {
        checkInit();
        return signalUpdateInterval;
    }

    private static void changeItemProperty(String oldTag, String newTag) {
        Map<String, Property> items = config.getCategory(Configuration.CATEGORY_ITEM);
        Property prop = items.remove(oldTag);
        if (prop != null) {
            prop.setName(newTag);
            items.put(newTag, prop);
        }

        loadItemProperty(newTag);
    }

    private static void loadItemProperty(String tag) {
        tag = MiscTools.cleanTag(tag);
        Map<String, Property> items = config.getCategory(Configuration.CATEGORY_ITEM);
        Property prop = items.get(tag);
        if (prop == null) {
            newItemIDs.add(tag);
            return;
        }
        itemIDs.put(tag, prop.getInt());
    }

    private static void loadNewItems() {
        for (String tag : newItemIDs) {
            while (Item.itemsList[nextItemId + Item.shovelSteel.itemID] != null || itemIDs.values().contains(nextItemId)) {
                nextItemId++;
            }
            Property prop = config.getItem(tag, nextItemId);
            itemIDs.put(tag, prop.getInt());
            nextItemId++;
        }
    }

    private static void loadBlockProperty(String tag, int defaultValue) {
        Property prop = config.getBlock(tag, defaultValue);
        blockIDs.put(tag, prop.getInt());
    }

    public static int getItemId(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Integer id = itemIDs.get(tag);
        if (id == null) {
            throw new IllegalArgumentException("RailcraftConfig: item tag not found: " + tag);
        }
        return id;
    }

    public static int getBlockId(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Integer id = blockIDs.get(tag);
        if (id == null) {
            return 0;
        }
        return id;
    }

    public static boolean isSubBlockEnabled(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Boolean b = enabledSubBlocks.get(tag);
        if (b == null) {
            return false;
        }
        return b;
    }

    public static boolean isCartEnabled(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Boolean enabled = carts.get(tag);
        if (enabled == null) {
            return false;
        }
        return enabled;
    }

    public static int getLootChance(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Integer chance = lootChances.get(tag);
        if (chance == null) {
            throw new RuntimeException("Railcraft Loot Chance Entry does not exist: " + tag);
        }
        return chance;
    }

    public static boolean isWorldGenEnabled(String tag) {
        checkInit();
        tag = MiscTools.cleanTag(tag);
        Boolean gen = worldGen.get(tag);
        if (gen == null) {
            throw new RuntimeException("Railcraft World Gen Entry does not exist: " + tag);
        }
        return gen;
    }

    private static boolean get(String tag, boolean defaultValue, String comment) {
        return get(Configuration.CATEGORY_GENERAL, tag, defaultValue, comment);
    }

    private static void loadRecipeProperty(String subcat, String tag, boolean defaultValue, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(CAT_RECIPES + "." + subcat, tag, defaultValue);
        prop.comment = comment;
        recipes.put(subcat + "." + tag, Boolean.valueOf(prop.value));
    }

    private static boolean get(String cat, String tag, boolean defaultValue, String comment) {
        return get(cat, tag, defaultValue, false, comment);
    }

    private static boolean get(String cat, String tag, boolean defaultValue, boolean reset, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(cat, tag, defaultValue);
        prop.comment = comment;
        boolean ret = prop.getBoolean(defaultValue);
        if (reset) {
            prop.value = Boolean.toString(defaultValue);
        }
        return ret;
    }

    private static boolean get(Configuration config, String cat, String tag, boolean defaultValue) {
        Property prop = config.get(cat, tag, defaultValue);
        return prop.getBoolean(defaultValue);
    }

    private static int get(String tag, int defaultValue, String comment) {
        return get(Configuration.CATEGORY_GENERAL, tag, defaultValue, comment);
    }

    private static int get(String cat, String tag, int defaultValue, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(cat, tag, defaultValue);
        prop.comment = comment;
        return prop.getInt();
    }

    private static int get(String cat, String tag, int min, int defaultValue, int max, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(cat, tag, defaultValue);
        prop.comment = comment;
        int value = prop.getInt(defaultValue);
        value = Math.max(value, min);
        value = Math.min(value, max);
        prop.value = Integer.toString(value);
        return value;
    }

    private static float get(String cat, String tag, float min, float defaultValue, float max, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(cat, tag, defaultValue);
        prop.comment = comment;
        double value = prop.getDouble(defaultValue);
        value = Math.max(value, min);
        value = Math.min(value, max);
        prop.value = Double.toString(value);
        return (float) value;
    }

    private static Property get(String tag, String defaultValue, String comment) {
        return get(Configuration.CATEGORY_GENERAL, tag, defaultValue, comment);
    }

    private static Property get(String cat, String tag, String defaultValue, String comment) {
        comment = COMMENT_PREFIX + comment.replace("{t}", tag) + COMMENT_SUFFIX;
        Property prop = config.get(cat, tag, defaultValue);
        prop.comment = comment;
        return prop;
    }

    private static void loadLootProperty(String tag, int defaultValue) {
        Property prop = config.get(CAT_LOOT, tag, defaultValue);
        if (!prop.isIntValue()) {
            prop.value = Integer.toString(defaultValue);
        }
        lootChances.put(tag, prop.getInt());
    }

    private static void loadCartProperty(String tag) {
        Property prop = config.get(CAT_CARTS, tag, true);
        carts.put(tag, Boolean.valueOf(prop.value));
    }
}
