package railcraft.common.lang;

import cpw.mods.fml.common.registry.LanguageRegistry;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;
import java.util.logging.Level;
import net.minecraft.util.StringTranslate;
import railcraft.common.core.Railcraft;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public class RailcraftLanguage {

    public static final String ENGLISH = "en_US";
    private Map<String, Properties> languageTable = new TreeMap<String, Properties>();
    private static RailcraftLanguage instance;

    public RailcraftLanguage() {
        loadDefaultLanguage(ENGLISH, "");
        loadDefaultLanguage("en_GB", ENGLISH);
        loadDefaultLanguage("de_DE", ENGLISH);
        loadDefaultLanguage("pt_PT", ENGLISH);
        loadDefaultLanguage("pt_BR", "pt_PT");
        loadDefaultLanguage("es_ES", ENGLISH);
        loadDefaultLanguage("es_AR", "es_ES");
        loadDefaultLanguage("es_MX", "es_ES");
        loadDefaultLanguage("es_UY", "es_ES");
        loadDefaultLanguage("es_VE", "es_ES");
        loadDefaultLanguage("ru_RU", ENGLISH);
        loadDefaultLanguage("fr_FR", ENGLISH);
        loadDefaultLanguage("ja_JP", ENGLISH);
        loadDefaultLanguage("it_IT", ENGLISH);
        loadDefaultLanguage("zh_TW", ENGLISH);
        loadDefaultLanguage("zh_CN", "zh_TW");
        loadDefaultLanguage("nl_NL", ENGLISH);
        loadDefaultLanguage("pl_PL", ENGLISH);

        File langFolder = new File(Railcraft.getMod().getConfigFolder(), "lang");
        if (!langFolder.exists()) {
            langFolder.mkdirs();
        }
        for (File file : langFolder.listFiles()) {
            String lang = file.getName();
            lang = lang.replace(".lang", "");
            loadCustomLanguage(file, lang);
        }
    }

    public static RailcraftLanguage getInstance() {
        if (instance == null) {
            instance = new RailcraftLanguage();
        }
        return instance;
    }

    public static String translate(String tag) {
        return getInstance().translate_do(tag);
    }

    private String translate_do(String tag) {
        String lang = StringTranslate.getInstance().getCurrentLanguage();
        return translate(tag, lang);
    }

    public static boolean hasTag(String tag) {
        return getInstance().hasTag_do(tag);
    }

    private boolean hasTag_do(String tag) {
        String lang = StringTranslate.getInstance().getCurrentLanguage();
        return hasTag(tag, lang);
    }

    private void loadDefaultLanguage(String lang, String parent) {
        Properties langFile = new Properties(languageTable.get(parent));
        InputStreamReader reader = null;
        try {
            InputStream inStream = getClass().getResourceAsStream(lang + ".lang");
            if (inStream == null) {
                throw new IOException();
            }
            reader = new InputStreamReader(inStream, "UTF-8");
            langFile.load(reader);
        } catch (IOException ex) {
            Game.log(Level.WARNING, "Default language file not found: {0}", lang);
        } finally {
            try {
                if (reader != null) reader.close();
            } catch (IOException ex) {
            }
        }
        languageTable.put(lang, langFile);
    }

    private void loadCustomLanguage(File file, String lang) {
        Properties parent = languageTable.get(lang);
        if (parent == null) {
            parent = languageTable.get(ENGLISH);
        }
        Properties langFile = new Properties(parent);
        InputStreamReader reader = null;
        try {
            reader = new InputStreamReader(new FileInputStream(file), "UTF-8");
            langFile.load(reader);
            Game.log(Level.INFO, "Custom language file loaded: {0}", lang);
        } catch (IOException ex) {
            Game.log(Level.INFO, "Custom language file could not be opened: {0}", lang);
        } finally {
            try {
                if (reader != null) reader.close();
            } catch (IOException ex) {
            }
        }
        languageTable.put(lang, langFile);
    }

    public String translate(String tag, String lang) {
        tag = MiscTools.cleanTag(tag);
        Properties languageSet = languageTable.get(lang);
        if (languageSet == null) {
            languageSet = languageTable.get(ENGLISH);
        }
        String value = languageSet.getProperty(tag);
        if (value == null) {
            Game.log(Level.WARNING, "Language Tag is unknown: {0}, {1}", lang, tag);
            return tag;
        }
        return value;
    }

    public boolean hasTag(String tag, String lang) {
        tag = MiscTools.cleanTag(tag);
        Properties languageSet = languageTable.get(lang);
        if (languageSet == null) {
            languageSet = languageTable.get(ENGLISH);
        }
        String value = languageSet.getProperty(tag);
        return value != null;
    }

    public void registerItemName(Object o, String tag) {
        tag = MiscTools.cleanTag(tag);
        for (String lang : languageTable.keySet()) {
            LanguageRegistry.instance().addNameForObject(o, lang, translate(tag, lang));
        }
    }

    public void registerDepreciatedItemName(Object o, String tag) {
        tag = MiscTools.cleanTag(tag);
        for (String lang : languageTable.keySet()) {
            LanguageRegistry.instance().addNameForObject(o, lang, translate(tag, lang) + " (" + translate("obsolete", lang) + ")");
        }
    }
}
