package railcraft.common.liquids;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import net.minecraft.inventory.Container;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.ICrafting;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.liquids.ILiquidTank;
import net.minecraftforge.liquids.ITankContainer;
import net.minecraftforge.liquids.LiquidStack;
import net.minecraftforge.liquids.LiquidTank;
import railcraft.common.liquids.tanks.StandardTank;
import railcraft.common.util.network.PacketGuiUpdate;

/**
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public class TankManager implements ITankContainer {

    private static final byte NETWORK_DATA = 3;
    private List<StandardTank> tanks = new ArrayList<StandardTank>();
    private List<LiquidStack> prevLiquids = new ArrayList<LiquidStack>();

    public void addTank(StandardTank tank) {
        tanks.add(tank);
        int index = tanks.indexOf(tank);
        tank.setTankIndex(index);
        prevLiquids.add(tank.getLiquid() == null ? null : tank.getLiquid().copy());
    }

    public void addTanks(Collection<StandardTank> tanks) {
        for (StandardTank tank : tanks) {
            addTank(tank);
        }
    }

    public void addTanks(StandardTank[] tanks) {
        for (StandardTank tank : tanks) {
            addTank(tank);
        }
    }

    public void writeTanksToNBT(NBTTagCompound data) {
        NBTTagList tagList = new NBTTagList();
        for (byte slot = 0; slot < tanks.size(); slot++) {
            StandardTank tank = tanks.get(slot);
            if (tank.getLiquid() != null) {
                NBTTagCompound tag = new NBTTagCompound();
                tag.setByte("tank", slot);
                tank.writeToNBT(tag);
                tagList.appendTag(tag);
            }
        }
        data.setTag("tanks", tagList);
    }

    public void readTanksFromNBT(NBTTagCompound data) {
        NBTTagList tagList = data.getTagList("tanks");
        for (int tagIndex = 0; tagIndex < tagList.tagCount(); tagIndex++) {
            NBTTagCompound tag = (NBTTagCompound) tagList.tagAt(tagIndex);
            int slot = tag.getByte("tank");
            if (slot >= 0 && slot < tanks.size()) {
                tanks.get(slot).readFromNBT(tag);
            }
        }
    }

    public void writePacketData(DataOutputStream data) throws IOException {
        for (int i = 0; i < tanks.size(); i++) {
            writePacketData(data, i);
        }
    }

    public void writePacketData(DataOutputStream data, int tankIndex) throws IOException {
        if (tankIndex >= tanks.size()) {
            return;
        }
        ILiquidTank tank = tanks.get(tankIndex);
        LiquidStack liquid = tank.getLiquid();
        if (liquid == null) {
            liquid = new LiquidStack(0, 0, 0);
        }
        data.writeShort(liquid.itemID);
        data.writeShort(liquid.itemMeta);
        data.writeInt(liquid.amount);
    }

    public void readPacketData(DataInputStream data) throws IOException {
        for (int i = 0; i < tanks.size(); i++) {
            readPacketData(data, i);
        }
    }

    public void readPacketData(DataInputStream data, int tankIndex) throws IOException {
        if (tankIndex >= tanks.size()) {
            return;
        }
        LiquidTank tank = tanks.get(tankIndex);
        LiquidStack liquid = new LiquidStack(0, 0, 0);
        liquid.itemID = data.readShort();
        liquid.itemMeta = data.readShort();
        liquid.amount = data.readInt();
        if (liquid.itemID == 0) {
            liquid = null;
        }
        tank.setLiquid(liquid);
    }

    public void initGuiData(Container container, ICrafting player, int tankIndex) {
        if (tankIndex >= tanks.size()) {
            return;
        }
        LiquidStack liquid = tanks.get(tankIndex).getLiquid();
        int liquidId = 0;
        int liquidMeta = 0;
        int liquidQty = 0;
        if (liquid != null) {
            liquidId = liquid.itemID;
            liquidMeta = liquid.itemMeta;
            liquidQty = liquid.amount;
        }


        player.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 0, liquidId);
        player.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 1, liquidMeta);
        PacketGuiUpdate pkt = new PacketGuiUpdate(container.windowId, tankIndex * NETWORK_DATA + 2, liquidQty);
        pkt.sendPacket((EntityPlayer) player);
    }

    public void updateGuiData(Container container, List crafters, int tankIndex) {
        for (int i = 0; i < crafters.size(); i++) {
            ICrafting crafter = (ICrafting) crafters.get(i);
            EntityPlayer player = (EntityPlayer) crafters.get(i);

            LiquidStack liquid = tanks.get(tankIndex).getLiquid();
            LiquidStack prev = prevLiquids.get(tankIndex);

            if (liquid == null ^ prev == null) {
                int liquidId = 0;
                int liquidMeta = 0;
                int liquidQty = 0;
                if (liquid != null) {
                    liquidId = liquid.itemID;
                    liquidMeta = liquid.itemMeta;
                    liquidQty = liquid.amount;
                }
                crafter.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 0, liquidId);
                crafter.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 1, liquidMeta);
                PacketGuiUpdate pkt = new PacketGuiUpdate(container.windowId, tankIndex * NETWORK_DATA + 2, liquidQty);
                pkt.sendPacket(player);
            } else if (liquid != null && prev != null) {
                if (liquid.itemID != prev.itemID) {
                    crafter.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 0, liquid.itemID);
                }

                if (liquid.itemMeta != prev.itemMeta) {
                    crafter.sendProgressBarUpdate(container, tankIndex * NETWORK_DATA + 1, liquid.itemMeta);
                }

                if (liquid.amount != prev.amount) {
                    PacketGuiUpdate pkt = new PacketGuiUpdate(container.windowId, tankIndex * NETWORK_DATA + 2, liquid.amount);
                    pkt.sendPacket(player);
                }
            }
        }

        ILiquidTank tank = tanks.get(tankIndex);
        prevLiquids.set(tankIndex, tank.getLiquid() == null ? null : tank.getLiquid().copy());
    }

    public void processGuiUpdate(int messageId, int data) {
        int tankIndex = messageId / NETWORK_DATA;

        if (tankIndex >= tanks.size()) {
            return;
        }
        LiquidStack liquid = tanks.get(tankIndex).getLiquid();
        if (liquid == null) {
            liquid = new LiquidStack(0, 0);
            tanks.get(tankIndex).setLiquid(liquid);
        }

        switch (messageId % NETWORK_DATA) {
            case 0:
                liquid.itemID = data;
                break;
            case 1:
                liquid.itemMeta = data;
                break;
            case 2:
                liquid.amount = data;
                break;
        }
    }

    @Override
    public int fill(ForgeDirection from, LiquidStack resource, boolean doFill) {
        return fill(0, resource, doFill);
    }

    @Override
    public int fill(int tankIndex, LiquidStack resource, boolean doFill) {
        if (tankIndex < 0 || tankIndex >= tanks.size() || resource == null) {
            return 0;
        }

        return tanks.get(tankIndex).fill(resource, doFill);
    }

    @Override
    public LiquidStack drain(ForgeDirection from, int maxDrain, boolean doDrain) {
        return drain(0, maxDrain, doDrain);
    }

    @Override
    public LiquidStack drain(int tankIndex, int maxDrain, boolean doDrain) {
        if (tankIndex < 0 || tankIndex >= tanks.size()) {
            return null;
        }

        return tanks.get(tankIndex).drain(maxDrain, doDrain);
    }

    @Override
    public ILiquidTank[] getTanks(ForgeDirection direction) {
        ILiquidTank[] result = new ILiquidTank[tanks.size()];
        for (int i = 0; i < result.length; i++) {
            result[i] = new ImmutableTankWrapper(tanks.get(i));
        }
        return result;
    }

    public ILiquidTank[] getTanks() {
        return getTanks(ForgeDirection.UNKNOWN);
    }

    public StandardTank getTank(int tankIndex) {
        if (tankIndex < 0 || tankIndex >= tanks.size()) {
            return null;
        }
        return tanks.get(tankIndex);
    }

    public void setCapacity(int tankIndex, int capacity) {
        LiquidTank tank = getTank(tankIndex);
        tank.setCapacity(capacity);
        LiquidStack liquid = tank.getLiquid();
        if (liquid != null && liquid.amount > capacity) {
            liquid.amount = capacity;
        }
    }

    public void outputLiquid(ITankContainer[] outputs, int tankIndex, int amount) {
        for (int side = 0; side < 6; side++) {
            ITankContainer nearbyTank = outputs[side];
            if (nearbyTank != null) {
                ForgeDirection dir = ForgeDirection.getOrientation(side);
                dir = dir.getOpposite();
                LiquidStack available = drain(tankIndex, amount, false);
                if (available != null) {
                    int used = nearbyTank.fill(dir, available, true);
                    drain(tankIndex, used, true);
                } else {
                    break;
                }
            }
        }
    }

    private boolean tankMatchesLiquid(ILiquidTank tank, LiquidStack liquid) {
        if (liquid == null) {
            return false;
        }
        if (tank.fill(liquid, false) > 0) {
            return true;
        }
        if (liquid.isLiquidEqual(tank.getLiquid())) {
            return true;
        }
        return false;
    }

    @Override
    public ILiquidTank getTank(ForgeDirection side, LiquidStack type) {
        if (type == null) {
            return null;
        }
        for (ILiquidTank tank : tanks) {
            if (tankMatchesLiquid(tank, type)) {
                return tank;
            }
        }
        return null;
    }
}
