package railcraft.common.modules;

import java.io.File;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import net.minecraftforge.common.Configuration;
import net.minecraftforge.common.Property;
import railcraft.common.core.Railcraft;
import railcraft.common.util.misc.Game;

public abstract class ModuleManager {

    public static final String MODULE_CONFIG_FILE_NAME = "modules.cfg";
    public static final String CATEGORY_MODULES = "modules";

    public enum Module {

        CORE, FACTORY, EXTRAS, TRACKS, TRACKS_HIGHSPEED, TRACKS_WOOD, TRACKS_REINFORCED, SIGNALS, STRUCTURES, AUTOMATION, TRANSPORT, IC2, FORESTRY, ENERGY, WORLD, CHUNK_LOADING, SEASONAL, TRAIN, LOCOMOTIVES
    };
    private static Map<Module, RailcraftModule> modules = new EnumMap<Module, RailcraftModule>(Module.class);
    private static Set<Module> loadedModules = EnumSet.noneOf(Module.class);
    private static Set<Module> unloadedModules = EnumSet.noneOf(Module.class);

    public static void preInit() {
        registerModule(Module.CORE, new ModuleCore());
        registerModule(Module.AUTOMATION, new ModuleAutomation());
        registerModule(Module.SIGNALS, new ModuleSignals());
        registerModule(Module.STRUCTURES, new ModuleStructures());
        registerModule(Module.EXTRAS, new ModuleExtras());
        registerModule(Module.FACTORY, new ModuleFactory());
        registerModule(Module.TRANSPORT, new ModuleTransport());
        registerModule(Module.TRACKS, new ModuleTrack());
        registerModule(Module.TRACKS_HIGHSPEED, new ModuleTracksHighSpeed());
        registerModule(Module.TRACKS_WOOD, new ModuleTracksWood());
        registerModule(Module.TRACKS_REINFORCED, new ModuleTracksReinforced());
        registerModule(Module.IC2, new ModuleIC2());
        registerModule(Module.FORESTRY, new ModuleForestry());
        registerModule(Module.ENERGY, new ModuleEnergy());
        registerModule(Module.WORLD, new ModuleWorld());
        registerModule(Module.CHUNK_LOADING, new ModuleChunkLoading());
        registerModule(Module.SEASONAL, new ModuleSeasonal());
        registerModule(Module.TRAIN, new ModuleTrain());
        registerModule(Module.LOCOMOTIVES, new ModuleLocomotives());

        Locale locale = Locale.getDefault();
        Locale.setDefault(Locale.ENGLISH);

        Configuration config = new Configuration(new File(Railcraft.getMod().getConfigFolder(), MODULE_CONFIG_FILE_NAME));

        config.load();
        config.addCustomCategoryComment(CATEGORY_MODULES, "Disabling these modules can greatly change how the mod functions.\n"
                + "For example, disabling the Train Module will prevent you from linking carts.\n"
                + "Disabling the World Module will disable all world gen.\n"
                + "Disabling the Energy Module will remove the energy requirement from machines, "
                + "but will only do so if Forestry or Buildcraft are not installed.");

        config.categories.remove(Configuration.CATEGORY_BLOCK);
        config.categories.remove(Configuration.CATEGORY_ITEM);
        config.categories.remove(Configuration.CATEGORY_GENERAL);

        for (Module m : Module.values()) {
            if (m == Module.CORE || isEnabled(config, m)) {
                RailcraftModule instance = modules.get(m);
                if (instance.canModuleLoad()) {
                    loadedModules.add(m);
                } else {
                    unloadedModules.add(m);
                    instance.printLoadError();
                }
            } else {
                unloadedModules.add(m);
                Game.log(Level.FINER, "Module disabled: {0}", m);
            }
        }
        config.save();

        Locale.setDefault(locale);

        for (Module m : loadedModules) {
            preInit(m);
        }
    }

    public static void init() {
        for (Module m : loadedModules) {
            initFirst(m);
        }

        for (Module m : loadedModules) {
            initSecond(m);
        }
    }

    public static void postInit() {
        for (Module m : loadedModules) {
            postInit(m);
        }

        for (Module m : unloadedModules) {
            postInitNotLoaded(m);
        }
    }

    private static boolean isEnabled(Configuration config, Module m) {
        boolean defaultValue = true;
        Property prop = config.get(CATEGORY_MODULES, m.toString().toLowerCase().replace('_', '.'), defaultValue);
        return Boolean.valueOf(prop.value);
    }

    public static boolean isModuleLoaded(Module module) {
        return loadedModules.contains(module);
    }

    private static void registerModule(Module module, RailcraftModule instance) {
        modules.put(module, instance);
    }

    private static void preInit(Module module) {
        RailcraftModule instance = modules.get(module);
        if (instance != null) {
            boolean override = false;
            try {
                override = instance.getClass().getMethod("preInit").getDeclaringClass() != RailcraftModule.class;
            } catch (Exception ex) {
            }
            if (override) {
                Game.log(Level.FINER, "Pre-Init Start: {0}", instance);
                instance.preInit();
                Game.log(Level.FINER, "Pre-Init Complete: {0}", instance);
            }
        }
    }

    private static void initFirst(Module module) {
        RailcraftModule instance = modules.get(module);
        if (instance != null) {
            boolean override = false;
            try {
                override = instance.getClass().getMethod("initFirst").getDeclaringClass() != RailcraftModule.class;
            } catch (Exception ex) {
            }
            if (override) {
                Game.log(Level.FINER, "Init-First Start: {0}", instance);
                instance.initFirst();
                Game.log(Level.FINER, "Init-First Complete: {0}", instance);
            }
        }
    }

    private static void initSecond(Module module) {
        RailcraftModule instance = modules.get(module);
        if (instance != null) {
            boolean override = false;
            try {
                override = instance.getClass().getMethod("initSecond").getDeclaringClass() != RailcraftModule.class;
            } catch (Exception ex) {
            }
            if (override) {
                Game.log(Level.FINER, "Init-Second Start: {0}", instance);
                instance.initSecond();
                Game.log(Level.FINER, "Init-Second Complete: {0}", instance);
            }
        }
    }

    private static void postInit(Module module) {
        RailcraftModule instance = modules.get(module);
        if (instance != null) {
            boolean override = false;
            try {
                override = instance.getClass().getMethod("postInit").getDeclaringClass() != RailcraftModule.class;
            } catch (Exception ex) {
            }
            if (override) {
                Game.log(Level.FINER, "Post-Init Start: {0}", instance);
                instance.postInit();
                Game.log(Level.FINER, "Post-Init Complete: {0}", instance);
            }
        }
    }

    private static void postInitNotLoaded(Module module) {
        RailcraftModule instance = modules.get(module);
        if (instance != null) {
            boolean override = false;
            try {
                override = instance.getClass().getMethod("postInitNotLoaded").getDeclaringClass() != RailcraftModule.class;
            } catch (Exception ex) {
            }
            if (override) {
                Game.log(Level.FINER, "Disabled-Init Start: {0}", instance);
                instance.postInitNotLoaded();
                Game.log(Level.FINER, "Disabled-Init Complete: {0}", instance);
            }
        }
    }
}
