package railcraft.common.util.inventory;

import buildcraft.api.inventory.ISelectiveInventory;
import buildcraft.api.inventory.ISpecialInventory;
import buildcraft.api.transport.PipeManager;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import net.minecraft.block.Block;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.InventoryLargeChest;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.nbt.NBTTagString;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.tileentity.TileEntityChest;
import net.minecraft.world.World;
import net.minecraftforge.common.ForgeDirection;
import net.minecraftforge.common.ISidedInventory;
import railcraft.common.api.core.items.IItemType;
import railcraft.common.lang.RailcraftLanguage;
import railcraft.common.util.misc.EnumColor;
import railcraft.common.util.misc.Game;
import railcraft.common.util.misc.MiscTools;

public abstract class InvTools {

    private static final String TAG_ID = "id";
    private static final String TAG_SIZE = "stackSize";
    private static final String TAG_DAMAGE = "Damage";
    private static final String TAG_COUNT = "Count";
    private static final String TAG_NBT = "tag";
    private static final String TAG_SLOT = "Slot";

    public static List<IInventory> getAdjacentInventories(World world, int i, int j, int k) {
        return getAdjacentInventories(world, i, j, k, null);
    }

    public static List<IInventory> getAdjacentInventories(World world, int i, int j, int k, Class<? extends IInventory> type) {
        List<IInventory> list = new ArrayList<IInventory>(5);
        for (int side = 0; side < 6; side++) {
            IInventory inv = getInventoryFromSide(world, i, j, k, ForgeDirection.getOrientation(side), type, null);
            if (inv != null) {
                list.add(inv);
            }
        }
        return list;
    }

    public static Map<Integer, IInventory> getAdjacentInventoryMap(World world, int i, int j, int k) {
        return getAdjacentInventoryMap(world, i, j, k, null);
    }

    public static Map<Integer, IInventory> getAdjacentInventoryMap(World world, int i, int j, int k, Class<? extends IInventory> type) {
        Map<Integer, IInventory> map = new TreeMap<Integer, IInventory>();
        for (int side = 0; side < 6; side++) {
            IInventory inv = getInventoryFromSide(world, i, j, k, ForgeDirection.getOrientation(side), type, null);
            if (inv != null) {
                map.put(side, inv);
            }
        }
        return map;
    }

    public static IInventory getInventoryFromSide(World world, int x, int y, int z, ForgeDirection side, Class<? extends IInventory> type, Class<? extends IInventory> exclude) {
        x = MiscTools.getXOnSide(x, side);
        y = MiscTools.getYOnSide(y, side);
        z = MiscTools.getZOnSide(z, side);

        TileEntity tile = world.getBlockTileEntity(x, y, z);
        if (tile == null || !(tile instanceof IInventory)
                || (type != null && !type.isAssignableFrom(tile.getClass()))
                || (exclude != null && exclude.isAssignableFrom(tile.getClass()))) {
            return null;
        }

        if (!PipeManager.canExtractItems(null, world, x, y, z)) {
            return null;
        }

        IInventory inv = (IInventory) tile;
        ForgeDirection oppositeSide = side.getOpposite();
        if (inv instanceof ISelectiveInventory) {
            inv = new SelectiveInventoryMapper((ISelectiveInventory) inv, side);
        } else if (inv instanceof ISpecialInventory) {
            inv = new SpecialInventoryMapper((ISpecialInventory) inv, oppositeSide);
        } else if (inv instanceof ISidedInventory) {
            inv = new InventoryMapper((ISidedInventory) inv, oppositeSide);
        } else if (inv instanceof TileEntityChest) {
            for (int s = 2; s < 6; s++) {
                tile = MiscTools.getBlockTileEntityOnSide(world, x, y, z, ForgeDirection.getOrientation(s));
                if (tile instanceof TileEntityChest) {
                    IInventory inv2 = (IInventory) tile;
                    return new InventoryLargeChest("Large Chest", inv, inv2);
                }
            }
        }
        return inv;
    }

    public static EnumColor getItemColor(ItemStack stack) {
        if (stack == null) {
            return null;
        }
        if (stack.itemID == Block.cloth.blockID) {
            return EnumColor.fromId(15 - stack.getItemDamage());
        }
        if (stack.itemID == Item.dyePowder.itemID) {
            return EnumColor.fromId(stack.getItemDamage());
        }
        return null;
    }

    public static boolean isSynthetic(ItemStack stack) {
        NBTTagCompound nbt = stack.getTagCompound();
        if (nbt != null && nbt.hasKey("synthetic")) return true;
        return false;
    }

    public static void markItemSynthetic(ItemStack stack) {
        NBTTagCompound nbt = getItemData(stack);
        nbt.setBoolean("synthetic", true);
        NBTTagCompound display = nbt.getCompoundTag("display");
        nbt.setCompoundTag("display", display);
        NBTTagList lore = display.getTagList("Lore");
        display.setTag("Lore", lore);
        lore.appendTag(new NBTTagString("Synthetic", "\u00a77\u00a7o" + RailcraftLanguage.translate("item.synthetic")));
    }

    public static void addItemToolTip(ItemStack stack, String tag, String msg) {
        NBTTagCompound nbt = getItemData(stack);
        NBTTagCompound display = nbt.getCompoundTag("display");
        nbt.setCompoundTag("display", display);
        NBTTagList lore = display.getTagList("Lore");
        display.setTag("Lore", lore);
        lore.appendTag(new NBTTagString(tag, msg));
    }

    public static NBTTagCompound getItemData(ItemStack stack) {
        NBTTagCompound nbt = stack.getTagCompound();
        if (nbt == null) {
            nbt = new NBTTagCompound(TAG_NBT);
            stack.setTagCompound(nbt);
        }
        return nbt;
    }

    public static ItemStack depleteItem(ItemStack stack) {
        if (stack.stackSize == 1) {
            return stack.getItem().getContainerItemStack(stack);
        } else {
            stack.splitStack(1);
            return stack;
        }
    }

    public static ItemStack damageItem(ItemStack stack, int damage) {
        if (!stack.isItemStackDamageable()) return stack;
        int curDamage = stack.getItemDamage();
        curDamage += damage;
        stack.setItemDamage(curDamage);
        if (stack.getItemDamage() >= stack.getMaxDamage()) {
            stack.stackSize--;
            stack.setItemDamage(0);
        }
        if (stack.stackSize <= 0) {
            stack = null;
        }
        return stack;
    }

    public static void dropItem(ItemStack stack, World world, double x, double y, double z) {
        if (stack == null || stack.stackSize < 1) {
            return;
        }
        EntityItem entityItem = new EntityItem(world, x, y + 1.5, z, stack);
        entityItem.delayBeforeCanPickup = 10;
        world.spawnEntityInWorld(entityItem);
    }

    public static void dropInventory(IInventory inv, World world, int x, int y, int z) {
        if (Game.isNotHost(world)) return;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            ItemStack stack = inv.getStackInSlot(slot);
            if (stack != null) {
                float xOffset = MiscTools.getRand().nextFloat() * 0.8F + 0.1F;
                float yOffset = MiscTools.getRand().nextFloat() * 0.8F + 0.1F;
                float zOffset = MiscTools.getRand().nextFloat() * 0.8F + 0.1F;
                while (stack.stackSize > 0) {
                    int numToDrop = MiscTools.getRand().nextInt(21) + 10;
                    if (numToDrop > stack.stackSize) {
                        numToDrop = stack.stackSize;
                    }
                    ItemStack newStack = stack.copy();
                    newStack.stackSize = numToDrop;
                    stack.stackSize -= numToDrop;
                    EntityItem entityItem = new EntityItem(world, (float) x + xOffset, (float) y + yOffset, (float) z + zOffset, newStack);
                    float variance = 0.05F;
                    entityItem.motionX = (float) MiscTools.getRand().nextGaussian() * variance;
                    entityItem.motionY = (float) MiscTools.getRand().nextGaussian() * variance + 0.2F;
                    entityItem.motionZ = (float) MiscTools.getRand().nextGaussian() * variance;
                    world.spawnEntityInWorld(entityItem);
                }
                inv.setInventorySlotContents(slot, null);
            }
        }
    }

    public static boolean isInventoryEmpty(IInventory inv) {
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack != null) {
                break;
            }
        }
        return stack == null;
    }

    public static boolean isInventoryEmpty(IInventory inv, ForgeDirection side) {
        if (inv instanceof ISidedInventory) {
            ISidedInventory sided = (ISidedInventory) inv;
            inv = new InventoryMapper(inv, sided.getStartInventorySide(side), sided.getSizeInventorySide(side));
        }
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack != null) {
                break;
            }
        }
        return stack == null;
    }

    public static boolean isInventoryFull(IInventory inv) {
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack == null) {
                break;
            }
        }
        return stack != null;
    }

    public static boolean isInventoryFull(IInventory inv, ForgeDirection side) {
        if (inv instanceof ISidedInventory) {
            ISidedInventory sided = (ISidedInventory) inv;
            inv = new InventoryMapper(inv, sided.getStartInventorySide(side), sided.getSizeInventorySide(side));
        }
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack == null) {
                break;
            }
        }
        return stack != null;
    }

    /**
     * Counts the number of items.
     *
     * @param inv
     * @return
     */
    public static int countItems(IInventory inv) {
        int count = 0;
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack != null) {
                count += stack.stackSize;
            }
        }
        return count;
    }

    public static int countItems(IInventory inv, ForgeDirection side) {
        if (inv instanceof ISidedInventory) {
            ISidedInventory sided = (ISidedInventory) inv;
            inv = new InventoryMapper(inv, sided.getStartInventorySide(side), sided.getSizeInventorySide(side));
        }
        return countItems(inv);
    }

    public static int countItems(IInventory inv, IItemType filter) {
        int count = 0;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            ItemStack stack = inv.getStackInSlot(slot);
            if (stack != null && filter.isItemType(stack)) {
                count += stack.stackSize;
            }
        }
        return count;
    }

    /**
     * Counts the number of items that match the filter. Ignores
     * ISpecialInventory and ISidedInventory, and bypasses InventoryMapper to
     * count from the base IInventory.
     *
     * @param inv
     * @param filters
     * @return
     */
    public static int countItems(IInventory inv, ItemStack... filters) {
        if (inv instanceof InventoryMapper) {
            inv = ((InventoryMapper) inv).getBaseInventory();
        }
        boolean hasFilter = false;
        for (ItemStack filter : filters) {
            if (filter != null) {
                hasFilter = true;
                break;
            }
        }

        if (!hasFilter) {
            return countItems(inv);
        }

        int count = 0;
        ItemStack stack = null;
        for (int slot = 0; slot < inv.getSizeInventory(); slot++) {
            stack = inv.getStackInSlot(slot);
            if (stack != null) {
                for (ItemStack filter : filters) {
                    if (filter != null && isItemEqual(stack, filter)) {
                        count += stack.stackSize;
                        break;
                    }
                }
            }
        }
        return count;
    }

    public static int countItems(Collection<IInventory> inventories, ItemStack... filter) {
        int count = 0;
        for (IInventory inv : inventories) {
            count += InvTools.countItems(inv, filter);
        }
        return count;
    }

    /**
     * Returns true if the inventory contains the specified item.
     *
     * @param inv The IIinventory to check
     * @param item The ItemStack to look for
     * @return true is exists
     */
    public static boolean containsItem(IInventory inv, ItemStack item) {
        return countItems(inv, item) > 0;
    }

    /**
     * Returns a map backed by an
     * <code>ItemStackMap</code> that lists the total number of each type of
     * item in the inventory.
     *
     * @param inv The <code>IInventory</code> to generate the manifest for
     * @return A <code>Map</code> that lists how many of each item is in * * *
     * the <code>IInventory</code>
     * @see ItemStackMap
     */
    public static Map<ItemStack, Integer> getManifest(IInventory inv) {
        Map<ItemStack, Integer> manifest = new ItemStackMap<Integer>();
        for (int i = 0; i < inv.getSizeInventory(); i++) {
            ItemStack slot = inv.getStackInSlot(i);
            if (slot != null) {
                Integer count = manifest.get(slot);
                if (count == null) {
                    count = 0;
                }
                count += slot.stackSize;
                manifest.put(slot, count);
            }
        }
        return manifest;
    }

    /**
     * Attempts to move a single item from one inventory to another.
     *
     * @param source
     * @param dest
     * @return null if nothing was moved, the stack moved otherwise
     */
    public static ItemStack moveOneItem(IInventory source, IInventory dest) {
        if (source instanceof ISpecialInventory) {
            ItemStack[] extracted = ((ISpecialInventory) source).extractItem(false, ForgeDirection.UNKNOWN, 1);
            if (extracted != null && extracted.length > 0) {
                ItemStack result = moveItemStack(extracted[0], dest);
                if (result == null) {
                    ((ISpecialInventory) source).extractItem(true, ForgeDirection.UNKNOWN, 1);
                    return extracted[0];
                }
            }
            return null;
        }
        for (int inputSlot = 0; inputSlot < source.getSizeInventory(); inputSlot++) {
            ItemStack inputStack = source.getStackInSlot(inputSlot);
            if (inputStack == null) {
                continue;
            }
            ItemStack stackToMove = inputStack.copy();
            stackToMove.stackSize = 1;
            ItemStack result = moveItemStack(stackToMove, dest);
            if (result == null) {
                source.decrStackSize(inputSlot, 1);
                return stackToMove;
            }
        }
        return null;
    }

    /**
     * Attempts to move a single item from one inventory to another.
     *
     * @param source
     * @param dest
     * @param filer an ItemStack[] to match against
     * @return null if nothing was moved, the stack moved otherwise
     */
    public static ItemStack moveOneItem(IInventory source, IInventory dest, ItemStack... filters) {
        boolean hasFilter = false;
        for (ItemStack filter : filters) {
            if (filter != null) {
                hasFilter = true;
                break;
            }
        }
        if (hasFilter && source instanceof ISelectiveInventory) {
            ItemStack[] extracted = ((ISelectiveInventory) source).extractItem(filters, false, false, ForgeDirection.UNKNOWN, 1);
            if (extracted != null && extracted.length == 1) {
                if (isItemEqual(extracted[0], filters)) {
                    ItemStack result = moveItemStack(extracted[0], dest);
                    if (result == null) {
                        ((ISelectiveInventory) source).extractItem(filters, true, false, ForgeDirection.UNKNOWN, 1);
                        return extracted[0];
                    }
                }
            }
            return null;
        } else if (source instanceof ISpecialInventory) {
            ItemStack[] extracted = ((ISpecialInventory) source).extractItem(false, ForgeDirection.UNKNOWN, 1);
            if (extracted != null && extracted.length > 0) {
                if (!hasFilter || isItemEqual(extracted[0], filters)) {
                    ItemStack result = moveItemStack(extracted[0], dest);
                    if (result == null) {
                        ((ISpecialInventory) source).extractItem(true, ForgeDirection.UNKNOWN, 1);
                        return extracted[0];
                    }
                }
            }
            return null;
        }
        for (int inputSlot = 0; inputSlot < source.getSizeInventory(); inputSlot++) {
            ItemStack inputStack = source.getStackInSlot(inputSlot);
            if (inputStack == null || (hasFilter && !isItemEqual(inputStack, filters))) {
                continue;
            }
            ItemStack stackToMove = inputStack.copy();
            stackToMove.stackSize = 1;
            ItemStack result = moveItemStack(stackToMove, dest);
            if (result == null) {
                source.decrStackSize(inputSlot, 1);
                return stackToMove;
            }
        }
        return null;
    }

    /**
     * Attempts to move a single item from one inventory to another.
     *
     * @param source
     * @param dest
     * @param filer an IItemType to match against
     * @return null if nothing was moved, the stack moved otherwise
     */
    public static ItemStack moveOneItem(IInventory source, IInventory dest, IItemType filter) {
        if (source instanceof ISpecialInventory) {
            ItemStack[] extracted = ((ISpecialInventory) source).extractItem(false, ForgeDirection.UNKNOWN, 1);
            if (extracted != null && extracted.length > 0) {
                if (filter.isItemType(extracted[0])) {
                    ItemStack result = moveItemStack(extracted[0], dest);
                    if (result == null) {
                        ((ISpecialInventory) source).extractItem(true, ForgeDirection.UNKNOWN, 1);
                        return extracted[0];
                    }
                }
            }
            return null;
        }
        for (int inputSlot = 0; inputSlot < source.getSizeInventory(); inputSlot++) {
            ItemStack inputStack = source.getStackInSlot(inputSlot);
            if (inputStack == null || !filter.isItemType(inputStack)) {
                continue;
            }
            ItemStack stackToMove = inputStack.copy();
            stackToMove.stackSize = 1;
            ItemStack result = moveItemStack(stackToMove, dest);
            if (result == null) {
                source.decrStackSize(inputSlot, 1);
                return stackToMove;
            }
        }
        return null;
    }

    /**
     * Attempts to move one item from a collection of inventories.
     *
     * @param sources
     * @param dest
     * @param filters
     * @return
     */
    public static ItemStack moveOneItem(List<IInventory> sources, IInventory dest, ItemStack... filters) {
        for (IInventory inv : sources) {
            ItemStack moved = InvTools.moveOneItem(inv, dest, filters);
            if (moved != null) {
                return moved;
            }
        }
        return null;
    }

    /**
     * Attempts to move one item to a collection of inventories.
     *
     * @param sources
     * @param dest
     * @param filters
     * @return
     */
    public static ItemStack moveOneItem(IInventory source, List<IInventory> destinations, ItemStack... filters) {
        for (IInventory dest : destinations) {
            ItemStack moved = InvTools.moveOneItem(source, dest, filters);
            if (moved != null) {
                return moved;
            }
        }
        return null;
    }

    /**
     * Attempts to move a single item from one inventory to another.
     *
     * Will not move any items in the filter.
     *
     * @param source
     * @param dest
     * @param filer an ItemStack[] to exclude
     * @return null if nothing was moved, the stack moved otherwise
     */
    public static ItemStack moveOneItemExcept(IInventory source, IInventory dest, ItemStack... filters) {
        boolean hasFilter = false;
        for (ItemStack filter : filters) {
            if (filter != null) {
                hasFilter = true;
                break;
            }
        }
        if (source instanceof ISpecialInventory) {
            ItemStack[] extracted = ((ISpecialInventory) source).extractItem(false, ForgeDirection.UNKNOWN, 1);
            if (extracted != null && extracted.length > 0) {
                if (!hasFilter || !isItemEqual(extracted[0], filters)) {
                    ItemStack result = moveItemStack(extracted[0], dest);
                    if (result == null) {
                        ((ISpecialInventory) source).extractItem(true, ForgeDirection.UNKNOWN, 1);
                        return extracted[0];
                    }
                }
            }
            return null;
        }
        for (int inputSlot = 0; inputSlot < source.getSizeInventory(); inputSlot++) {
            ItemStack inputStack = source.getStackInSlot(inputSlot);
            if (inputStack == null || (hasFilter && isItemEqual(inputStack, filters))) {
                continue;
            }
            ItemStack stackToMove = inputStack.copy();
            stackToMove.stackSize = 1;
            ItemStack result = moveItemStack(stackToMove, dest);
            if (result == null) {
                source.decrStackSize(inputSlot, 1);
                return stackToMove;
            }
        }
        return null;
    }

    /**
     * Attempts to move one item from a collection of inventories.
     *
     * @param sources
     * @param dest
     * @param filters
     * @return
     */
    public static ItemStack moveOneItemExcept(List<IInventory> sources, IInventory dest, ItemStack... filters) {
        for (IInventory inv : sources) {
            ItemStack moved = InvTools.moveOneItemExcept(inv, dest, filters);
            if (moved != null) {
                return moved;
            }
        }
        return null;
    }

    /**
     * Attempts to move one item to a collection of inventories.
     *
     * @param sources
     * @param dest
     * @param filters
     * @return
     */
    public static ItemStack moveOneItemExcept(IInventory source, List<IInventory> destinations, ItemStack... filters) {
        for (IInventory dest : destinations) {
            ItemStack moved = InvTools.moveOneItemExcept(source, dest, filters);
            if (moved != null) {
                return moved;
            }
        }
        return null;
    }

    /**
     * A more robust item comparison function. Supports items with damage = -1
     * matching any sub-type.
     *
     * @param a An ItemStack
     * @param b An ItemStack
     * @return True if equal
     */
    public static boolean isItemEqual(ItemStack a, ItemStack b) {
        if (a == null || b == null) {
            return false;
        }
        if (a.itemID != b.itemID) {
            return false;
        }
        if (a.stackTagCompound != null && !a.stackTagCompound.equals(b.stackTagCompound)) {
            return false;
        }
        if (a.getHasSubtypes()) {
            if (a.getItemDamage() == -1 || b.getItemDamage() == -1) {
                return true;
            }
            if (a.getItemDamage() != b.getItemDamage()) {
                return false;
            }
        }
        return true;
    }

    /**
     * A more robust item comparison function.
     *
     * Compares stackSize as well.
     *
     * Two null stacks will return true, unlike the other functions.
     *
     * This function is primarily intended to be used to track changes to an
     * ItemStack.
     *
     * @param a An ItemStack
     * @param b An ItemStack
     * @return True if equal
     */
    public static boolean isItemEqualStrict(ItemStack a, ItemStack b) {
        if (a == null && b == null) {
            return true;
        }
        if (a == null || b == null) {
            return false;
        }
        if (a.itemID != b.itemID) {
            return false;
        }
        if (a.stackSize != b.stackSize) {
            return false;
        }
        if (a.getHasSubtypes() && a.getItemDamage() != b.getItemDamage()) {
            return false;
        }
        if (a.stackTagCompound != null && !a.stackTagCompound.equals(b.stackTagCompound)) {
            return false;
        }
        return true;
    }

    /**
     * A more robust item comparison function. Supports items with damage = -1
     * matching any sub-type.
     *
     * @param a An ItemStack
     * @param b An ItemStack
     * @return True if equal
     */
    public static boolean isItemEqualIgnoreNBT(ItemStack a, ItemStack b) {
        if (a == null || b == null) {
            return false;
        }
        if (a.itemID != b.itemID) {
            return false;
        }
        if (a.getHasSubtypes()) {
            if ((a.getItemDamage() == -1 || b.getItemDamage() == -1)) {
                return true;
            }
            if (a.getItemDamage() != b.getItemDamage()) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns true if the item is equal to any one of several possible matches.
     *
     * @param stack
     * @param matches
     * @return
     */
    public static boolean isItemEqual(ItemStack stack, ItemStack... matches) {
        for (ItemStack match : matches) {
            if (isItemEqual(stack, match)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns true if the item is equal to any one of several possible matches.
     *
     * @param stack
     * @param matches
     * @return
     */
    public static boolean isItemEqual(ItemStack stack, Collection<ItemStack> matches) {
        for (ItemStack match : matches) {
            if (isItemEqual(stack, match)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Places an ItemStack in a destination IInventory. Will attempt to move as
     * much of the stack as possible, returning any remainder.
     *
     * @param stack The ItemStack to put in the inventory.
     * @param dest The destination IInventory.
     * @return Null if itemStack was completely moved, a new itemStack with
     * remaining stackSize if part or none of the stack was moved.
     */
    public static ItemStack moveItemStack(ItemStack stack, IInventory dest) {
        if (stack == null) {
            return null;
        }
        stack = stack.copy();
        if (dest == null) {
            return stack;
        }
        if (dest instanceof ISpecialInventory) {
            int used = ((ISpecialInventory) dest).addItem(stack, true, ForgeDirection.UNKNOWN);
            if (used >= stack.stackSize) {
                return null;
            } else {
                stack.stackSize -= used;
                return stack;
            }
        }
        boolean movedItem;
        do {
            movedItem = false;
            ItemStack destStack;
            for (int ii = 0; ii < dest.getSizeInventory(); ii++) {
                destStack = dest.getStackInSlot(ii);
                if (destStack != null && isItemEqual(destStack, stack)) {
                    destStack = destStack.copy();
                    int maxStack = Math.min(destStack.getMaxStackSize(), dest.getInventoryStackLimit());
                    int room = maxStack - destStack.stackSize;
                    if (room > 0) {
                        int move = Math.min(room, stack.stackSize);
                        destStack.stackSize += move;
                        stack.stackSize -= move;
                        dest.setInventorySlotContents(ii, destStack);
                        if (stack.stackSize <= 0) {
                            return null;
                        }
                        movedItem = true;
                    }
                }
            }
            if (!movedItem) {
                for (int ii = 0; ii < dest.getSizeInventory(); ii++) {
                    destStack = dest.getStackInSlot(ii);
                    if (destStack == null) {
                        if (stack.stackSize > dest.getInventoryStackLimit()) {
                            dest.setInventorySlotContents(ii, stack.splitStack(dest.getInventoryStackLimit()));
                        } else {
                            dest.setInventorySlotContents(ii, stack);
                            return null;
                        }
                        movedItem = true;
                    }
                }
            }
        } while (movedItem);
        return stack;
    }

//    /**
//     * Places an ItemStack in a destination IInventory. Will attempt to move as
//     * much of the stack as possible, returning any remainder.
//     * @param stack The ItemStack to put in the inventory.
//     * @param dest The destination IInventory.
//     * @return Null if itemStack was completely moved, the itemStack with remaining stackSize if part or none of the stack was moved.
//     */
//    public static ItemStack moveItemStack(ItemStack stack, IInventory dest, int side)
//    {
//        if(stack == null) {
//            return null;
//        }
//        stack = stack.copy();
//        if(dest == null) {
//            return stack;
//        }
//        if(dest instanceof ISpecialInventory) {
//            ((ISpecialInventory)dest).addItem(stack, true, ForgeDirection.values()[side]);
//            if(stack.stackSize <= 0) {
//                return null;
//            } else {
//                return stack;
//            }
//        }
//        if(dest instanceof ISidedInventory) {
//            ISidedInventory sided = (ISidedInventory)dest;
//            dest = new InventoryMapper(dest, sided.getStartInventorySide(side), sided.getSizeInventorySide(side));
//        }
//        return moveItemStack(stack, dest);
//    }
    /**
     * Checks if there is room for the ItemStack in the inventory.
     *
     * @param stack The ItemStack
     * @param dest The IInventory
     * @return true if room for stack
     */
    public static boolean isRoomForStack(ItemStack stack, IInventory dest) {
        if (stack == null) {
            return false;
        }
        if (dest == null) {
            return false;
        }
        if (dest instanceof ISpecialInventory) {
            return ((ISpecialInventory) dest).addItem(stack, false, ForgeDirection.UNKNOWN) > 0;
        }

        for (int ii = 0; ii < dest.getSizeInventory(); ii++) {
            ItemStack slot = dest.getStackInSlot(ii);
            if (slot == null) {
                return true;
            } else if (isItemEqual(slot, stack)) {
                int maxStack = Math.min(slot.getMaxStackSize(), dest.getInventoryStackLimit());
                int room = maxStack - slot.stackSize;
                if (room >= stack.stackSize) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Removes a up to numItems worth of items from the inventory, not caring
     * about what the items are.
     *
     * @param inv
     * @param numItems
     * @return
     */
    public static ItemStack[] removeItems(IInventory inv, int numItems) {
        if (inv instanceof ISpecialInventory) {
            return ((ISpecialInventory) inv).extractItem(true, ForgeDirection.UNKNOWN, numItems);
        }
        StandaloneInventory output = new StandaloneInventory(27);
        for (int i = 0; i < inv.getSizeInventory(); i++) {
            if (numItems <= 0) {
                break;
            }
            ItemStack slot = inv.getStackInSlot(i);
            if (slot == null) {
                continue;
            }
            ItemStack removed = inv.decrStackSize(i, numItems);
            numItems -= removed.stackSize;
            ItemStack remainder = moveItemStack(removed, output);
            if (remainder != null) {
                moveItemStack(remainder, inv);
                numItems += remainder.stackSize;
                break;
            }
        }

        List<ItemStack> list = new LinkedList<ItemStack>();
        for (ItemStack stack : output.getContents()) {
            if (stack != null) {
                list.add(stack);
            }
        }
        return list.toArray(new ItemStack[0]);
    }

    /**
     * Removes and returns a single item from the inventory.
     *
     * @param inv The inventory
     * @return An ItemStack
     */
    public static ItemStack removeOneItem(IInventory inv) {
        for (int i = 0; i < inv.getSizeInventory(); i++) {
            ItemStack slot = inv.getStackInSlot(i);
            if (slot != null) {
                return inv.decrStackSize(i, 1);
            }
        }
        return null;
    }

    /**
     * Removes and returns a single item from the inventory that matches the
     * filter.
     *
     * @param inv The inventory
     * @param filter ItemStack to match against
     * @return An ItemStack
     */
    public static ItemStack removeOneItem(IInventory inv, ItemStack filter) {
        for (int i = 0; i < inv.getSizeInventory(); i++) {
            ItemStack slot = inv.getStackInSlot(i);
            if (slot != null && isItemEqual(slot, filter)) {
                return inv.decrStackSize(i, 1);
            }
        }
        return null;
    }

    /**
     * Removes and returns a single item from the inventory that matches the
     * filter.
     *
     * @param inv The inventory
     * @param filter EnumItemType to match against
     * @return An ItemStack
     */
    public static ItemStack removeOneItem(IInventory inv, IItemType filter) {
        for (int i = 0; i < inv.getSizeInventory(); i++) {
            ItemStack slot = inv.getStackInSlot(i);
            if (slot != null && filter.isItemType(slot)) {
                return inv.decrStackSize(i, 1);
            }
        }
        return null;
    }

    public static void writeInvToNBT(IInventory inv, String tag, NBTTagCompound data) {
        NBTTagList list = new NBTTagList();
        for (byte slot = 0; slot < inv.getSizeInventory(); slot++) {
            ItemStack stack = inv.getStackInSlot(slot);
            if (stack != null) {
                NBTTagCompound itemTag = new NBTTagCompound();
                itemTag.setByte(TAG_SLOT, slot);
                writeItemToNBT(stack, itemTag);
                list.appendTag(itemTag);
            }
        }
        data.setTag(tag, list);
    }

    public static void readInvFromNBT(IInventory inv, String tag, NBTTagCompound data) {
        NBTTagList list = data.getTagList(tag);
        for (byte entry = 0; entry < list.tagCount(); entry++) {
            NBTTagCompound itemTag = (NBTTagCompound) list.tagAt(entry);
            int slot = itemTag.getByte(TAG_SLOT);
            if (slot >= 0 && slot < inv.getSizeInventory()) {
                ItemStack stack = readItemFromNBT(itemTag);
                inv.setInventorySlotContents(slot, stack);
            }
        }
    }

    public static void writeItemToNBT(ItemStack stack, NBTTagCompound data) {
        if (stack == null || stack.stackSize <= 0) {
            return;
        }
        if (stack.stackSize > 127) {
            stack.stackSize = 127;
        }
        stack.writeToNBT(data);
    }

    public static ItemStack readItemFromNBT(NBTTagCompound data) {
        int itemID = data.getShort(TAG_ID);
        int stackSize = data.getShort(TAG_SIZE);
        int damage = data.getShort(TAG_DAMAGE);

        if (stackSize <= 0) {
            stackSize = data.getByte(TAG_COUNT);
        }

        if (itemID <= 0 || stackSize <= 0) {
            return null;
        }

        ItemStack stack = new ItemStack(itemID, stackSize, damage);

        if (data.hasKey(TAG_NBT)) {
            stack.stackTagCompound = data.getCompoundTag(TAG_NBT);
        }

        return stack.getItem() != null ? stack : null;
    }
}
