package railcraft.common.util.inventory;

import com.google.common.collect.Iterators;
import java.util.Iterator;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import railcraft.common.blocks.RailcraftTileEntity;
import railcraft.common.lang.RailcraftLanguage;

/**
 * Creates a standalone instance of IInventory.
 *
 * Useful for hiding parts of an inventory from outsiders.
 *
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public class StandaloneInventory implements IInventory, Iterable<ItemStack> {

    private Callback callback;
    private final String name;
    private ItemStack contents[];

    public StandaloneInventory(int size, String name, IInventory callback) {
        this.name = name;
        contents = new ItemStack[size];
        this.callback = callback == null ? null : new InventoryCallback(callback);
    }

    public StandaloneInventory(int size, String name, RailcraftTileEntity callback) {
        this.name = name;
        contents = new ItemStack[size];
        this.callback = callback == null ? null : new TileCallback(callback);
    }

    public StandaloneInventory(int size, IInventory callback) {
        this(size, null, callback);
    }

    public StandaloneInventory(int size, RailcraftTileEntity callback) {
        this(size, null, callback);
    }

    public StandaloneInventory(int size, String name) {
        this(size, name, (RailcraftTileEntity) null);
    }

    public StandaloneInventory(int size) {
        this(size, null, (RailcraftTileEntity) null);
    }

    @Override
    public int getSizeInventory() {
        return contents.length;
    }

    @Override
    public ItemStack getStackInSlot(int i) {
        return contents[i];
    }

    @Override
    public ItemStack decrStackSize(int i, int j) {
        if (contents[i] != null) {
            if (contents[i].stackSize <= j) {
                ItemStack itemstack = contents[i];
                contents[i] = null;
                onInventoryChanged();
                return itemstack;
            }
            ItemStack itemstack1 = contents[i].splitStack(j);
            if (contents[i].stackSize <= 0) {
                contents[i] = null;
            }
            onInventoryChanged();
            return itemstack1;
        } else {
            return null;
        }
    }

    @Override
    public void setInventorySlotContents(int i, ItemStack itemstack) {
        contents[i] = itemstack;
        if (itemstack != null && itemstack.stackSize > getInventoryStackLimit()) {
            itemstack.stackSize = getInventoryStackLimit();
        }
        onInventoryChanged();
    }

    @Override
    public String getInvName() {
        if (name != null) {
            return RailcraftLanguage.translate(name);
        }
        if (callback != null) {
            return callback.getInvName();
        }
        return invTypeName();
    }

    protected String invTypeName() {
        return "Standalone";
    }

    @Override
    public int getInventoryStackLimit() {
        return 64;
    }

    @Override
    public void onInventoryChanged() {
        if (callback != null) {
            callback.onInventoryChanged();
        }
    }

    @Override
    public boolean isUseableByPlayer(EntityPlayer entityplayer) {
        if (callback != null) {
            return callback.isUseableByPlayer(entityplayer);
        }
        return true;
    }

    @Override
    public void openChest() {
        if (callback != null) {
            callback.openChest();
        }
    }

    @Override
    public void closeChest() {
        if (callback != null) {
            callback.closeChest();
        }
    }

    @Override
    public ItemStack getStackInSlotOnClosing(int slot) {
        return null;
    }

    public ItemStack[] getContents() {
        return contents;
    }

    public void writeToNBT(String tag, NBTTagCompound data) {
        InvTools.writeInvToNBT(this, tag, data);
    }

    public void readFromNBT(String tag, NBTTagCompound data) {
        InvTools.readInvFromNBT(this, tag, data);
    }

    @Override
    public Iterator<ItemStack> iterator() {
        return Iterators.forArray(contents);
    }

    private abstract class Callback {

        public boolean isUseableByPlayer(EntityPlayer entityplayer) {
            return true;
        }

        public void openChest() {
        }

        public void closeChest() {
        }

        public abstract void onInventoryChanged();

        public abstract String getInvName();
    }

    private class InventoryCallback extends Callback {

        private IInventory inv;

        public InventoryCallback(IInventory inv) {
            this.inv = inv;
        }

        @Override
        public boolean isUseableByPlayer(EntityPlayer entityplayer) {
            return inv.isUseableByPlayer(entityplayer);
        }

        @Override
        public void openChest() {
            inv.openChest();
        }

        @Override
        public void closeChest() {
            inv.closeChest();
        }

        @Override
        public void onInventoryChanged() {
            inv.onInventoryChanged();
        }

        @Override
        public String getInvName() {
            return inv.getInvName();
        }
    }

    private class TileCallback extends Callback {

        private RailcraftTileEntity inv;

        public TileCallback(RailcraftTileEntity inv) {
            this.inv = inv;
        }

        @Override
        public void onInventoryChanged() {
            inv.onInventoryChanged();
        }

        @Override
        public String getInvName() {
            return inv.getInvName();
        }
    }
}
